//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-2000, O.Mizno     All rights reserved.
//
//     Dib.cpp
//       DibStream class
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <math.h>
#include "Dib.h"

#define WIDTHBYTES(i)           ((unsigned)((i+31)&(~31))/8)
#define DibWidthBytesN(lpbi, n) (UINT)WIDTHBYTES((UINT)((LPBITMAPINFOHEADER)(lpbi))->biWidth * (UINT)(n))
#define DibWidthBytes(lpbi)     DibWidthBytesN(lpbi, ((LPBITMAPINFOHEADER)(lpbi))->biBitCount)


static const BYTE aDividedBy51[256] =
{
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
  1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
  2, 2, 2, 2, 2, 2, 2, 2, 2, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
  3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
  4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 5,
};

static const BYTE aModulo51[256] =
{
  0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19,
  20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37,
  38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0, 1, 2, 3, 4, 5, 6,
  7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25,
  26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43,
  44, 45, 46, 47, 48, 49, 50, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12,
  13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30,
  31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48,
  49, 50, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17,
  18, 19, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35,
  36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0, 1, 2, 3,
  4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22,
  23, 24, 25, 26, 27, 28, 29, 30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40,
  41, 42, 43, 44, 45, 46, 47, 48, 49, 50, 0,
};

static const BYTE aTimes6[6] =
{
  0, 6, 12, 18, 24, 30
};

static const BYTE aTimes36[6] =
{
  0, 36, 72, 108, 144, 180
};

static const BYTE aHalftone8x8[64] =
{
   0, 38,  9, 47,  2, 40, 11, 50,
  25, 12, 35, 22, 27, 15, 37, 24,
   6, 44,  3, 41,  8, 47,  5, 43,
  31, 19, 28, 15, 34, 21, 31, 18,
   1, 39, 11, 49,  0, 39, 10, 48,
  27, 14, 36, 23, 26, 13, 35, 23,
   7, 46,  4, 43,  7, 45,  3, 42,
  33, 20, 30, 17, 32, 19, 29, 16,
};


// sum=256
#define BRIGHT_R   73
#define BRIGHT_G  146
#define BRIGHT_B   37


static void *MemAlloc(int type, int size)
{
	switch(type) {
	case DIB_MEMORYTYPE_HEAP:
		return malloc(size);

	case DIB_MEMORYTYPE_LOCALLOCKED:
			{
		HLOCAL hLocal = LocalAlloc(LMEM_MOVEABLE, size);
		if(hLocal != NULL)
			return LocalLock(hLocal);
		break;
			}

	case DIB_MEMORYTYPE_GLOBALLOCKED:
			{
		HGLOBAL hGlobal = GlobalAlloc(GMEM_MOVEABLE, size);
		if(hGlobal != NULL)
			return GlobalLock(hGlobal);
		break;
			}

	case DIB_MEMORYTYPE_LOCALFIXED:
		return LocalAlloc(LMEM_FIXED, size);

	case DIB_MEMORYTYPE_GLOBALFIXED:
		return GlobalAlloc(GMEM_FIXED, size);;
	}

	return NULL;
}


#define HEADER_MEMTYPE   DIB_MEMORYTYPE_GLOBALFIXED
#define BUFFER_MEMTYPE   DIB_MEMORYTYPE_GLOBALFIXED


//---------------------------------------------------------------------------
//  DibStream::DibStream
//---------------------------------------------------------------------------
DibStream::DibStream()
{
	RefCount = 1;
	Header = Buffer = NULL;
	HeaderMemoryType = BufferMemoryType = DIB_MEMORYTYPE_NONE;
}	//DibStream::DibStream()

//---------------------------------------------------------------------------
//  DibStream::~DibStream
//---------------------------------------------------------------------------
DibStream::~DibStream()
{
	HeaderFree();
	BufferFree();
}	//DibStream::~DibStream()

//---------------------------------------------------------------------------
//  DibStream::AddRef
//---------------------------------------------------------------------------
ULONG DibStream::AddRef(void)
{
	return ++RefCount;
}

//---------------------------------------------------------------------------
//  DibStream::Release
//---------------------------------------------------------------------------
ULONG DibStream::Release(void)
{
	if(--RefCount == 0){
		delete this;
		return 0;
	}
	return RefCount;
}

//---------------------------------------------------------------------------
//  DibStream::HeaderFree          1999/10/31
//---------------------------------------------------------------------------
void DibStream::HeaderFree(void)
{
	if(Header == NULL) return;

	switch(HeaderMemoryType){
	case DIB_MEMORYTYPE_HEAP:
		free(Header);
		break;
	case DIB_MEMORYTYPE_LOCALLOCKED:{
		HLOCAL hlocal = LocalHandle(Header);
		if(hlocal != NULL){
			LocalUnlock(hlocal);
			LocalFree(hlocal);
		}
		break;}
	case DIB_MEMORYTYPE_GLOBALLOCKED:{
		HGLOBAL hglobal = GlobalHandle(Header);
		if(hglobal != NULL){
			GlobalUnlock(hglobal);
			GlobalFree(hglobal);
		}
		break;}
	case DIB_MEMORYTYPE_LOCALFIXED:
		LocalFree((HLOCAL)Header);
		break;
	case DIB_MEMORYTYPE_GLOBALFIXED:
		GlobalFree((HGLOBAL)Header);
		break;
	}
	Header = NULL;
	HeaderMemoryType = DIB_MEMORYTYPE_NONE;
}	//DibStream::HeaderFree()

//---------------------------------------------------------------------------
//  DibStream::BufferFree          1999/11/1
//---------------------------------------------------------------------------
void DibStream::BufferFree(void)
{
	if(Buffer == NULL) return;

	switch(BufferMemoryType){
	case DIB_MEMORYTYPE_HEAP:
		free(Buffer);
		break;
	case DIB_MEMORYTYPE_LOCALLOCKED:{
		HLOCAL hlocal = LocalHandle(Buffer);
		if(hlocal != NULL){
			LocalUnlock(hlocal);
			LocalFree(hlocal);
		}
		break;}
	case DIB_MEMORYTYPE_GLOBALLOCKED:{
		HGLOBAL hglobal = GlobalHandle(Buffer);
		if(hglobal != NULL){
			GlobalUnlock(hglobal);
			GlobalFree(hglobal);
		}
		break;}
	case DIB_MEMORYTYPE_LOCALFIXED:
		LocalFree((HLOCAL)Buffer);
		break;
	case DIB_MEMORYTYPE_GLOBALFIXED:
		GlobalFree((HGLOBAL)Buffer);
		break;
	}
	Buffer = NULL;
	BufferMemoryType = DIB_MEMORYTYPE_NONE;
}	//DibStream::BufferFree()

//---------------------------------------------------------------------------
//  DibStream::Clear               2000/3/11
//---------------------------------------------------------------------------
void DibStream::Clear(void)
{
	HeaderFree();
	BufferFree();
}	//DibStream::Clear()

//---------------------------------------------------------------------------
//  DibStream::Clone               1999/10/31
//---------------------------------------------------------------------------
DibStream *DibStream::Clone(void)
{
	DibStream *Dest = new DibStream;
	if(Header != NULL){
		int size = sizeof(BITMAPINFOHEADER) /*+ sizeof(DWORD) * 3*/
			+ GetNumColors() * sizeof(RGBQUAD);
		Dest->Header = MemAlloc(
			Dest->HeaderMemoryType = HEADER_MEMTYPE,
			size);
		memcpy(Dest->Header, Header, size);
	}
	if(Buffer != NULL){
		Dest->Buffer = MemAlloc(
			Dest->BufferMemoryType = BUFFER_MEMTYPE,
			GetImageSize());
		memcpy(Dest->Buffer, Buffer, GetImageSize());
	}
	return Dest;
}	//DibStream::Clone()

//---------------------------------------------------------------------------
//  DibStream::SetDib
//---------------------------------------------------------------------------
bool DibStream::SetDib(LPVOID SourceHeader, DWORD SHMemType, LPVOID SourceBuffer, DWORD SBMemType)
{
	if(SourceHeader == NULL || SourceBuffer == NULL)
		return false;

	HeaderFree();
	BufferFree();

	Header = SourceHeader;
	HeaderMemoryType = SHMemType;

	Buffer = SourceBuffer;
	BufferMemoryType = SBMemType;

	return true;
}	//DibStream::SetDib()

//---------------------------------------------------------------------------
//  DibStream::SetBuffer
//---------------------------------------------------------------------------
bool DibStream::SetBuffer(LPVOID SourceBuffer, DWORD SBMemType)
{
	BufferFree();

	Buffer = SourceBuffer;
	BufferMemoryType = SBMemType;

	return true;
}	//DibStream::SetBuffer()

//---------------------------------------------------------------------------
//  DibStream::Copy                1999/10/31
//---------------------------------------------------------------------------
bool DibStream::Copy(LPVOID SourceHeader, LPVOID SourceBuffer)
{
	HeaderFree();
	BufferFree();

#define BIH ((LPBITMAPINFOHEADER)SourceHeader)

	int ColNum = (BIH->biClrUsed == 0 && BIH->biBitCount <= 8
			? (int)(1 << (int)BIH->biBitCount)
			: (int)BIH->biClrUsed);
	Header = MemAlloc(
			HeaderMemoryType = HEADER_MEMTYPE,
			sizeof(BITMAPINFOHEADER) + ColNum*sizeof(RGBQUAD));
	if(Header == NULL) return false;
	memcpy(Header, SourceHeader, sizeof(BITMAPINFOHEADER) + ColNum*sizeof(RGBQUAD));

	int Size = (BIH->biSizeImage == 0
			? ((DWORD)(UINT)DibWidthBytes(SourceHeader) * (DWORD)abs(BIH->biHeight))
			: BIH->biSizeImage);
	Buffer = MemAlloc(
		BufferMemoryType = BUFFER_MEMTYPE,
		Size);
	if(Buffer == NULL) {
		HeaderFree();
		return false;
	}
	memcpy(Buffer, SourceBuffer, Size);

#undef BIH
	return true;
}	//DibStream::Copy()

//---------------------------------------------------------------------------
//  DibStream::CreateAlphaDib      1999/12/28
//---------------------------------------------------------------------------
DibStream *DibStream::CreateAlphaDib(void)
{
	// if not 32-bit, fail
	if(GetBpp() != 32)
		return NULL;

	// create monochrome DIB
	DibStream *Dest = new DibStream;
	Dest->Create(GetWidth(), GetHeader()->biHeight, 8);
	Dest->SetMonoPalette();

	// set alpha image
	int w = GetWidth();
	int h = GetHeight();
	int spitch = GetPitch();
	int dpitch = Dest->GetPitch();
	BYTE *src = (BYTE *)GetBuffer();
	BYTE *dst = (BYTE *)Dest->GetBuffer();
	for(int y=0; y<h; y++,src+=spitch,dst+=dpitch) {
		DibBGRA *sp = (DibBGRA *)src;
		BYTE *dp = dst;
		for(int x=0; x<w; x++,sp++,dp++)
			*dp = sp->a;
	}

	return Dest;
}	//DibStream::CreateAlphaDib()

//---------------------------------------------------------------------------
//  DibStream::SetAlphaPlane       2000/5/21
//---------------------------------------------------------------------------
bool DibStream::SetAlphaPlane(DibStream *AlphaDib)
{
	int x,y,w,h,spitch,dpitch;
	BYTE *src,*dst;

	if(AlphaDib == NULL)
		return false;

	w = GetWidth();
	h = GetHeight();

	if(GetBpp()!=32 || AlphaDib->GetBpp()!=8)
		return false;

	if(w != AlphaDib->GetWidth() || h != AlphaDib->GetHeight())
		return false;

	spitch = AlphaDib->GetPitch();
	dpitch = GetPitch();
	src = (BYTE *)AlphaDib->GetBuffer();
	dst = (BYTE *)GetBuffer();
	for(y=0; y<h; y++,src+=spitch,dst+=dpitch) {
		BYTE *sp = src;
		DibBGRA *dp = (DibBGRA *)dst;
		for(x=0; x<w; x++,sp++,dp++)
			dp->a = *sp;
	}

	return true;
}	//DibStream::SetAlphaPlane()

//---------------------------------------------------------------------------
//  DibStream::Create              1999/10/31
//---------------------------------------------------------------------------
bool DibStream::Create(int Width, int Height, int Bpp)
{
	HeaderFree();
	BufferFree();

	Header = CreateHeader(Width, Height, Bpp);
	HeaderMemoryType = HEADER_MEMTYPE;
	if(Header == NULL)
		return false;

	Buffer = CreateBuffer((BITMAPINFOHEADER *)Header);
	BufferMemoryType = BUFFER_MEMTYPE;
	if(Buffer == NULL){
		HeaderFree();
		return false;
	}

	return true;
}	//DibStream::Create()

//---------------------------------------------------------------------------
//  DibStream::CreateHeader        2000/2/8
//---------------------------------------------------------------------------
LPVOID DibStream::CreateHeader(int Width, int Height, int Bpp)
{
	BITMAPINFOHEADER info;
	int i;

	if( !SetBitmapInfo(info, Width, Height, Bpp) )
		return NULL;

	LPVOID NewHeader = GlobalAlloc(GMEM_FIXED,
		sizeof(info)/* + sizeof(DWORD) * 3*/ + info.biClrUsed * sizeof(RGBQUAD));
	*(BITMAPINFOHEADER *)NewHeader = info;

	RGBQUAD *rgb = (RGBQUAD *)((char*)NewHeader + sizeof(info));
	switch(Bpp) {
	case 1:
		rgb[0].rgbBlue = rgb[0].rgbGreen = rgb[0].rgbRed = 0;
		rgb[1].rgbBlue = rgb[1].rgbGreen = rgb[1].rgbRed = 255;
		rgb[0].rgbReserved = rgb[1].rgbReserved = 0;
		break;
	case 8:
		for(i=0; i<256; i++,rgb++) {
			rgb->rgbBlue = rgb->rgbGreen = rgb->rgbRed = (BYTE)i;
			rgb->rgbReserved = 0;
		}
		break;
	/*case 16:
		*(DWORD *)&rgb[0] = 0x00007C00; //r mask
		*(DWORD *)&rgb[1] = 0x000003E0; //g mask
		*(DWORD *)&rgb[2] = 0x0000001F; //b mask
		break;
	case 32:
		*(DWORD *)&rgb[0] = 0x00FF0000; //r mask
		*(DWORD *)&rgb[1] = 0x0000FF00; //g mask
		*(DWORD *)&rgb[2] = 0x000000FF; //b mask
		break;*/
	}

	return NewHeader;
}	//DibStream::CreateHeader()

//---------------------------------------------------------------------------
//  DibStream::CreateBuffer        2000/2/8
//---------------------------------------------------------------------------
LPVOID DibStream::CreateBuffer(BITMAPINFOHEADER *info)
{
	int size = DibWidthBytes(info) * abs(info->biHeight);
	if(size == 0) return NULL;
	return GlobalAlloc(GMEM_FIXED, size);
}	//DibStream::CreateBuffer()

//---------------------------------------------------------------------------
//  DibStream::SetBitmapInfo           2000/2/8
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibStream::SetBitmapInfo(BITMAPINFOHEADER &info, int Width, int Height, int Bpp)
{
	int Pitch;

	info.biSize = sizeof(BITMAPINFOHEADER);
	info.biWidth = Width;
	info.biHeight = Height;
	info.biPlanes = 1;
	info.biBitCount = (WORD)Bpp;
	Pitch = WIDTHBYTES(Width * Bpp);
	switch(Bpp){
	case 1:
		//Pitch = ((Width+31) & (-32)) >> 3;
		info.biCompression = BI_RGB;
		info.biClrUsed = 2;
		break;
	case 4:
		//Pitch = ((Width+7) & (-8)) >> 1;
		info.biCompression = BI_RGB;
		info.biClrUsed = 16;
		break;
	case 8:
		//Pitch = (Width+3) & (-4);
		info.biCompression = BI_RGB;
		info.biClrUsed = 256;
		break;
	case 16:
		//Pitch = (Width*2+2) & (-4);
		info.biCompression = BI_RGB;//BI_BITFIELDS;
		info.biClrUsed = 0;//3;
		break;
	case 24:
		//Pitch = (Width*3+3) & (-4);
		info.biCompression = BI_RGB;
		info.biClrUsed = 0;
		break;
	case 32:
		//Pitch = (Width*4+3) & (-4);
		info.biCompression = BI_RGB;//BI_BITFIELDS;
		info.biClrUsed = 0;//3;
		break;
	default:
		return false;
	}
	info.biSizeImage = Pitch * abs(Height);
	info.biClrImportant = 0;

	info.biXPelsPerMeter = 0;
	info.biYPelsPerMeter = 0;

	return true;
}	//DibStream::SetBitmapInfo()

//---------------------------------------------------------------------------
//  DibStream::GetPitch              1999/12/28
//      corresponded bpp  - all
//---------------------------------------------------------------------------
int DibStream::GetPitch(void)
{
	return (int)WIDTHBYTES(
		(UINT)(BMPHeader->biWidth * (UINT)BMPHeader->biBitCount) );
}

//---------------------------------------------------------------------------
//  DibStream::Pixel
//      corresponded bpp  - all
//---------------------------------------------------------------------------
BYTE *DibStream::Pixel(int x, int y)
{
	BYTE *ptr = ScanLine(y);
	switch(GetBpp()){
	case  1: return &ptr[x>>3];
	case  4: return &ptr[x>>1];
	case  8: return &ptr[x];
	case 16: return &ptr[x*2];
	case 24: return &ptr[x*3];
	case 32: return &ptr[x*4];
	}
	return NULL;
}	//DibStream::Pixel()

//---------------------------------------------------------------------------
//  DibStream::LoadFromFile          2000/2/8
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibStream::LoadFromFile(LPCSTR FileName, DWORD LoadBit)
{
	HANDLE hFile;

	HeaderFree();
	BufferFree();

	hFile = CreateFile(FileName, GENERIC_READ,
		0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, (HANDLE)NULL);
	if(hFile == INVALID_HANDLE_VALUE)
		return false;

	BITMAPFILEHEADER BMPFile;
	DWORD nBytesRead, OrgImgSize;
	BOOL bResult;
	int nNumColors, WidthByte;

	// read BITMAPFILEHEADER
	bResult = ReadFile(hFile, (LPVOID)&BMPFile, sizeof(BMPFile), &nBytesRead, NULL);
	if(!bResult || nBytesRead != sizeof(BMPFile))
		return false;
	// check if this file is old bmp or not bmp
	if(BMPFile.bfType != 0x4D42)	//!="BM"
		goto LoadError;

	// read BITMAPINFOHEADER
	BITMAPINFOHEADER info;
	bResult = ReadFile(hFile, &info, sizeof(info), &nBytesRead, NULL);
	if(!bResult || nBytesRead != sizeof(info))
		goto LoadError;
	if(info.biSize != sizeof(info))
		goto LoadError;

	// check type of compression
	switch(info.biCompression) {
	case BI_RGB: break;
	case BI_RLE8:
		OrgImgSize = info.biSizeImage;
		break;
	default:
		goto LoadError;
	}

	// check biBitCount
	switch(info.biBitCount) {
#define LOAD(b,m) case b: if(!(LoadBit & m))goto LoadError; break;
	LOAD( 1, DIB_LOAD_1);
	LOAD( 4, DIB_LOAD_4);
	LOAD( 8, DIB_LOAD_8);
	LOAD(16, DIB_LOAD_16);
	LOAD(24, DIB_LOAD_24);
	LOAD(32, DIB_LOAD_32);
	default: goto LoadError;
	}

	// fix the number of palette
	if(info.biClrUsed == 0 && info.biBitCount <= 8)
		nNumColors = 1 << (int)info.biBitCount;
	else if(info.biBitCount > 8)
		nNumColors = 0;
	else
		nNumColors = info.biClrUsed;
	info.biClrUsed = nNumColors;
	// fix bitmap-info
	WidthByte = DibWidthBytes(&info);
	info.biSizeImage = WidthByte * abs(info.biHeight);

	// allocate memory
	Header = MemAlloc(
			HeaderMemoryType = HEADER_MEMTYPE,
			sizeof(info) + nNumColors * sizeof(RGBQUAD));
	Buffer = MemAlloc(
			BufferMemoryType = BUFFER_MEMTYPE,
			(int)info.biSizeImage);
	if(Header == NULL || Buffer == NULL)
		goto LoadError;

	*(BITMAPINFOHEADER *)Header = info;

	// read color palettes
	if(nNumColors){
		bResult = ReadFile(hFile, (char*)Header + sizeof(info),
			nNumColors * sizeof(RGBQUAD), &nBytesRead, NULL);
		if(!bResult || nBytesRead != nNumColors * sizeof(RGBQUAD))
			goto LoadError;
	}
	// set file pointer
	if(BMPFile.bfOffBits != 0){
		if(SetFilePointer(hFile, BMPFile.bfOffBits, NULL, FILE_BEGIN) == (DWORD)-1)
			goto LoadError;
	}

	// read bitmaps
	switch(info.biCompression) {
	case BI_RGB:
		bResult = ReadFile(hFile, Buffer,
			info.biSizeImage, &nBytesRead, NULL);
		if(!bResult || nBytesRead != info.biSizeImage)
			goto LoadError;
		break;
	case BI_RLE8:
			{
		int x,y;
		BYTE *TmpBuffer = (BYTE *)malloc(OrgImgSize);
		bResult = ReadFile(hFile, TmpBuffer, OrgImgSize, &nBytesRead, NULL);
		if(!bResult || nBytesRead != OrgImgSize) {
			free(TmpBuffer);
			goto LoadError;
		}
		memset(Buffer, 0, info.biSizeImage);

		BYTE *stream = TmpBuffer;
		BYTE *endstr = &TmpBuffer[OrgImgSize];
		int h = abs(info.biHeight);
		x = y = 0;

		for(; stream<endstr && y<h; )
		{
			BYTE com = *stream++;
			BYTE *ptr = (BYTE *)GetBuffer() + y*WidthByte + x;
			if(com == 0) {
				switch(com = *stream++) {
				case 0: // end of line
					x=0; y++;
					break;
				case 1: // end of image
					goto RLE8_END;
				case 2: // offset
					x += *stream++;
					y += *stream++;
					break;
				default: // raw data
					if(com & 1) {
						x += com;
						for(; com>0; com--)
							*ptr++ = *stream++;
						stream++;
					} else {
						x += com;
						for(; com>0; com--)
							*ptr++ = *stream++;
					}
				}
			} else {
				x += com;
				for(; com>0; com--)
					*ptr++ = *stream;
				stream++;
			}
		}
RLE8_END:;
		free(TmpBuffer);

		info.biCompression = BI_RGB;
		break;
			}
	}

	CloseHandle(hFile);
	return true;

LoadError:
	HeaderFree();
	BufferFree();
	if(hFile)
		CloseHandle(hFile);
	return false;
}	//DibStream::LoadFromFile()

//---------------------------------------------------------------------------
//  DibStream::SaveToFile
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibStream::SaveToFile(LPCSTR FileName)
{
	HANDLE hFile;
	BITMAPFILEHEADER BMPFile;
	BITMAPINFOHEADER info;
	DWORD nBytesWrite;

	hFile = CreateFile(FileName, GENERIC_WRITE,
		0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, (HANDLE)NULL);
	if(hFile == INVALID_HANDLE_VALUE)
		return false;

	// write file header
	BMPFile.bfType = 0x4D42;	//"BM"
	BMPFile.bfReserved1 = BMPFile.bfReserved2 = 0;
	BMPFile.bfOffBits = sizeof(BMPFile)
					  + sizeof(BITMAPINFOHEADER)
					  + GetNumColors() * sizeof(RGBQUAD);
	BMPFile.bfSize = BMPFile.bfOffBits + GetImageSize();
	WriteFile(hFile, &BMPFile, sizeof(BMPFile), &nBytesWrite, NULL);

	// write info header
	info = *(BITMAPINFOHEADER *)Header;
	info.biHeight    = GetHeight();
	info.biSizeImage = GetPitch() * GetHeight();
	WriteFile(hFile, &info, sizeof(info), &nBytesWrite, NULL);

	// write palette
	if(GetNumColors() > 0)
		WriteFile(hFile, GetPalette(), GetNumColors()*sizeof(RGBQUAD),
			&nBytesWrite, NULL);

	// write image
	if(IsBottomUp()) {
		WriteFile(hFile, Buffer, GetImageSize(), &nBytesWrite, NULL);
	} else {
		int pitch = GetPitch();
		for(int y=GetHeight()-1; y>=0; y--)
			WriteFile(hFile, ScanLine(y), pitch, &nBytesWrite, NULL);
	}

	CloseHandle(hFile);

	return true;
}	//DibStream::SaveToFile()

//---------------------------------------------------------------------------
//  DibStream::Create8Bit            1999/10/31
//---------------------------------------------------------------------------
BYTE *DibStream::Create8Bit(void)
{
	int w=GetWidth(), h=GetHeight();
	int Pitch = GetPitch();
	int x,y;

	BYTE tbl[256];
	RGBQUAD *rgb = GetPalette();
	int palc = GetNumColors();

	BYTE *buf = (BYTE *)GlobalAlloc(GMEM_FIXED, w * h);

	BYTE *rp = (BYTE *)Buffer;
	BYTE *wp = (BYTE *)buf;

	switch(GetBpp()){
	case 1:
		for(x=0; x<palc; x++){
			tbl[x] = (BYTE)((
				(DWORD)rgb[x].rgbRed   * BRIGHT_R +
				(DWORD)rgb[x].rgbGreen * BRIGHT_G +
				(DWORD)rgb[x].rgbBlue  * BRIGHT_B) >> 8);
		}
		for(; x<2; x++) tbl[x] = 0;

		for(y=0; y<h; y++,rp+=Pitch){
			BYTE *cr = rp;
			BYTE shift=0x80;
			for(x=0; x<w; x++){
				*wp++ = tbl[(*cr & shift) ? 1 : 0];
				if((shift >>= 1) == 0){
					cr++;
					shift = 0x80;
				}
			}
		}
		break;
	case 4:
		for(x=0; x<palc; x++){
			tbl[x] = (BYTE)((
				(DWORD)rgb[x].rgbRed   * BRIGHT_R +
				(DWORD)rgb[x].rgbGreen * BRIGHT_G +
				(DWORD)rgb[x].rgbBlue  * BRIGHT_B) >> 8);
		}
		for(; x<16; x++) tbl[x] = 0;

		for(y=0; y<h; y++,rp+=Pitch){
			BYTE *cr = rp;
			BYTE shift=0xF0;
			for(x=0; x<w; x++){
				if(shift == 0xF0){
					*wp++ = tbl[(*cr & 0xF0) >> 4];
					shift = 0x0F;
				} else {
					*wp++ = tbl[*cr++ & 0x0F];
					shift = 0xF0;
				}
			}
		}
		break;
	case 8:
		for(x=0; x<palc; x++){
			tbl[x] = (BYTE)((
				(DWORD)rgb[x].rgbRed   * BRIGHT_R +
				(DWORD)rgb[x].rgbGreen * BRIGHT_G +
				(DWORD)rgb[x].rgbBlue  * BRIGHT_B) >> 8);
		}
		for(; x<256; x++) tbl[x] = 0;

		for(y=h; y>0; y--,rp+=Pitch){
			BYTE *cr = rp;
			for(x=w; x>0; x--)
				*wp++ = tbl[*cr++];
		}
		break;
	case 16:
		for(y=0; y<h; y++,rp+=Pitch){
			WORD *cr = (WORD *)rp;
			for(x=0; x<w; x++,cr++){
				DWORD c;
				c  = (((DWORD)*cr & 0x7C00) >> 7) * BRIGHT_R; //r
				c += (((DWORD)*cr & 0x03E0) >> 2) * BRIGHT_G; //g
				c += (((DWORD)*cr & 0x001F) << 3) * BRIGHT_B; //b
				*wp++ = (BYTE)(c >> 8);
			}
		}
		break;
	case 24:
		for(y=h; y>0; y--,rp+=Pitch){
			BYTE *cr = rp;
			for(x=w; x>0; x--){
				DWORD c;
				c  = (DWORD)*cr++ * BRIGHT_B; //b
				c += (DWORD)*cr++ * BRIGHT_G; //g
				c += (DWORD)*cr++ * BRIGHT_R; //r
				*wp++ = (BYTE)(c >> 8);
			}
		}
		break;
	case 32:
		for(y=h; y>0; y--,rp+=Pitch){
			DibBGRA *cr = (DibBGRA *)rp;
			for(x=w; x>0; x--,cr++)
				*wp++ = (BYTE)((
					(DWORD)cr->r*BRIGHT_R + (DWORD)cr->g*BRIGHT_G + (DWORD)cr->b*BRIGHT_B) >> 8);
		}
		break;
	default:
		GlobalFree(buf);
		return NULL;
	}

	return buf;
}	//DibStream::Create8Bit()

//---------------------------------------------------------------------------
//  DibStream::Create24Bit           1999/12/29
//---------------------------------------------------------------------------
DibBGR *DibStream::Create24Bit(void)
{
	int w=GetWidth(), h=GetHeight();
	int Pitch = GetPitch();

	DibBGR *buf = (DibBGR *)GlobalAlloc(GMEM_FIXED, w * h * sizeof(DibBGR));

	switch(GetBpp()) {
	case 1:
			{
		RGBQUAD *rgb = GetPalette();
		BYTE *ss = (BYTE *)Buffer;
		DibBGR *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			BYTE shift=0x80, *s=ss;
			for(int x=0; x<w; x++,d++){
				if(*s & shift)
					*d = DibBGR(rgb[1].rgbBlue, rgb[1].rgbGreen, rgb[1].rgbRed);
				else
					*d = DibBGR(rgb[0].rgbBlue, rgb[0].rgbGreen, rgb[0].rgbRed);
				if((shift>>=1)==0){s++; shift=0x80;}
			}
		}
		break;
			}

	case 4:
			{
		RGBQUAD *rgb = GetPalette();
		BYTE *ss = (BYTE *)Buffer;
		DibBGR *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			BYTE shift=0xF0, *s=ss;
			for(int x=0; x<w; x++,d++){
				BYTE c = *s & shift;
				if(shift == 0xF0) c >>= 4;
				*d = DibBGR(rgb[c].rgbBlue, rgb[c].rgbGreen, rgb[c].rgbRed);
				if(shift == 0xF0) shift = 0x0F;
				else {s++; shift = 0xF0;}
			}
		}
		break;
			}

	case 8:
			{
		RGBQUAD *rgb = GetPalette();
		BYTE *ss = (BYTE *)Buffer;
		DibBGR *d = buf;
		for(int y=h; y>0; y--,ss+=Pitch){
			BYTE *s = ss;
			for(int x=w; x>0; x--){
				BYTE c = *s++;
				*d++ = DibBGR(rgb[c].rgbBlue, rgb[c].rgbGreen, rgb[c].rgbRed);
			}
		}
		break;
			}

	case 16:
			{
		BYTE *ss = (BYTE *)Buffer;
		DibBGR *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			WORD *s = (WORD *)ss;
			for(int x=0; x<w; x++,s++){
				DWORD c;
				c  = ((DWORD)*s & 0x7C00) << 9; //r
				c |= ((DWORD)*s & 0x03E0) << 6; //g
				c |= ((DWORD)*s & 0x001F) << 3; //b
				*d++ = DibBGR(c);
			}
		}
		break;
			}

	case 24:
			{
		BYTE *ss = (BYTE *)Buffer;
		DibBGR *d = buf;
		for(int y=h; y>0; y--,ss+=Pitch,d+=w)
			memcpy(d, ss, w*sizeof(DibBGR));
		break;
			}

	case 32:
			{
		BYTE *ss = (BYTE *)Buffer;
		DibBGR *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			DWORD *s = (DWORD *)ss;
			for(int x=0; x<w; x++,s++)
				*d++ = DibBGR(*s);
		}
		break;
			}

	default:
		GlobalFree(buf);
		return NULL;
	}

	return buf;
}	//DibStream::Create24Bit()

//---------------------------------------------------------------------------
//  DibStream::Create32Bit           1999/10/31
//---------------------------------------------------------------------------
DibBGRA *DibStream::Create32Bit(void)
{
	int w=GetWidth(), h=GetHeight();
	int Pitch = GetPitch();

	DWORD *buf = (DWORD *)GlobalAlloc(GMEM_FIXED, w * h * sizeof(DWORD));

	switch(GetBpp()) {
	case 1:
			{
		RGBQUAD *rgb = GetPalette();
		BYTE *ss = (BYTE *)Buffer;
		DWORD *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			BYTE shift=0x80, *s=ss;
			for(int x=0; x<w; x++,d++){
				if(*s & shift)
					*d = (rgb[1].rgbRed<<16) | (rgb[1].rgbGreen<<8) |
						rgb[1].rgbBlue | 0xFF000000;
				else
					*d = (rgb[0].rgbRed<<16) | (rgb[0].rgbGreen<<8) |
						rgb[0].rgbBlue | 0xFF000000;
				if((shift>>=1)==0){s++; shift=0x80;}
			}
		}
		break;
			}
	case 4:
			{
		RGBQUAD *rgb = GetPalette();
		BYTE *ss = (BYTE *)Buffer;
		DWORD *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			BYTE shift=0xF0, *s=ss;
			for(int x=0; x<w; x++,d++){
				BYTE c = *s & shift;
				if(shift == 0xF0) c >>= 4;
				*d = (rgb[c].rgbRed<<16) | (rgb[c].rgbGreen<<8) |
					rgb[c].rgbBlue | 0xFF000000;
				if(shift == 0xF0) shift = 0x0F;
				else {s++; shift = 0xF0;}
			}
		}
		break;
			}
	case 8:
			{
		RGBQUAD *rgb = GetPalette();
		BYTE *ss = (BYTE *)Buffer;
		DWORD *d = buf;
		for(int y=h; y>0; y--,ss+=Pitch){
			BYTE *s = ss;
			for(int x=w; x>0; x--){
				BYTE c = *s++;
				*d++ = (rgb[c].rgbRed<<16) | (rgb[c].rgbGreen<<8) |
					rgb[c].rgbBlue | 0xFF000000;
			}
		}
		break;
			}
	case 16:
			{
		BYTE *ss = (BYTE *)Buffer;
		DWORD *d = buf;
		for(int y=0; y<h; y++,ss+=Pitch){
			WORD *s = (WORD *)ss;
			for(int x=0; x<w; x++,s++){
				DWORD c;
				c  = ((DWORD)*s & 0x7C00) << 9; //r
				c |= ((DWORD)*s & 0x03E0) << 6; //g
				c |= ((DWORD)*s & 0x001F) << 3; //b
				*d++ = c | 0xFF000000; //a
			}
		}
		break;
			}
	case 24:
			{
		BYTE *ss = (BYTE *)Buffer;
		DWORD *d = buf;
		for(int y=h; y>0; y--,ss+=Pitch){
			BYTE *s = ss;
			for(int x=w; x>0; x--){
				DWORD c;
				c  = (DWORD)*s++; //b
				c |= (DWORD)*s++ << 8; //g
				c |= (DWORD)*s++ << 16; //r
				*d++ = c | 0xFF000000; //a
			}
		}
		break;
			}
	case 32:
		memcpy(buf, Buffer, w*h*sizeof(DWORD));
		break;
	default:
		GlobalFree(buf);
		return NULL;
	}
	return (DibBGRA *)buf;
}	//DibStream::Create32Bit()

//---------------------------------------------------------------------------
//  DibStream::ChangeTo1Bit
//---------------------------------------------------------------------------
bool DibStream::ChangeTo1Bit(DWORD Flag)
{
	DibBGRA *buf = Create32Bit();
	if(buf == NULL)
		return false;

	int w=GetWidth(), h=GetHeight();
	Create(w, GetHeader()->biHeight, 1);
	if(h < 0) h = -h;

	int Pitch = GetPitch();

	memset(Buffer, 0, Pitch * h);

	int *errbuf = (int *)malloc(sizeof(int) * w * 3);
	memset(errbuf, 0, sizeof(int) * w * 3);

	int x,y;

#define E1 8
#define E2 4
#define E3 2
#define E4 1
#define SUM  (E1*2+E2*4+E3*4+E4*2)

	DibBGRA *rs;
	BYTE *ws = (BYTE *)Buffer;
	for(y=0,rs=buf; y<h; y++,rs+=w,ws+=Pitch)
	{
		DibBGRA *rp = rs;
		BYTE *wp = ws;
		BYTE shift=0x80;
		int *cerr = errbuf;
		for(x=0; x<w; x++,rp++,cerr++)
		{
			int errval = *cerr / SUM;
			int errmod = *cerr - errval*SUM;
			errval += (int)((
				(DWORD)rp->r*BRIGHT_R + (DWORD)rp->g*BRIGHT_G + (DWORD)rp->b*BRIGHT_B) >> 8);
			if(errval >= 128){
				*wp |= shift;
				errval = errval - 255;
			} else {
				errval = errval;
			}
			int errnum = errval * SUM + errmod;
			int cev,*diferr;

#define SETERR(ex,ev) cev=errval*ev; diferr[ex]+=cev; errnum-=cev;
			diferr = cerr;
			if(x+1 < w){ SETERR(1,E1) }
			if(x+2 < w){ SETERR(2,E2) }
			if(y+1 < h){
				diferr += w;
				if(x-2 >=0){ SETERR(-2,E3) }
				if(x-1 >=0){ SETERR(-1,E2) }
				SETERR(0,E1)
				if(x+1 < w){ SETERR(+1,E2) }
				if(x+2 < w){ SETERR(+2,E3) }
			}
			if(y+2 < h){
				diferr += w;
				if(x-2 >=0){ SETERR(-2,E4) }
				if(x-1 >=0){ SETERR(-1,E3) }
							 SETERR( 0,E2)
				if(x+1 < w){ SETERR(+1,E3) }
				if(x+2 < w){ diferr[2]+=errnum; }
			}
#undef SETERR

			if((shift >>= 1) == 0){
				wp++;
				shift=0x80;
			}
		}

		memcpy(errbuf, &errbuf[w], sizeof(int) * w);
		memcpy(&errbuf[w], &errbuf[w*2], sizeof(int) * w);
		memset(&errbuf[w*2], 0, sizeof(int) * w);
	}

	free(errbuf);

	GlobalFree(buf);
	return true;
}	//DibStream::ChangeTo1Bit()

//---------------------------------------------------------------------------
//  DibStream::ChangeTo4Bit          1999/12/30
//---------------------------------------------------------------------------
bool DibStream::ChangeTo4Bit(DWORD Flag)
{
	switch(Flag & DIB_CHANGEBPP_MASK) {
	case DIB_CHANGEBPP_4_OPTIMIZE:
		return MedianCut4(Flag);

	case DIB_CHANGEBPP_4_MONO:
	default:
			{
		BYTE *buf = Create8Bit();
		if(buf == NULL)
			return false;

		Create(GetWidth(), GetHeader()->biHeight, 4);

		// set monochrome palette
		SetMonoPalette();

		// set image
		int width  = GetWidth();
		int height = GetHeight();
		int pitch  = GetPitch();
		int count  = width / 2;
		BYTE *rp = buf;
		BYTE *wp = (BYTE *)Buffer;
		for(int y=height; y>0; y--,rp+=width,wp+=pitch)
		{
			BYTE *r = rp;
			BYTE *w = wp;
			for(int x=count; x>0; x--,r+=2,w++)
				*w = (BYTE)(((r[1] & 0xF0) >> 4) | (r[0] & 0xF0));
			if(width & 1)
				*w = (BYTE)(*r & 0xF0);
		}

		GlobalFree(buf);
		break;
			}
	}

	return true;
}	//DibStream::ChangeTo4Bit()

//---------------------------------------------------------------------------
//  DibStream::ChangeTo8Bit          1999/12/30
//---------------------------------------------------------------------------
bool DibStream::ChangeTo8Bit(DWORD Flag)
{
	switch(Flag & DIB_CHANGEBPP_MASK) {
	case DIB_CHANGEBPP_8_DITHER:
			{
		DibBGR *buf = Create24Bit();
		if(buf == NULL)
			return false;

		Create(GetWidth(), GetHeader()->biHeight, 8);

		// set palette
		RGBQUAD *rgb = GetPalette();
		for(int b=0; b<=255; b+=51) {
			for(int g=0; g<=255; g+=51) {
				for(int r=0; r<=255; r+=51,rgb++) {
					rgb->rgbBlue  = (BYTE)b;
					rgb->rgbGreen = (BYTE)g;
					rgb->rgbRed   = (BYTE)r;
					rgb->rgbReserved = 0;
				}
			}
		}

		// set image
		int width  = GetWidth();
		int height = GetHeight();
		int pitch  = GetPitch();
		DibBGR *ptr = buf;
		BYTE *ps = (BYTE *)Buffer;

		for(int y=0; y<height; y++,ps+=pitch)
		{
			BYTE *p = ps;
			for(int x=0; x<width; x++,ptr++)
			{
				BYTE border = aHalftone8x8[((x&7)<<3) + (y&7)];
				BYTE tr = (BYTE)(aDividedBy51[ptr->r] + (aModulo51[ptr->r] > border));
				BYTE tg = (BYTE)(aDividedBy51[ptr->g] + (aModulo51[ptr->g] > border));
				BYTE tb = (BYTE)(aDividedBy51[ptr->b] + (aModulo51[ptr->b] > border));
				*p++ = (BYTE)(tr + aTimes6[tg] + aTimes36[tb]);
			}
		}

		GlobalFree(buf);
		break;
			}

	case DIB_CHANGEBPP_8_OPTIMIZE:
		return MedianCut8(Flag);

	case DIB_CHANGEBPP_8_MONO:
	default:
			{
		BYTE *buf = Create8Bit();
		if(buf == NULL)
			return false;

		Create(GetWidth(), GetHeader()->biHeight, 8);

		// set monochrome palette
		SetMonoPalette();

		// set image
		int w=GetWidth(), h=GetHeight();
		int Pitch = GetPitch();
		BYTE *rp = buf;
		BYTE *wp = (BYTE *)Buffer;
		for(int y=h; y>0; y--,rp+=w,wp+=Pitch)
			memcpy(wp, rp, w);

		GlobalFree(buf);
		break;
			}
	}

	return true;
}	//DibStream::ChangeTo8Bit()

//---------------------------------------------------------------------------
//  namespace dibmedian            1999/12/30
//---------------------------------------------------------------------------
namespace dibmedian {

class coltree
{
public:
	int type; //0.none 1.red 2.green 3.blue
	int pos; // (less) - (equal or more)
	coltree *left, *right;
	int pal_index;

	coltree() { type=0; left=right=NULL; }
	~coltree() {
		if(left)  delete left;
		if(right) delete right;
	}
};

class colinfo
{
public:
	int r_pos,  g_pos,  b_pos;
	int r_size, g_size, b_size;
	int pixel_count;
	coltree *tree; // link

	colinfo() { pixel_count = 0; }
	colinfo(int p1, int p2, int p3, int s1, int s2, int s3) {
		r_pos = p1;
		g_pos = p2;
		b_pos = p3;
		r_size = s1;
		g_size = s2;
		b_size = s3;
		pixel_count = 0;
	}

	inline int get_r_min(void) { return r_pos; }
	inline int get_g_min(void) { return g_pos; }
	inline int get_b_min(void) { return b_pos; }
	inline int get_r_max(void) { return r_pos + r_size - 1; }
	inline int get_g_max(void) { return g_pos + g_size - 1; }
	inline int get_b_max(void) { return b_pos + b_size - 1; }
	inline void set_r_min(int val) { r_pos = val; }
	inline void set_g_min(int val) { g_pos = val; }
	inline void set_b_min(int val) { b_pos = val; }
	inline void set_r_max(int val) { r_size = val - r_pos + 1; }
	inline void set_g_max(int val) { g_size = val - g_pos + 1; }
	inline void set_b_max(int val) { b_size = val - b_pos + 1; }

	inline int maxlen(void) {
		return max(r_size, max(g_size, b_size));
	}

	inline int get_pixel_weit(void) {
		int weit = 0;
		for(int i=pixel_count; i>0; i>>=1,weit++);
		return weit;
	}

	int maxtype(void) {
		int len = maxlen();
		if(r_size == len) return 1;
		if(g_size == len) return 2;
		if(b_size == len) return 3;
		return 0;
	}
};

class medianhistdata
{
private:
	int size;
public:
	DibBGR *col;
	int count;

	medianhistdata() {
		col = NULL;
		count = size = 0;
	}
	~medianhistdata() {
		if(col)
			free(col);
	}
	void add(DibBGR newcol) {
		if(count >= size) {
			size = size * 2 + 128;
			col = (DibBGR *)realloc(col, size * sizeof(DibBGR));
		}
		col[count++] = newcol;
	}
};

#define MEDIANHIST_SHIFT   4
#define MEDIANHIST_SIZE    16

class medianhist
{
public:
	medianhistdata hist[MEDIANHIST_SIZE][MEDIANHIST_SIZE][MEDIANHIST_SIZE];

	medianhist() {
	}
	inline medianhistdata& get(int r, int g, int b) { return hist[r][g][b]; }
	inline void add(DibBGR col) {
		get(col.r>>MEDIANHIST_SHIFT, col.g>>MEDIANHIST_SHIFT, col.b>>MEDIANHIST_SHIFT).add(col);
	}
	int search(
		int rmin, int rmax, int gmin, int gmax, int bmin, int bmax,
		int *count, int offset)
	{
		int r1 = rmin >> MEDIANHIST_SHIFT;
		int r2 = rmax >> MEDIANHIST_SHIFT;
		int g1 = gmin >> MEDIANHIST_SHIFT;
		int g2 = gmax >> MEDIANHIST_SHIFT;
		int b1 = bmin >> MEDIANHIST_SHIFT;
		int b2 = bmax >> MEDIANHIST_SHIFT;

		int total = 0;

		for(int r=r1; r<=r2; r++) {
			for(int g=g1; g<=g2; g++) {
				for(int b=b1; b<=b2; b++) {
					medianhistdata& data = get(r, g, b);
					DibBGR *ptr = data.col;
					for(int i=data.count; i>0; i--,ptr++) {
						if(rmin <= ptr->r && ptr->r <= rmax
						&& gmin <= ptr->g && ptr->g <= gmax
						&& bmin <= ptr->b && ptr->b <= bmax)
						{
							count[ptr->col[offset]]++;
							total++;
						}
					}
				}
			}
		}

		return total;
	}

	void rect(
		int rbox_min, int rbox_max, int gbox_min, int gbox_max, int bbox_min, int bbox_max,
		int& rmin, int& rmax, int& gmin, int& gmax, int& bmin, int& bmax)
	{
		int r1 = rbox_min >> MEDIANHIST_SHIFT;
		int r2 = rbox_max >> MEDIANHIST_SHIFT;
		int g1 = gbox_min >> MEDIANHIST_SHIFT;
		int g2 = gbox_max >> MEDIANHIST_SHIFT;
		int b1 = bbox_min >> MEDIANHIST_SHIFT;
		int b2 = bbox_max >> MEDIANHIST_SHIFT;

		for(int r=r1; r<=r2; r++) {
			for(int g=g1; g<=g2; g++) {
				for(int b=b1; b<=b2; b++) {
					medianhistdata& data = get(r, g, b);
					DibBGR *ptr = data.col;
					for(int i=data.count; i>0; i--,ptr++) {
						if(rbox_min <= ptr->r && ptr->r <= rbox_max
						&& gbox_min <= ptr->g && ptr->g <= gbox_max
						&& bbox_min <= ptr->b && ptr->b <= bbox_max)
						{
							if(rmin > ptr->r) rmin = ptr->r;
							if(rmax < ptr->r) rmax = ptr->r;
							if(gmin > ptr->g) gmin = ptr->g;
							if(gmax < ptr->g) gmax = ptr->g;
							if(bmin > ptr->b) bmin = ptr->b;
							if(bmax < ptr->b) bmax = ptr->b;
						}
					}
				}
			}
		}
	}
};

class mediancut
{
public:
	coltree *tree;
	colinfo info[256+4];
	int count;

	mediancut() {
		tree = NULL;
		count = 0;
	}

	~mediancut() {
		if(tree)
			delete tree;
	}

	int add(colinfo& newinfo) {
		info[count] = newinfo;
		info[count].tree = new coltree;
		if(tree == NULL)
			tree = info[count].tree;
		return count++;
	}

	int remove(int index) {
		if(count > 0)
			info[index] = info[--count];
		return count;
	}

	void slice(int index, int type, int pos, int& i1, int& i2)
	{
		colinfo col = info[index];
		coltree *parent = col.tree;

		// remove old box
		remove(index);

		// add new two boxes
		switch(type) {
		case 1: //red
			i1 = add(colinfo(
				col.r_pos, col.g_pos, col.b_pos,
				pos-col.r_pos, col.g_size, col.b_size));
			i2 = add(colinfo(
				pos, col.g_pos, col.b_pos,
				col.r_size+col.r_pos-pos, col.g_size, col.b_size));
			break;
		case 2: // green
			i1 = add(colinfo(
				col.r_pos, col.g_pos, col.b_pos,
				col.r_size, pos-col.g_pos, col.b_size));
			i2 = add(colinfo(
				col.r_pos, pos, col.b_pos,
				col.r_size, col.g_size+col.g_pos-pos, col.b_size));
			break;
		case 3: // blue
			i1 = add(colinfo(
				col.r_pos, col.g_pos, col.b_pos,
				col.r_size, col.g_size, pos-col.b_pos));
			i2 = add(colinfo(
				col.r_pos, col.g_pos, pos,
				col.r_size, col.g_size, col.b_size+col.b_pos-pos));
			break;
		}
		parent->type = type;
		parent->pos  = pos;
		parent->left  = info[i1].tree;
		parent->right = info[i2].tree;
	}

	int get_index(void)
	{
		int weit = 0;
		int index = -1;
		// search maximum weit of 'colinfo'
		for(int i=0; i<count; i++) {
			int infoweit = info[i].maxlen() * info[i].get_pixel_weit();
			if(weit < infoweit) {
				weit = infoweit;
				index = i;
			}
		}
		return index;
	}

	bool execute(DibBGR *buf, int width, int height, int palette_max);
};

bool mediancut::execute(DibBGR *buf, int width, int height, int palette_max)
{
	int i;
	DibBGR *ptr;

	dibmedian::medianhist *hists = new dibmedian::medianhist;
	dibmedian::medianhist& histgram = *hists;

	int rmin, rmax, gmin, gmax, bmin, bmax;

	// set first box
	ptr = buf;
	rmin = rmax = ptr->r;
	gmin = gmax = ptr->g;
	bmin = bmax = ptr->b;
	for(int y=0; y<height; y++) {
		for(int x=0; x<width; x++,ptr++) {
			histgram.add(*ptr);
			if(rmin > ptr->r) rmin = ptr->r;
			if(rmax < ptr->r) rmax = ptr->r;
			if(gmin > ptr->g) gmin = ptr->g;
			if(gmax < ptr->g) gmax = ptr->g;
			if(bmin > ptr->b) bmin = ptr->b;
			if(bmax < ptr->b) bmax = ptr->b;
		}
	}
	i = add(dibmedian::colinfo(
		rmin, gmin, bmin, rmax-rmin+1, gmax-gmin+1, bmax-bmin+1));
	info[i].pixel_count = width * height;

	// median-cut
	while(count < palette_max)
	{
		int hist[256];
		int index = get_index();
		dibmedian::colinfo& col = info[index];
		int len = col.maxlen();
		if(len <= 1)
			break; // cannot slice more

		int total, counter, pos;
		int type = col.maxtype();

		rmin = col.get_r_min();
		rmax = col.get_r_max();
		gmin = col.get_g_min();
		gmax = col.get_g_max();
		bmin = col.get_b_min();
		bmax = col.get_b_max();

		switch(type) {
		case 1: // red
			for(i=rmin; i<=rmax; i++)
				hist[i] = 0;
			// get a histgram of color
			total = histgram.search(rmin, rmax, gmin, gmax, bmin, bmax, hist, 2);
			// search middle point
			for(pos=rmin+1,counter=hist[rmin];
				pos<rmax && 2*counter<total;
				counter+=hist[pos++]);
			break;

		case 2: // green
			for(i=gmin; i<=gmax; i++)
				hist[i] = 0;
			// get a histgram of color
			total = histgram.search(rmin, rmax, gmin, gmax, bmin, bmax, hist, 1);
			// search middle point
			for(pos=gmin+1,counter=hist[gmin];
				pos<gmax && 2*counter<total;
				counter+=hist[pos++]);
			break;

		case 3: // blue
			for(i=bmin; i<=bmax; i++)
				hist[i] = 0;
			// get a histgram of color
			total = histgram.search(rmin, rmax, gmin, gmax, bmin, bmax, hist, 0);
			// search middle point
			for(pos=bmin+1,counter=hist[bmin];
				pos<bmax && 2*counter<total;
				counter+=hist[pos++]);
			break;
		}

		if(total == 0) {
			remove(index);
			continue;
		}

		// slice
		int box[2];
		slice(index, type, pos, box[0], box[1]);
		info[box[0]].pixel_count = counter;
		info[box[1]].pixel_count = total - counter;

#if 0
		char mes[64];
		static const char *coltext[] = {"", "  red", "green", "blue"};
		sprintf(mes, "%s: %d %d:%d", coltext[type], pos, counter, total);
		OutputDebugString(mes);
#endif

		// resize bounding boxes
		for(i=0; i<2; i++)
		{
			dibmedian::colinfo& col = info[box[i]];
			int rbox_min, rbox_max;
			int gbox_min, gbox_max;
			int bbox_min, bbox_max;
			rmin = rbox_max = col.get_r_max();
			rmax = rbox_min = col.get_r_min();
			gmin = gbox_max = col.get_g_max();
			gmax = gbox_min = col.get_g_min();
			bmin = bbox_max = col.get_b_max();
			bmax = bbox_min = col.get_b_min();

			histgram.rect(
				rbox_min, rbox_max, gbox_min, gbox_max, bbox_min, bbox_max,
				rmin, rmax, gmin, gmax, bmin, bmax);

			col.set_r_min(rmin);
			col.set_r_max(rmax);
			col.set_g_min(gmin);
			col.set_g_max(gmax);
			col.set_b_min(bmin);
			col.set_b_max(bmax);
		}
	}

	delete hists;

	// make indices of palettes
	for(i=0; i<count; i++)
		info[i].tree->pal_index = i;

	return true;
}

}; //namespace dibmedian

//---------------------------------------------------------------------------
//  DibStream::MedianCut4            1999/12/30
//---------------------------------------------------------------------------
bool DibStream::MedianCut4(DWORD Flag)
{
	int i;

	DibBGR *buf = Create24Bit();
	if(buf == NULL)
		return false;

	int width  = GetWidth();
	int height = GetHeight();

	// median cut
	dibmedian::mediancut median;
	if( !median.execute(buf, width, height,
		(Flag & DIB_CHANGEBPP_OPTIMIZE_MASK) ?
			min((int)(Flag & DIB_CHANGEBPP_OPTIMIZE_MASK),16) : 16) )
	{
		GlobalFree(buf);
		return false;
	}

	Create(GetWidth(), GetHeader()->biHeight, 4);

	// set palette
	RGBQUAD *rgb = GetPalette();
	for(i=0; i<median.count; i++,rgb++)
	{
		dibmedian::colinfo& info = median.info[i];
		rgb->rgbBlue  = (BYTE)(info.b_pos + info.b_size/2);
		rgb->rgbGreen = (BYTE)(info.g_pos + info.g_size/2);
		rgb->rgbRed   = (BYTE)(info.r_pos + info.r_size/2);
		rgb->rgbReserved = 0;
	}
	for(; i<16; i++,rgb++)
		rgb->rgbBlue = rgb->rgbGreen = rgb->rgbRed = rgb->rgbReserved = 0;

	// set image
	int count = width / 2;
	int pitch = GetPitch();
	BYTE *wp = (BYTE *)Buffer;
	DibBGR *ptr = buf;
	for(int y=height; y>0; y--,wp+=pitch)
	{
		BYTE *p = wp;
		for(int x=count; x>0; x--)
		{
			// search a nearest color in the binary tree
			BYTE pal_index[2];

			for(int i=0; i<2; i++,ptr++)
			{
				dibmedian::coltree* tree = median.tree;
				for(; tree->type!=0; ) {
					switch(tree->type) {
					case 1:
						if(ptr->r < tree->pos) tree = tree->left;
						else                   tree = tree->right;
						break;
					case 2:
						if(ptr->g < tree->pos) tree = tree->left;
						else                   tree = tree->right;
						break;
					case 3:
						if(ptr->b < tree->pos) tree = tree->left;
						else                   tree = tree->right;
						break;
					}
				}
				pal_index[i] = (BYTE)tree->pal_index;
			}
			// set index of palette
			*p++ = (BYTE)(
				((pal_index[0] << 4) & 0xF0) |
				(pal_index[1] & 0x0F) );
		}

		if(width & 1)
		{
			dibmedian::coltree* tree = median.tree;
			for(; tree->type!=0; ) {
				switch(tree->type) {
				case 1:
					if(ptr->r < tree->pos) tree = tree->left;
					else                   tree = tree->right;
					break;
				case 2:
					if(ptr->g < tree->pos) tree = tree->left;
					else                   tree = tree->right;
					break;
				case 3:
					if(ptr->b < tree->pos) tree = tree->left;
					else                   tree = tree->right;
					break;
				}
			}
			// set index of palette
			*p = (BYTE)((tree->pal_index << 4) & 0xF0);
		}
	}

	GlobalFree(buf);

	return true;
}

//---------------------------------------------------------------------------
//  DibStream::MedianCut8            1999/12/30
//---------------------------------------------------------------------------
bool DibStream::MedianCut8(DWORD Flag)
{
	int i;

	DibBGR *buf = Create24Bit();
	if(buf == NULL)
		return false;

	int width  = GetWidth();
	int height = GetHeight();

	// median cut
	dibmedian::mediancut median;
	if( !median.execute(buf, width, height,
		(Flag & DIB_CHANGEBPP_OPTIMIZE_MASK) ?
			(Flag & DIB_CHANGEBPP_OPTIMIZE_MASK) : 256) )
	{
		GlobalFree(buf);
		return false;
	}

	Create(GetWidth(), GetHeader()->biHeight, 8);

	// set palette
	RGBQUAD *rgb = GetPalette();
	for(i=0; i<median.count; i++,rgb++)
	{
		dibmedian::colinfo& info = median.info[i];
		rgb->rgbBlue  = (BYTE)(info.b_pos + info.b_size/2);
		rgb->rgbGreen = (BYTE)(info.g_pos + info.g_size/2);
		rgb->rgbRed   = (BYTE)(info.r_pos + info.r_size/2);
		rgb->rgbReserved = 0;
	}
	for(; i<256; i++,rgb++)
		rgb->rgbBlue = rgb->rgbGreen = rgb->rgbRed = rgb->rgbReserved = 0;

	// set image
	int pitch = GetPitch();
	BYTE *wp = (BYTE *)Buffer;
	DibBGR *ptr = buf;
	for(int y=height; y>0; y--,wp+=pitch)
	{
		BYTE *p = wp;
		for(int x=width; x>0; x--,ptr++)
		{
			// search a nearest color in the binary tree
			dibmedian::coltree* tree = median.tree;
			for(; tree->type!=0; ) {
				switch(tree->type) {
				case 1:
					if(ptr->r < tree->pos) tree = tree->left;
					else                   tree = tree->right;
					break;
				case 2:
					if(ptr->g < tree->pos) tree = tree->left;
					else                   tree = tree->right;
					break;
				case 3:
					if(ptr->b < tree->pos) tree = tree->left;
					else                   tree = tree->right;
					break;
				}
			}
			// set index of palette
			*p++ = (BYTE)tree->pal_index;
		}
	}

	GlobalFree(buf);

	return true;
}

//---------------------------------------------------------------------------
//  DibStream::ChangeTo16Bit         1999/12/30
//---------------------------------------------------------------------------
bool DibStream::ChangeTo16Bit(DWORD Flag)
{
	DibBGRA *buf = Create32Bit();
	if(buf == NULL)
		return false;

	int w=GetWidth(), h=GetHeight();
	Create(w, GetHeader()->biHeight, 16);
	if(h < 0) h = -h;

	int Pitch = GetPitch();
	DibBGRA *bs = buf;
	BYTE *ps = (BYTE *)Buffer;
	switch(Flag & DIB_CHANGEBPP_MASK) {
	case DIB_CHANGEBPP_16_DITHER:
			{
		DWORD mb=0,mg=0,mr=0;
		for(int y=h; y>0; y--,ps+=Pitch) {
			WORD *p = (WORD *)ps;
			for(int x=w; x>0; x--,bs++) {
				mb+=bs->b; if(mb>0xFF)mb=0xFF;
				mg+=bs->g; if(mg>0xFF)mg=0xFF;
				mr+=bs->r; if(mr>0xFF)mr=0xFF;
				*p++ = (WORD)(((mb>>3)&0x1F) | ((mg<<2)&0x3E0) | ((mr<<7)&0x7C00));
				mb&=0x7; mg&=0x7; mr&=0x7;
			}
		}
		break;
			}
	default:
			{
		for(int y=0; y<h; y++,bs=&bs[w],ps+=Pitch) {
			DWORD *b = (DWORD *)bs;
			WORD *p = (WORD *)ps;
			for(int x=w; x>0; x--,b++)
				*p++ = (WORD)(((*b>>3)&0x1F) | ((*b>>6)&0x3E0) | ((*b>>9)&0x7C00));
		}
		break;
			}
	}

	GlobalFree(buf);
	return true;
}	//DibStream::ChangeTo16Bit()

//---------------------------------------------------------------------------
//  DibStream::ChangeTo24Bit         1999/12/29
//---------------------------------------------------------------------------
bool DibStream::ChangeTo24Bit(DWORD Flag)
{
	DibBGR *buf = Create24Bit();
	if(buf == NULL)
		return false;

	int w=GetWidth(), h=GetHeight();
	Create(w, GetHeader()->biHeight, 24);
	if(h < 0) h = -h;

	int Pitch = GetPitch();
	DibBGR *bs = buf;
	BYTE *ps = (BYTE *)Buffer;
	for(int y=h; y>0; y--,ps+=Pitch,bs+=w)
		memcpy(ps, bs, sizeof(DibBGR) * w);

	GlobalFree(buf);
	return true;
}	//DibStream::ChangeTo24Bit()

//---------------------------------------------------------------------------
//  DibStream::ChangeTo32Bit
//---------------------------------------------------------------------------
bool DibStream::ChangeTo32Bit(DWORD Flag)
{
	DibBGRA *buf = Create32Bit();
	if(buf == NULL)
		return false;

	void *NewHeader = CreateHeader(GetWidth(), GetHeader()->biHeight, 32);
	if(NewHeader == NULL)
		return false;

	SetDib(NewHeader, HEADER_MEMTYPE, buf, BUFFER_MEMTYPE);

	return true;
}	//DibStream::ChangeTo32Bit()

//---------------------------------------------------------------------------
//  DibStream::ChangeBpp
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibStream::ChangeBpp(int Bpp, DWORD Flag)
{
	if(Buffer == NULL) return false;

	switch(Bpp){
	case 1:  return ChangeTo1Bit(Flag);
	case 4:  return ChangeTo4Bit(Flag);
	case 8:  return ChangeTo8Bit(Flag);
	case 16: return ChangeTo16Bit(Flag);
	case 24: return ChangeTo24Bit(Flag);
	case 32: return ChangeTo32Bit(Flag);
	}
	return false;
}	//DibStream::ChangeBpp()

//---------------------------------------------------------------------------
//  DibStream::SetMonoPalette      1999/12/29
//      corresponded bpp  - 1,4,8
//---------------------------------------------------------------------------
void DibStream::SetMonoPalette(void)
{
	RGBQUAD *rgb;
	unsigned count;

	switch(GetBpp()) {
	case 1:
		rgb = GetPalette();
		for(count=0; count<2; rgb++,count++) {
			rgb->rgbRed = rgb->rgbGreen = rgb->rgbBlue = (BYTE)(count*255);
			rgb->rgbReserved = 0;
		}
		break;

	case 4:
		rgb = GetPalette();
		for(count=0; count<16; rgb++,count++) {
			rgb->rgbRed = rgb->rgbGreen = rgb->rgbBlue = (BYTE)(count*17);
			rgb->rgbReserved = 0;
		}
		break;

	case 8:
		rgb = GetPalette();
		for(count=0; count<256; rgb++,count++) {
			rgb->rgbRed = rgb->rgbGreen = rgb->rgbBlue = (BYTE)count;
			rgb->rgbReserved = 0;
		}
		break;
	}
}

//---------------------------------------------------------------------------
//  DibStream::UpDown
//      corresponded bpp  - all
//---------------------------------------------------------------------------
void DibStream::UpDown(bool TopDown)
{
	if(Buffer == NULL) return;

	if(!TopDown && IsBottomUp()) return;
	if( TopDown && IsTopDown())  return;

	int h = GetHeight();
	int halfh = h / 2;
	int pitch = GetPitch();
	LPBYTE top = (LPBYTE)GetBuffer();
	LPBYTE bottom = &top[pitch*(h-1)];

	void *buf = malloc(pitch);

	if(top == NULL || buf == NULL) return;

	// reverse top and bottom
	for(int y=0; y<halfh; y++,top+=pitch,bottom-=pitch)
	{
#if 0
		for(int x=0; x<pitch; x++){
			BYTE c = top[x];
			top[x] = bottom[x];
			bottom[x] = c;
		}
#else
		memcpy(buf, top, pitch);
		memcpy(top, bottom, pitch);
		memcpy(bottom, buf, pitch);
#endif
	}

	GetHeader()->biHeight = -GetHeader()->biHeight;

	free(buf);
}	//DibStream::UpDown()

//---------------------------------------------------------------------------
//  DibStream::CreatePalette       1999/12/30
//      corresponded bpp  - 1,4,8
//---------------------------------------------------------------------------
HPALETTE DibStream::CreatePalette(void)
{
	int i;
	struct {
		WORD         palVersion;
		WORD         palNumEntries;
		PALETTEENTRY palPalEntry[256];
	} LogPalette;

	if(Header == NULL)
		return NULL;

	// set version
	LogPalette.palVersion = 0x300;

	// set number
	switch(GetBpp()) {
	case 1:
	case 4:
	case 8:
		LogPalette.palNumEntries = (WORD)GetNumColors();
		break;
	default:
		return NULL;
	}

	// set palette
	RGBQUAD *rgb = GetPalette();
	PALETTEENTRY *pal = LogPalette.palPalEntry;
	for(i=LogPalette.palNumEntries; i>0; i--,pal++,rgb++) {
		pal->peRed = rgb->rgbRed;
		pal->peGreen = rgb->rgbGreen;
		pal->peBlue = rgb->rgbBlue;
		pal->peFlags = PC_NOCOLLAPSE;
	}

	return ::CreatePalette((LOGPALETTE *)&LogPalette);
}	//DibStream::CreatePalette()

//---------------------------------------------------------------------------
//  DibStream::Draw
//      corresponded bpp  - all
//---------------------------------------------------------------------------
void DibStream::Draw(HDC hdc, int X, int Y)
{
	if(Buffer == NULL) return;

	StretchDraw(hdc, X, Y, GetWidth(), GetHeight());
}	//DibStream::Draw()

//---------------------------------------------------------------------------
//  DibStream::StretchDraw
//      corresponded bpp  - all
//---------------------------------------------------------------------------
void DibStream::StretchDraw(HDC hdc, int X, int Y, int W, int H)
{
	if(Buffer == NULL) return;

	StretchDIBits(hdc,
		X, Y, W, H, //dest
		0, 0, GetWidth(), GetHeight(), //source
		GetBuffer(), GetInfo(),
		DIB_RGB_COLORS, SRCCOPY);
}	//DibStream::StretchDraw()

//---------------------------------------------------------------------------
//  DibStream::StretchDrawRect       1998/8/16
//      corresponded bpp  - all
//---------------------------------------------------------------------------
void DibStream::StretchDrawRect(HDC hdc,
	int DstX, int DstY, int DstW, int DstH,
	int SrcX, int SrcY, int SrcW, int SrcH)
{
	if(Buffer == NULL) return;

	StretchDIBits(hdc,
		DstX, DstY, DstW, DstH, //dest
		SrcX, GetHeight()-SrcY-SrcH, SrcW, SrcH, //source
		GetBuffer(), GetInfo(),
		DIB_RGB_COLORS, SRCCOPY);
}	//DibStream::StretchDrawRect()

//---------------------------------------------------------------------------
//  DibStream::DrawWithPalette
//      corresponded bpp  - all
//---------------------------------------------------------------------------
void DibStream::DrawWithPalette(HDC hdc, int X, int Y)
{
	HPALETTE hpal,hpalold;

	if(Buffer == NULL) return;

	hpal = CreatePalette();
	if(hpal != NULL){
		hpalold = SelectPalette(hdc, hpal, FALSE);
		RealizePalette(hdc);
	}

	StretchDraw(hdc, X, Y, GetWidth(), GetHeight());

	if(hpal != NULL){
		SelectPalette(hdc, hpalold, FALSE);
		DeleteObject(hpal);
	}
}	//DibStream::DrawWithPalette()

//---------------------------------------------------------------------------
//  DibStream::Resize                1999/2/13
//      corresponded bpp  - 8,16,24,32
//---------------------------------------------------------------------------
bool DibStream::Resize(int w, int h)
{
	if(Buffer == NULL) return false;

typedef struct {
	int pos;
	WORD k0,k1;
} BiLinearInfo;

	BITMAPINFOHEADER info;
	void *NewBuf = NULL;
	BYTE *src,*dst;
	int x,y,newpitch,bufpitch;
	double fx,fy,dfx,dfy,ipart,dpart;
	BiLinearInfo *gradx,*grady,*blix,*bliy;
	DWORD ddw0,ddw1;

	//set new bitmap-info
	if( !SetBitmapInfo(info, w, IsBottomUp() ? abs(h) : -abs(h), GetBpp()) )
		return false;
	newpitch = DibWidthBytes(&info);
	bufpitch = GetPitch();

	dfx = (double)(GetWidth()-1) / (w-1);
	dfy = (double)(GetHeight()-1) / (h-1);

	//create table of x-axis for bi-linear-filting
	blix = (BiLinearInfo *)malloc(sizeof(BiLinearInfo) * w);
	for(x=0,fx=0,gradx=blix; x<w; x++,fx+=dfx,gradx++)
	{
		dpart = modf(fx, &ipart);
		gradx->pos = (int)ipart;
		gradx->k1 = (WORD)((-2*dpart+3)*dpart*dpart * 256);
		gradx->k0 = (WORD)(256 - gradx->k1);
	}
	blix[w-1].pos=GetWidth()-2; blix[w-1].k0=0; blix[w-1].k1=256;

	//create table of y-axis for bi-linear-filting
	bliy = (BiLinearInfo *)malloc(sizeof(BiLinearInfo) * h);
	for(y=0,fy=0,grady=bliy; y<h; y++,fy+=dfy,grady++)
	{
		dpart = modf(fy, &ipart);
		grady->pos = (int)ipart;
		grady->k1 = (WORD)((-2*dpart+3)*dpart*dpart * 256);
		grady->k0 = (WORD)(256 - grady->k1);
	}
	bliy[h-1].pos=GetHeight()-2; bliy[h-1].k0=0; bliy[h-1].k1=256;

	//create and set resized buffer
	src = (BYTE *)GetBuffer();
	switch(GetBpp()){
	case 8:
		NewBuf = CreateBuffer(&info);
		dst = (BYTE *)NewBuf;
		for(y=0,grady=bliy; y<h; y++,grady++)
		{
			BiLinearInfo by = *grady;
			BYTE *src0 = &src[(by.pos)*bufpitch];
			BYTE *src1 = &src[(by.pos+1)*bufpitch];
			BYTE *dst0 = &dst[y*newpitch];
			for(x=0,gradx=blix; x<w; x++,gradx++)
			{
				BiLinearInfo bx = *gradx;
				*dst0++ = (by.k0 >= 0.5) ?
					((bx.k0>=0.5) ? src0[bx.pos] : src0[bx.pos+1]) :
					((bx.k0>=0.5) ? src1[bx.pos] : src1[bx.pos+1]);
			}
		}
		break;
	case 16:
		NewBuf = CreateBuffer(&info);
		dst = (BYTE *)NewBuf;
		for(y=0,grady=bliy; y<h; y++,grady++)
		{
			BiLinearInfo by = *grady;
			WORD *src0 = (WORD *)&src[(by.pos)*bufpitch];
			WORD *src1 = (WORD *)&src[(by.pos+1)*bufpitch];
			WORD *dst0 = (WORD *)&dst[y*newpitch];
			for(x=0,gradx=blix; x<w; x++,gradx++)
			{
				BiLinearInfo bx = *gradx;
				*dst0++ = (by.k0 >= 0.5) ?
					((bx.k0>=0.5) ? src0[bx.pos] : src0[bx.pos+1]) :
					((bx.k0>=0.5) ? src1[bx.pos] : src1[bx.pos+1]);
			}
		}
		break;
	case 24:
		NewBuf = CreateBuffer(&info);
		dst = (BYTE *)NewBuf;
		for(y=0,grady=bliy; y<h; y++,grady++)
		{
			BiLinearInfo by = *grady;
			BYTE *src0 = &src[(by.pos)*bufpitch];
			BYTE *src1 = &src[(by.pos+1)*bufpitch];
			BYTE *dst0 = &dst[y*newpitch];
			for(x=0,gradx=blix; x<w; x++,gradx++)
			{
				BiLinearInfo bx = *gradx;
				BYTE *s0 = &src0[bx.pos*3];
				BYTE *s1 = &src1[bx.pos*3];
				*dst0++ = (BYTE)((
					((DWORD)s0[0]*bx.k0 + (DWORD)s0[3]*bx.k1)*by.k0 +
					((DWORD)s1[0]*bx.k0 + (DWORD)s1[3]*bx.k1)*by.k1) >> 16);
				*dst0++ = (BYTE)((
					((DWORD)s0[1]*bx.k0 + (DWORD)s0[4]*bx.k1)*by.k0 +
					((DWORD)s1[1]*bx.k0 + (DWORD)s1[4]*bx.k1)*by.k1) >> 16);
				*dst0++ = (BYTE)((
					((DWORD)s0[2]*bx.k0 + (DWORD)s0[5]*bx.k1)*by.k0 +
					((DWORD)s1[2]*bx.k0 + (DWORD)s1[5]*bx.k1)*by.k1) >> 16);
			}
		}
		break;
	case 32:
		NewBuf = CreateBuffer(&info);
		dst = (BYTE *)NewBuf;
		for(y=0,grady=bliy; y<h; y++,grady++)
		{
			BiLinearInfo by = *grady;
			DibBGRA *src0 = (DibBGRA *)&src[(by.pos)*bufpitch];
			DibBGRA *src1 = (DibBGRA *)&src[(by.pos+1)*bufpitch];
			DibBGRA *dst0 = (DibBGRA *)&dst[y*newpitch];
			for(x=0,gradx=blix; x<w; x++,gradx++,dst0++)
			{
				BiLinearInfo bx = *gradx;
				ddw0 = (DWORD)src0[bx.pos  ].b * bx.k0
					 + (DWORD)src0[bx.pos+1].b * bx.k1;
				ddw1 = (DWORD)src1[bx.pos  ].b * bx.k0
					 + (DWORD)src1[bx.pos+1].b * bx.k1;
				dst0->b = (BYTE)((ddw0*by.k0 + ddw1*by.k1) >> 16);
				ddw0 = (DWORD)src0[bx.pos  ].g * bx.k0
					 + (DWORD)src0[bx.pos+1].g * bx.k1;
				ddw1 = (DWORD)src1[bx.pos  ].g * bx.k0
					 + (DWORD)src1[bx.pos+1].g * bx.k1;
				dst0->g = (BYTE)((ddw0*by.k0 + ddw1*by.k1) >> 16);
				ddw0 = (DWORD)src0[bx.pos  ].r * bx.k0
					 + (DWORD)src0[bx.pos+1].r * bx.k1;
				ddw1 = (DWORD)src1[bx.pos  ].r * bx.k0
					 + (DWORD)src1[bx.pos+1].r * bx.k1;
				dst0->r = (BYTE)((ddw0*by.k0 + ddw1*by.k1) >> 16);
				ddw0 = (DWORD)src0[bx.pos  ].a * bx.k0
					 + (DWORD)src0[bx.pos+1].a * bx.k1;
				ddw1 = (DWORD)src1[bx.pos  ].a * bx.k0
					 + (DWORD)src1[bx.pos+1].a * bx.k1;
				dst0->a = (BYTE)((ddw0*by.k0 + ddw1*by.k1) >> 16);
			}
		}
		break;
	}

	free(blix); free(bliy);

	if(NewBuf == NULL)
		return false;

	//set BITMAPINFOHEADER
	*(BITMAPINFOHEADER *)Header = info;
	//free old buffer and set new one
	BufferFree();
	Buffer = NewBuf;
	BufferMemoryType = BUFFER_MEMTYPE;

	return true;
}	//DibStream::Resize()

//---------------------------------------------------------------------------
//  DibStream::ChangeSize            2000/3/4
//      corresponded bpp  - 8,16,24,32
//---------------------------------------------------------------------------
bool DibStream::ChangeSize(int w, int h, DWORD BackColor)
{
	if(Buffer == NULL) return false;

	BITMAPINFOHEADER info;
	BYTE *NewBuf;
	BYTE *src,*dst;
	WORD *dstw;
	DWORD *dstdw;
	int x, y, ch, newpitch, bufpitch, copysize;

	// set new bitmap-info
	if( !SetBitmapInfo(info, w, IsBottomUp() ? abs(h) : -abs(h), GetBpp()) )
		return false;
	newpitch = DibWidthBytes(&info);
	bufpitch = GetPitch();
	ch = min(GetHeight(), abs(h));

	src = (BYTE *)GetBuffer();
	switch(GetBpp()) {
	case 8:
		// alloc buffer and clear
		NewBuf = (BYTE *)CreateBuffer(&info);
		memset(NewBuf, (BYTE)BackColor, newpitch * h);
		//set buffer
		dst = (BYTE *)NewBuf;
		if(GetWidth() > w) src += (GetWidth() - w) / 2;
		else               dst += (w - GetWidth()) / 2;
		if(GetHeight() > h) src += ((GetHeight() - h) / 2) * bufpitch;
		else                dst += ((h - GetHeight()) / 2) * newpitch;
		copysize = min(w, GetWidth());
		for(y=0; y<ch; y++,src+=bufpitch,dst+=newpitch)
			memcpy(dst, src, copysize);
		break;

	case 16:
		// alloc buffer and clear
		NewBuf = (BYTE *)CreateBuffer(&info);
		for(y=0; y<h; y++){
			dstw = (WORD *)&NewBuf[newpitch*y];
			for(x=0; x<w; x++)
				*dstw++ = (WORD)BackColor;
		}
		// set buffer
		dst = (BYTE *)NewBuf;
		if(GetWidth() > w) src += ((GetWidth() - w) / 2) * 2;
		else               dst += ((w - GetWidth()) / 2) * 2;
		if(GetHeight() > h) src += ((GetHeight() - h) / 2) * bufpitch;
		else                dst += ((h - GetHeight()) / 2) * newpitch;
		copysize = min(w, GetWidth()) * 2;
		for(y=0; y<ch; y++,src+=bufpitch,dst+=newpitch)
			memcpy(dst, src, copysize);
		break;

	case 24:
		// alloc buffer and clear
		NewBuf = (BYTE *)CreateBuffer(&info);
		for(y=0; y<h; y++){
			dst = (BYTE *)&NewBuf[newpitch*y];
			for(x=0; x<w; x++){
				*dst++ = (BYTE)(BackColor & 0xFF);
				*dst++ = (BYTE)((BackColor & 0xFF00) >> 8);
				*dst++ = (BYTE)((BackColor & 0xFF0000) >> 16);
			}
		}
		// set buffer
		dst = (BYTE *)NewBuf;
		if(GetWidth() > w) src += ((GetWidth() - w) / 2) * 3;
		else               dst += ((w - GetWidth()) / 2) * 3;
		if(GetHeight() > h) src += ((GetHeight() - h) / 2) * bufpitch;
		else                dst += ((h - GetHeight()) / 2) * newpitch;
		copysize = min(w, GetWidth()) * 3;
		for(y=0; y<ch; y++,src+=bufpitch,dst+=newpitch)
			memcpy(dst, src, copysize);
		break;

	case 32:
		// alloc buffer and clear
		NewBuf = (BYTE *)CreateBuffer(&info);
		for(y=0; y<h; y++){
			dstdw = (DWORD *)&NewBuf[newpitch*y];
			for(x=0; x<w; x++)
				*dstdw++ = BackColor;
		}
		// set buffer
		dst = (BYTE *)NewBuf;
		if(GetWidth() > w) src += ((GetWidth() - w) / 2) * 4;
		else               dst += ((w - GetWidth()) / 2) * 4;
		if(GetHeight() > h) src += ((GetHeight() - h) / 2) * bufpitch;
		else                dst += ((h - GetHeight()) / 2) * newpitch;
		copysize = min(w, GetWidth()) * 4;
		for(y=0; y<ch; y++,src+=bufpitch,dst+=newpitch)
			memcpy(dst, src, copysize);
		break;

	default:
		return false;
	}

	// set BITMAPINFOHEADER
	*(BITMAPINFOHEADER *)Header = info;

	// free old buffer and set new one
	BufferFree();
	Buffer = NewBuf;
	BufferMemoryType = BUFFER_MEMTYPE;

	return true;
}	//DibStream::ChangeSize()

//---------------------------------------------------------------------------
//  DibStream::ClipCut               1999/10/31
//     x,y     position of left-top
//     w       new width  (>0)
//     h       new height (>0)
//
//      corresponded bpp  - 4,8,16,24,32
//---------------------------------------------------------------------------
bool DibStream::ClipCut(int x, int y, int w, int h)
{
	BITMAPINFOHEADER info;
	BYTE *buf;
	int ix,iy,pitch;

	// check position and size
	if(x+w > GetWidth() || y+h > GetHeight())
		return false;
	if(w <= 0 || h <= 0)
		return false;

	// set bitmap-info
	if( !SetBitmapInfo(info, w, h, GetBpp()) )
		return false;
	pitch = DibWidthBytes(&info);

	// set image to new buffer
	switch(GetBpp()){
	case 4:
		buf = (BYTE *)CreateBuffer(&info);
		if(x & 1){
			for(iy=0; iy<h; iy++){
				BYTE *pr = (BYTE *)ScanLine(iy+y);
				BYTE *pw = &buf[pitch*(h-iy-1)];
				for(ix=0; ix<w; ix+=2,*pw++){
					*pw = (BYTE)(pr[ix] >> 4);
					if(ix+1<w)
						*pw |= (BYTE)(pr[ix+1] << 4);
				}
			}
		} else {
			for(iy=0; iy<h; iy++)
				memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		}
		break;
	case 8:
		buf = (BYTE *)CreateBuffer(&info);
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	case 16:
		buf = (BYTE *)CreateBuffer(&info);
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	case 24:
		buf = (BYTE *)CreateBuffer(&info);
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	case 32:
		buf = (BYTE *)CreateBuffer(&info);
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	default:
		return false;
	}

	//set BITMAPINFOHEADER
	*(BITMAPINFOHEADER *)Header = info;
	//free old buffer and set new one
	BufferFree();
	Buffer = buf;
	BufferMemoryType = BUFFER_MEMTYPE;

	return true;
}	//DibStream::ClipCut()

//---------------------------------------------------------------------------
//  DibStream::ClipClone           1998/8/12
//     x,y     position of left-top
//     w       new width  (>0)
//     h       new height (>0)
//
//      corresponded bpp  - 4,8,16,24,32
//---------------------------------------------------------------------------
DibStream *DibStream::ClipClone(int x, int y, int w, int h)
{
	BYTE *buf;
	int ix,iy,pitch;

	// check position,size,bpp
	if(x+w > GetWidth() || y+h > GetHeight())
		return NULL;
	if(w <= 0 || h <= 0)
		return NULL;
	switch(GetBpp()){
	case  4:
	case  8:
	case 16:
	case 24:
	case 32:
		break;
	default:
		return NULL;
	}

	// create clone of DibStream
	DibStream *Dest = new DibStream;
	Dest->Create(w, h, GetBpp());
	buf = (BYTE *)Dest->GetBuffer();
	pitch = Dest->GetPitch();

	// set image to new buffer
	switch(GetBpp()){
	case 4:
		memcpy(Dest->GetPalette(), GetPalette(), GetNumColors()*sizeof(RGBQUAD));
		if(x & 1){
			for(iy=0; iy<h; iy++){
				BYTE *pr = (BYTE *)ScanLine(iy+y);
				BYTE *pw = &buf[pitch*(h-iy-1)];
				for(ix=0; ix<w; ix+=2,*pw++){
					*pw = (BYTE)(pr[ix] >> 4);
					if(ix+1<w)
						*pw |= (BYTE)(pr[ix+1] << 4);
				}
			}
		} else {
			for(iy=0; iy<h; iy++)
				memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		}
		break;
	case 8:
		memcpy(Dest->GetPalette(), GetPalette(), GetNumColors()*sizeof(RGBQUAD));
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	case 16:
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	case 24:
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	case 32:
		for(iy=0; iy<h; iy++)
			memcpy(&buf[pitch*(h-iy-1)], Pixel(x,iy+y), pitch);
		break;
	}

	return Dest;
}	//DibStream::ClipClone()

//---------------------------------------------------------------------------
//  DibStream::InvertTone            1998/8/23
//      corresponded bpp  - 1,4(M),8(M),16,24,32
//---------------------------------------------------------------------------
bool DibStream::InvertTone(void)
{
	BYTE *p = (BYTE *)GetBuffer();
	if(p == NULL) return false;
	BYTE *ep = &p[GetImageSize()];
	for(; p<ep; p++)
		*p = (BYTE)~*p;
	return true;
}	//DibStream::InvertTone()

//---------------------------------------------------------------------------
//  DibStream::Brighten            1999/12/30
//      corresponded bpp  - 1,4,8,24,32
//---------------------------------------------------------------------------
bool DibStream::Brighten(float val)
{
	int x,y,w,h;
	int pitch,size;
	int i;

	if(Buffer == NULL) return false;
	if(val < 0) return false;
	if(val == 1) return true;

	w = GetWidth();
	h = GetHeight();
	pitch = GetPitch();
	size = GetImageSize();

	switch(GetBpp()) {
	case 1:
	case 4:
	case 8:
			{
		int c = GetNumColors();
		RGBQUAD *rgb = GetPalette();
		if(val>1){ //bright
			float div = 1 - 1/val;
			for(i=0; i<c; i++,rgb++){
				rgb->rgbBlue  += (BYTE)((float)(255-rgb->rgbBlue ) * div);
				rgb->rgbGreen += (BYTE)((float)(255-rgb->rgbGreen) * div);
				rgb->rgbRed   += (BYTE)((float)(255-rgb->rgbRed  ) * div);
			}
		} else { //dark
			for(i=0; i<c; i++,rgb++){
				rgb->rgbBlue  = (BYTE)((float)rgb->rgbBlue  * val + 0.5);
				rgb->rgbGreen = (BYTE)((float)rgb->rgbGreen * val + 0.5);
				rgb->rgbRed   = (BYTE)((float)rgb->rgbRed   * val + 0.5);
			}
		}
		break;
			}

	case 24:
			{
		LPBYTE ptr = (LPBYTE)GetBuffer();
		if(val>1){ //bright
			float div = 1 - 1/val;
			for(i=0; i<size; i++,ptr++)
				*ptr += (BYTE)((float)(255-*ptr) * div + 0.0);
		} else { //dark
			for(i=0; i<size; i++,ptr++)
				*ptr = (BYTE)((float)(*ptr) * val + 0.5);
		}
		break;
			}

	case 32:
			{
		LPBYTE buf = (LPBYTE)GetBuffer();
		if(val>1){ //bright
			float div = 1 - 1/val;
			for(y=0; y<h; y++){
				LPBYTE ptr = &buf[y*pitch];
				for(x=0; x<w; x++){
					*ptr += (BYTE)((float)(255-*ptr) * div + 0.0);  ptr++;
					*ptr += (BYTE)((float)(255-*ptr) * div + 0.0);  ptr++;
					*ptr += (BYTE)((float)(255-*ptr) * div + 0.0);  ptr++;
					ptr++;
				}
			}
		} else { //dark
			for(y=0; y<h; y++){
				LPBYTE ptr = &buf[y*pitch];
				for(x=0; x<w; x++){
					*ptr = (BYTE)((float)(*ptr) * val + 0.5);  ptr++;
					*ptr = (BYTE)((float)(*ptr) * val + 0.5);  ptr++;
					*ptr = (BYTE)((float)(*ptr) * val + 0.5);  ptr++;
					ptr++;
				}
			}
		}
		break;
			}

	default:
		return false;
	}

	return true;
}	//DibStream::Brighten()

//---------------------------------------------------------------------------
//  DibStream::Scroll            1998/8/9
//      corresponded bpp  - 4,8,16,24,32
//---------------------------------------------------------------------------
bool DibStream::Scroll(int x, int y)
{
	if(Buffer == NULL) return false;

	int w,h,p,byte;
	int i,j;

	w=GetWidth(); h=GetHeight(); p=GetPitch();
	if(IsBottomUp()) y = -y;

	if(x<0)x=w+x; if(y<0)y=h+y;
	x%=w; y%=h;

	switch(GetBpp()){
	case  4: break;
	case  8: byte=1; break;
	case 16: byte=2; break;
	case 24: byte=3; break;
	case 32: byte=4; break;
	default: return false;
	}

	BYTE *src = (BYTE *)GetBuffer();
	BYTE *lpb, *buf;

	if(x > 0)
	{	//x scroll
		switch(GetBpp()){
		case  4:
			buf = (BYTE *)malloc(p*2);
			for(i=0; i<h; i++)
			{
				lpb = &src[p*i];
				for(j=0; j<p; j++){
					buf[(j<<1)+1] = (BYTE)(lpb[j] & 0x0F);
					buf[(j<<1)+0] = (BYTE)((lpb[j] & 0xF0) >> 4);
				}
				memset(lpb, 0, p);
				for(j=0; j<w-x; j++)
					lpb[(j+x)>>1] |= (BYTE)((j+x)&1 ? buf[j] : buf[j]<<4);
				for(j=0; j<x; j++)
					lpb[j>>1] |= (BYTE)(j&1 ? buf[w-x+j] : buf[w-x+j]<<4);
			}
			free(buf);
			break;
		case  8:
		case 16:
		case 24:
		case 32:
			buf = (BYTE *)malloc(p);
			for(i=0; i<h; i++){
				lpb = &src[p*i];
				memcpy(buf, lpb, p);
				memcpy(&lpb[x*byte], buf, (w-x)*byte);
				memcpy(lpb, &buf[(w-x)*byte], x*byte);
			}
			free(buf);
			break;
		}
	}

	if(y > 0)
	{	//y scroll
		buf = (BYTE *)malloc(p*h);
		memcpy(buf, src, p*h);
		memcpy(&src[p*y], buf, p*(h-y));
		memcpy(src, &buf[p*(h-y)], p*y);
		free(buf);
	}

	return true;
}	//DibStream::Scroll()




