//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-2000, O.Mizno     All rights reserved.
//
//     DibEffect.cpp
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <math.h>
#include "DibLib.h"
#include "FileLib.h"

#undef PI
#define PI 3.14159265658979f
#undef PI2
#define PI2 (PI * 2.0f)

#undef RAD
#define RAD(x) ((x)/180.0f*PI)

#undef ABS
#define ABS(x) ((x)>=0 ? (x) : -(x))


#if !defined(_MSC_VER)

inline float fmodf(float x, float y) { return (double)fmod(x,y); }
inline float sqrtf(float x) { return (double)sqrt(x); }
inline float sinf(float x) { return (double)sin(x); }
inline float cosf(float x) { return (double)cos(x); }
inline float atan2f(float y, float x) { return (double)atan2(y,x); }

#endif



#define MASK1(_x)      ((BYTE)(0x80 >> ((_x) & 7)))
#define INV_MASK1(_x)  ((BYTE)~MASK1(_x))
#define MASK4(_x)      ((BYTE)(0xF0 >> (((_x) & 1) << 2))

#define CLAMP(_x,_a,_b)  if((_x)<(_a)) (_x)=(_a); else if((_x)>(_b)) (_x)=(_b);

inline int clamp(int x, int a, int b)
{
	if(x < a) return a;
	if(x > b) return b;
	return x;
}

inline BYTE clampByte(int x)
{
	if(x <   0) return   0;
	if(x > 255) return 255;
	return (BYTE)x;
}

inline BYTE clampByte(unsigned x)
{
	if(x > 255) return 255;
	return (BYTE)x;
}


struct clamp_rect {
	int x0, x1;
	int y0, y1;
};

void setClampRect(clamp_rect& rect, DibStream *dib, int pixel)
{
	// [0:x0-1],[x0,x1-1],[x1,w-1]
	int w = dib->GetWidth();
	int h = dib->GetHeight();
	rect.x0 = pixel;
	rect.x1 = w - pixel;
	rect.y0 = pixel;
	rect.y1 = h - pixel;
	if(rect.x0 > w) rect.x0 = w;
	if(rect.x0 > rect.x1) rect.x1 = rect.x0;
	if(rect.y0 > h) rect.y0 = h;
	if(rect.y0 > rect.y1) rect.y1 = rect.y0;
}


//---------------------------------------------------------------------------
//  initRow,freeRow              2000/9/11
//---------------------------------------------------------------------------
static void initRow(DibStream *Dib, BYTE **& row)
{
	int h = Dib->GetHeight();
	row = (BYTE **)malloc(sizeof(BYTE *) * h);
	for(int i=0; i<h; i++)
		row[i] = (BYTE *)Dib->ScanLine(i);
}

static void freeRow(BYTE **& row)
{
	free(row);
}


//---------------------------------------------------------------------------
//  getLoopX                     2000/9/13
//    return:  x (0,w-1)
//---------------------------------------------------------------------------
int __fastcall getLoopX(int x, int w, DWORD loop)
{
	if(loop & DIB_EFFECT_LOOP_X)
	{
		if(x >= w) return x % w;
		else if(x < 0) {
			int ix = w - ((-x)%w);
			if(ix == w) return 0;
			return ix;
		}
	} else {
		if(x < 0) return 0;
		else if(x >= w) return w-1;
	}

	return x;
}

int __fastcall getLoopX(int x, int w)
{
	if(x >= w) return x % w;
	if(x < 0) {
		int ix = w - ((-x)%w);
		if(ix == w) return 0;
		return ix;
	}
	return x;
}

//---------------------------------------------------------------------------
//  getLoopY                     2000/9/13
//    return:  y (0,h-1)
//---------------------------------------------------------------------------
int __fastcall getLoopY(int y, int h, DWORD loop)
{
	if(loop & DIB_EFFECT_LOOP_Y)
	{
		if(y >= h) return y % h;
		if(y < 0) {
			int iy = h - ((-y)%h);
			if(iy == h) return 0;
			return iy;
		}
	} else {
		if(y < 0) return 0;
		if(y >= h) return h-1;
	}

	return y;
}

int __fastcall getLoopY(int y, int h)
{
	if(y >= h) return y % h;
	if(y < 0) {
		int iy = h - ((-y)%h);
		if(iy == h) return 0;
		return iy;
	}
	return y;
}


//---------------------------------------------------------------------------
//  GetBright                    2000/9/7
//    return:  brightness (0,65280)
//---------------------------------------------------------------------------
inline int GetBright(const DibBGRA& bgra)
{
	return bgra.r*73 + bgra.g*146 + bgra.b*37;
}

inline int GetBright(int r, int g, int b)
{
	return r*73 + g*146 + b*37;
}

inline int GetBright(const RGBQUAD& pal)
{
	return pal.rgbRed*73 + pal.rgbGreen*146 + pal.rgbBlue*37;
}

//---------------------------------------------------------------------------
//  Get16B,Get16G,Get16R         98.8.9
//    return:  component of color  (0,0x1F)
//---------------------------------------------------------------------------
inline int Get16B(WORD v) { return v & 0x1F; }
inline int Get16G(WORD v) { return (v >> 5) & 0x1F; }
inline int Get16R(WORD v) { return (v >> 10) & 0x1F; }

//---------------------------------------------------------------------------
//  Get24B,Get24G,Get24R         98.8.9
//    return:  component of color  (0,0xFF)
//---------------------------------------------------------------------------
inline int Get24B(DWORD v) { return v & 0xFF; }
inline int Get24G(DWORD v) { return (v >> 8) & 0xFF; }
inline int Get24R(DWORD v) { return (v >> 16) & 0xFF; }

//---------------------------------------------------------------------------
//  Get32B,Get32G,Get32R,Get32A         98.8.9
//    return:  component of color  (0,0xFF)
//---------------------------------------------------------------------------
#define Get32B Get24B
#define Get32G Get24G
#define Get32R Get24R
inline int Get32A(WORD v) { return (v >> 24) & 0xFF; }

//---------------------------------------------------------------------------
//  Set16Color                   98.8.9
//    b,g,r  component of color  (0,0x1F)
//---------------------------------------------------------------------------
inline WORD Set16Color(int b, int g, int r)
{
	return (WORD)(b | (g<<5) | (r<<10));
}

//---------------------------------------------------------------------------
//  Set24Color                   98.8.9
//    b,g,r  component of color  (0,0xFF)
//---------------------------------------------------------------------------
inline DWORD Set24Color(int b, int g, int r)
{
	return (DWORD)(b | (g<<8) | (r<<16));
}

//---------------------------------------------------------------------------
//  Set32Color                   98.8.9
//    b,g,r  component of color  (0,0xFF)
//---------------------------------------------------------------------------
inline DWORD Set32Color(int b, int g, int r, int a)
{
	return (DWORD)(b | (g<<8) | (r<<16) | (a<<24));
}


#define WRAP_X								\
	if(loop & DIB_EFFECT_LOOP_X) {			\
		if(x < 0) {							\
			x = w - fmodf(-x,w);			\
			if(x == w) x = 0;				\
		} else if(x >= w) x = fmodf(x,w);	\
	} else {								\
		if(x < 0) x = 0;					\
		else if(x > w-1) x = w-1;			\
	}										\

#define WRAP_Y								\
	if(loop & DIB_EFFECT_LOOP_Y) {			\
		if(y < 0) {							\
			y = h - fmodf(-y,h);			\
			if(y == h) y = 0;				\
		} else if(y >= h) y = fmodf(y,h);	\
	} else {								\
		if(y < 0) y = 0;					\
		else if(y > h-1) y = h-1;			\
	}										\


//---------------------------------------------------------------------------
//  DibGetColor1                 2000/9/6
//      Get color of pixel in 16bit-DIB
//    return:  index of palette
//---------------------------------------------------------------------------
BYTE DibGetColor1(DibStream *Dib, float x, float y, DWORD loop)
{
	float w = (float)Dib->GetWidth();
	float h = (float)Dib->GetHeight();

	WRAP_X; WRAP_Y;

	int ix = (int)x;

	BYTE p = *(BYTE *)Dib->Pixel(ix,(int)y);
	return (BYTE)((p >> (7-(ix&7))) & 1);
}	//DibGetColor1()

//---------------------------------------------------------------------------
//  DibGetColor4                 2000/9/6
//      Get color of pixel in 16bit-DIB
//    return:  index of palette
//---------------------------------------------------------------------------
BYTE DibGetColor4(DibStream *Dib, float x, float y, DWORD loop)
{
	float w = (float)Dib->GetWidth();
	float h = (float)Dib->GetHeight();

	WRAP_X; WRAP_Y;

	int ix = (int)x;

	BYTE *p = (BYTE *)Dib->Pixel(ix,(int)y);
	return (BYTE)((ix&1) ? (*p&0xF) : ((*p>>4)&0xF));
}	//DibGetColor4()

//---------------------------------------------------------------------------
//  DibGetColor8                 2000/9/6
//      Get color of pixel in 16bit-DIB
//    return:  index of palette
//---------------------------------------------------------------------------
BYTE DibGetColor8(DibStream *Dib, float x, float y, DWORD loop)
{
	float w = (float)Dib->GetWidth();
	float h = (float)Dib->GetHeight();

	WRAP_X; WRAP_Y;

	return *(BYTE *)Dib->Pixel((int)x,(int)y);
}	//DibGetColor8()

//---------------------------------------------------------------------------
//  DibGetColor16                2000/9/6
//      Get color of pixel in 16bit-DIB
//    return:  BGR color
//---------------------------------------------------------------------------
WORD DibGetColor16(DibStream *Dib, float x, float y, DWORD loop)
{
	float w = (float)Dib->GetWidth();
	float h = (float)Dib->GetHeight();

	WRAP_X; WRAP_Y;

	return *(WORD *)Dib->Pixel((int)x,(int)y);
}	//DibGetColor16()

//---------------------------------------------------------------------------
//  DibGetColor24                2000/9/6
//      Get color of pixel in 24bit-DIB (bi-linear)
//    return:  BGR color
//---------------------------------------------------------------------------
DWORD DibGetColor24(DibStream *Dib, float x, float y, DWORD loop)
{
	float w = (float)Dib->GetWidth();
	float h = (float)Dib->GetHeight();

	WRAP_X; WRAP_Y;

	int ix = (int)x;
	int iy = (int)y;
	int offx = ix*3;

	DWORD cb1,cg1,cr1,cb2,cg2,cr2;

	BYTE *p1 = (BYTE *)Dib->ScanLine(iy) + offx;
	int mx = (int)((x - (float)ix) * 256);
	int my = (int)((y - (float)iy) * 256);
	if(iy+1<h || (loop & DIB_EFFECT_LOOP_Y))
	{
		BYTE *p2 = (BYTE *)Dib->ScanLine(iy+1<h ? iy+1 : 0) + offx;

		if(ix+1 < w) {
			cb1 = p1[0]*(256-mx) + p1[3]*mx;
			cg1 = p1[1]*(256-mx) + p1[4]*mx;
			cr1 = p1[2]*(256-mx) + p1[5]*mx;
			cb2 = p2[0]*(256-mx) + p2[3]*mx;
			cg2 = p2[1]*(256-mx) + p2[4]*mx;
			cr2 = p2[2]*(256-mx) + p2[5]*mx;
			cb1 = (cb1*(256-my) + cb2*my) >> 16;
			cg1 = (cg1*(256-my) + cg2*my) >> 16;
			cr1 = (cr1*(256-my) + cr2*my) >> 16;
		} else if(loop & DIB_EFFECT_LOOP_X) {
			cb1 = p1[0]*(256-mx) + p1[ -offx]*mx;
			cg1 = p1[1]*(256-mx) + p1[1-offx]*mx;
			cr1 = p1[2]*(256-mx) + p1[2-offx]*mx;
			cb2 = p2[0]*(256-mx) + p2[0-offx]*mx;
			cg2 = p2[1]*(256-mx) + p2[1-offx]*mx;
			cr2 = p2[2]*(256-mx) + p2[2-offx]*mx;
			cb1 = (cb1*(256-my) + cb2*my) >> 16;
			cg1 = (cg1*(256-my) + cg2*my) >> 16;
			cr1 = (cr1*(256-my) + cr2*my) >> 16;
		} else {
			cb1=p1[0]; cg1=p1[1]; cr1=p1[2];
			cb2=p2[0]; cg2=p2[1]; cr2=p2[2];
			cb1 = (cb1*(256-my) + cb2*my) >> 8;
			cg1 = (cg1*(256-my) + cg2*my) >> 8;
			cr1 = (cr1*(256-my) + cr2*my) >> 8;
		}
	} else {
		if(ix+1 < w) {
			cb1 = (p1[0]*(256-mx) + p1[3]*mx) >> 8;
			cg1 = (p1[1]*(256-mx) + p1[4]*mx) >> 8;
			cr1 = (p1[2]*(256-mx) + p1[5]*mx) >> 8;
		} else if(loop & DIB_EFFECT_LOOP_X) {
			cb1 = (p1[0]*(256-mx) + p1[ -offx]*mx) >> 8;
			cg1 = (p1[1]*(256-mx) + p1[1-offx]*mx) >> 8;
			cr1 = (p1[2]*(256-mx) + p1[2-offx]*mx) >> 8;
		} else {
			cb1=p1[0]; cg1=p1[1]; cr1=p1[2];
		}
	}

	return cb1 | (cg1<<8) | (cr1<<16);
}	//DibGetColor24()

//---------------------------------------------------------------------------
//  DibGetColor32                2000/9/6
//      Get color of pixel in 32bit-DIB (bi-linear)
//    return:  BGRA color
//---------------------------------------------------------------------------
DWORD DibGetColor32(DibStream *Dib, float x, float y, DWORD loop)
{
	float w = (float)Dib->GetWidth();
	float h = (float)Dib->GetHeight();

	WRAP_X; WRAP_Y;

	int ix = (int)x;
	int iy = (int)y;
	int offx = ix*4;

	DWORD cb1,cg1,cr1,ca1,cb2,cg2,cr2,ca2;

	BYTE *p1 = (BYTE *)Dib->ScanLine(iy) + offx;
	int mx = (int)((x - (float)ix) * 256);
	int my = (int)((y - (float)iy) * 256);
	if((loop & DIB_EFFECT_LOOP_Y) || iy+1<h)
	{
		BYTE *p2 = (BYTE *)Dib->ScanLine(iy+1<h ? iy+1 : 0) + offx;

		if(ix+1 < w) {
			cb1 = p1[0]*(256-mx) + p1[4]*mx;
			cg1 = p1[1]*(256-mx) + p1[5]*mx;
			cr1 = p1[2]*(256-mx) + p1[6]*mx;
			ca1 = p1[3]*(256-mx) + p1[7]*mx;
			cb2 = p2[0]*(256-mx) + p2[4]*mx;
			cg2 = p2[1]*(256-mx) + p2[5]*mx;
			cr2 = p2[2]*(256-mx) + p2[6]*mx;
			ca2 = p2[3]*(256-mx) + p2[7]*mx;
			cb1 = (cb1*(256-my) + cb2*my) >> 16;
			cg1 = (cg1*(256-my) + cg2*my) >> 16;
			cr1 = (cr1*(256-my) + cr2*my) >> 16;
			ca1 = (ca1*(256-my) + ca2*my) >> 16;
		} else if(loop & DIB_EFFECT_LOOP_X) {
			cb1 = p1[0]*(256-mx) + p1[ -offx]*mx;
			cg1 = p1[1]*(256-mx) + p1[1-offx]*mx;
			cr1 = p1[2]*(256-mx) + p1[2-offx]*mx;
			ca1 = p1[3]*(256-mx) + p1[3-offx]*mx;
			cb2 = p2[0]*(256-mx) + p2[0-offx]*mx;
			cg2 = p2[1]*(256-mx) + p2[1-offx]*mx;
			cr2 = p2[2]*(256-mx) + p2[2-offx]*mx;
			ca2 = p2[3]*(256-mx) + p2[3-offx]*mx;
			cb1 = (cb1*(256-my) + cb2*my) >> 16;
			cg1 = (cg1*(256-my) + cg2*my) >> 16;
			cr1 = (cr1*(256-my) + cr2*my) >> 16;
			ca1 = (ca1*(256-my) + ca2*my) >> 16;
		} else {
			cb1=p1[0]; cg1=p1[1]; cr1=p1[2]; ca1=p1[2];
			cb2=p2[0]; cg2=p2[1]; cr2=p2[2]; ca2=p2[2];
			cb1 = (cb1*(256-my) + cb2*my) >> 8;
			cg1 = (cg1*(256-my) + cg2*my) >> 8;
			cr1 = (cr1*(256-my) + cr2*my) >> 8;
			ca1 = (ca1*(256-my) + ca2*my) >> 8;
		}
	} else {
		if(ix+1 < w) {
			cb1 = (p1[0]*(256-mx) + p1[4]*mx) >> 8;
			cg1 = (p1[1]*(256-mx) + p1[5]*mx) >> 8;
			cr1 = (p1[2]*(256-mx) + p1[6]*mx) >> 8;
			ca1 = (p1[3]*(256-mx) + p1[7]*mx) >> 8;
		} else if(loop & DIB_EFFECT_LOOP_X) {
			cb1 = (p1[0]*(256-mx) + p1[ -offx]*mx) >> 8;
			cg1 = (p1[1]*(256-mx) + p1[1-offx]*mx) >> 8;
			cr1 = (p1[2]*(256-mx) + p1[2-offx]*mx) >> 8;
			ca1 = (p1[3]*(256-mx) + p1[3-offx]*mx) >> 8;
		} else {
			cb1=p1[0]; cg1=p1[1]; cr1=p1[2]; ca1=p1[3];
		}
	}

	return cb1 | (cg1<<8) | (cr1<<16) | (ca1<<24);
}	//DibGetColor32()


#undef WRAP_X
#undef WRAP_Y


//---------------------------------------------------------------------------
//  DibEffectReverse             98.10.2
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectReverse(bool callback(float,void*), void *data, DibStream *Dib,
	DEReverseInfo *info)
{
	return DibEffectReverse(callback, data, Dib);
}

bool DibEffectReverse(bool callback(float,void*), void *data, DibStream *Dib)
{
	DWORD *p, *ep;

	p = (DWORD *)Dib->GetBuffer();
	ep = (DWORD *)((BYTE *)p + Dib->GetPitch() * Dib->GetHeight());
	for(; p<ep; p++)
		*p = ~(*p);

	return true;
}	//DibEffectReverse()

//---------------------------------------------------------------------------
//  DibEffectRipple              2000/9/7
//    cx,cy      point of center
//    wavelen    length of wave (0+,+NAN)pixel
//    amp        amplitude  (-PI,PI)radian
//    phase      phase of wave (0,2PI) or (-PI,PI)radian
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectRipple(bool callback(float,void*), void *data, DibStream *Dib,
	DERippleInfo *info)
{
	return DibEffectRipple(callback, data, Dib,
		info->center_x, info->center_y, info->wave_length,
		info->amplitude, info->phase, info->loop);
}

bool DibEffectRipple(bool callback(float,void*), void *data, DibStream *Dib,
	float cx, float cy, float wavelen, float amp, float phase, DWORD loop)
{
	int x,y,w,h,pitch;
	BYTE *pw;
	WORD *pww;
	DWORD *pwdw;
	DibStream *tmpdib;
	float dx,dy,rx,ry,r,ang;
	bool result = false;

	if(wavelen == 0) return false;
	if(amp == 0) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	switch(Dib->GetBpp()) {
	case 1:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = sinf(r/wavelen*PI2 + phase) * amp;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				if(DibGetColor1(tmpdib, rx, ry, loop))
					*pw |= MASK1(x);
				if((x&7) == 7)
					pw++;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = sinf(r/wavelen*PI2 + phase) * amp;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				BYTE col = DibGetColor4(tmpdib, rx, ry, loop);
				if(x & 1) *pw++ |= col;
				else      *pw |= (BYTE)(col << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = sinf(r/wavelen*PI2 + phase) * amp;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				*pw++ = DibGetColor8(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = sinf(r/wavelen*PI2 + phase) * amp;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				*pww++ = DibGetColor16(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = sinf(r/wavelen*PI2 + phase) * amp;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				DWORD col = DibGetColor24(tmpdib, rx, ry, loop);
				*pw++ = ((BYTE *)&col)[0];
				*pw++ = ((BYTE *)&col)[1];
				*pw++ = ((BYTE *)&col)[2];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pwdw = (DWORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = sinf(r/wavelen*PI2 + phase) * amp;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				*pwdw++ = DibGetColor32(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		return false;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectRipple()

//---------------------------------------------------------------------------
//  DibEffectRotate              2000/9/7
//    angle      angle  (0,2PI) or (-PI,PI)radian
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectRotate(bool callback(float,void*), void *data, DibStream *Dib,
	DERotateInfo *info)
{
	return DibEffectRotate(callback, data, Dib,
		info->angle, info->loop);
}

bool DibEffectRotate(bool callback(float,void*), void *data, DibStream *Dib,
	float angle, DWORD loop)
{
	int x,y,w,h,pitch;
	BYTE *pw;
	WORD *pww;
	DWORD *pwdw;
	DibStream *tmpdib;
	float cx,cy,dx,dy,rx,ry,cosang,sinang;
	bool result = false;

	if(angle == 0.0f) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	// determine point of center
	cx = (float)(w-1) * 0.5f;
	cy = (float)(h-1) * 0.5f;

	cosang = cosf(angle);
	sinang = sinf(angle);

	switch(Dib->GetBpp()) {
	case 1:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				rx = dx * cosang - dy * sinang + cx;
				ry = dx * sinang + dy * cosang + cy;
				if(DibGetColor1(tmpdib, rx, ry, loop))
					*pw |= MASK1(x);
				if((x & 7) == 7) pw++;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				rx = dx * cosang - dy * sinang + cx;
				ry = dx * sinang + dy * cosang + cy;
				BYTE col = DibGetColor4(tmpdib, rx, ry, loop);
				if(x & 1) *pw++ |= col;
				else      *pw |= (BYTE)(col << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++) {
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++) {
				dx = (float)x - cx;
				rx = dx * cosang - dy * sinang + cx;
				ry = dx * sinang + dy * cosang + cy;
				*pw++ = DibGetColor8(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++) {
			pww = (WORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++) {
				dx = (float)x - cx;
				rx = dx * cosang - dy * sinang + cx;
				ry = dx * sinang + dy * cosang + cy;
				*pww++ = DibGetColor16(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				rx = dx * cosang - dy * sinang + cx;
				ry = dx * sinang + dy * cosang + cy;
				DWORD col = DibGetColor24(tmpdib, rx, ry, loop);
				pw[0] = ((BYTE *)&col)[0];
				pw[1] = ((BYTE *)&col)[1];
				pw[2] = ((BYTE *)&col)[2];
				pw += 3;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pwdw = (DWORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				rx = dx * cosang - dy * sinang + cx;
				ry = dx * sinang + dy * cosang + cy;
				*pwdw++ = DibGetColor32(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		return false;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectRotate()

//---------------------------------------------------------------------------
//  DibEffectWave                2000/9/7
//    wavelen    length of wave  (0+,+NAN)pixel
//    amp        amplitude  (-NAN,+NAN)pixel
//    phase      phase of wave (0,2PI) or (-PI,PI)radian
//    angle      angle(direction) of wave  (0,2PI) or (-PI,PI)radian
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectWave(bool callback(float,void*), void *data, DibStream *Dib,
	DEWaveInfo *info)
{
	return DibEffectWave(callback, data, Dib,
		info->wave_length, info->amplitude, info->phase,
		info->angle, info->loop);
}

bool DibEffectWave(bool callback(float,void*), void *data, DibStream *Dib,
	float wavelen, float amp, float phase, float angle, DWORD loop)
{
	int x,y,w,h,pitch;
	BYTE *pw;
	WORD *pww;
	DWORD *pwdw;
	DibStream *tmpdib;
	float cx,cy,dx,dy,rx,ry,r,k,l,kx,ky;
	float angcos, angsin;
	bool result = false;

	if(wavelen == 0)
		return false;

	if(amp == 0)
		return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	// determine point of center
	cx = (float)(w-1) * 0.5f;
	cy = (float)(h-1) * 0.5f;

	angcos = cosf(angle);
	angsin = sinf(angle);

	switch(Dib->GetBpp()) {
	case 1:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				k = dx*angsin - dy*angcos;
				l = (dx*dx + dy*dy) - k*k;
				l = (l > 0) ? sqrtf(l) : 0;
				r = ABS(k);
				kx = r * angsin;
				ky = r * angcos;
				if(kx*(ky-dy) - ky*(kx+dx) >= 0)
					r = sinf(l/wavelen*PI2 + phase) * amp;
				else
					r = sinf((-l/wavelen)*PI2 + phase) * amp;
				rx = x - r * angsin;
				ry = y + r * angcos;
				if(DibGetColor1(tmpdib, rx, ry, loop))
					*pw |= MASK1(x);
				if((x & 7) == 7) pw++;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				k = dx*angsin - dy*angcos;
				l = (dx*dx + dy*dy) - k*k;
				l = (l > 0) ? sqrtf(l) : 0;
				r = ABS(k);
				kx = r * angsin;
				ky = r * angcos;
				if(kx*(ky-dy) - ky*(kx+dx) >= 0)
					r = sinf(l/wavelen*PI2 + phase) * amp;
				else
					r = sinf((-l/wavelen)*PI2 + phase) * amp;
				rx = x - r * angsin;
				ry = y + r * angcos;
				BYTE col = DibGetColor4(tmpdib, rx, ry, loop);
				if(x & 1) *pw++ |= col;
				else      *pw |= (BYTE)(col << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				k = dx*angsin - dy*angcos;
				l = (dx*dx + dy*dy) - k*k;
				l = (l > 0) ? sqrtf(l) : 0;
				r = ABS(k);
				kx = r * angsin;
				ky = r * angcos;
				if(kx*(ky-dy) - ky*(kx+dx) >= 0)
					r = sinf(l/wavelen*PI2 + phase) * amp;
				else
					r = sinf((-l/wavelen)*PI2 + phase) * amp;
				rx = x - r * angsin;
				ry = y + r * angcos;
				*pw++ = DibGetColor8(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++){
			pww = (WORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++){
				dx = (float)x - cx;
				k = dx*angsin - dy*angcos;
				l = (dx*dx + dy*dy) - k*k;
				l = (l > 0) ? sqrtf(l) : 0;
				r = ABS(k);
				kx = r * angsin;
				ky = r * angcos;
				if(kx*(ky-dy) - ky*(kx+dx) >= 0)
					r = sinf(l/wavelen*PI2 + phase) * amp;
				else
					r = sinf((-l/wavelen)*PI2 + phase) * amp;
				rx = x - r * angsin;
				ry = y + r * angcos;
				*pww++ = DibGetColor16(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++){
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++){
				dx = (float)x - cx;
				k = dx*angsin - dy*angcos;
				l = (dx*dx + dy*dy) - k*k;
				l = (l > 0) ? sqrtf(l) : 0.0f;
				r = ABS(k);
				kx = r * angsin;
				ky = r * angcos;
				if(kx*(ky-dy) - ky*(kx+dx) >= 0)
					r = sinf(l/wavelen*PI2 + phase) * amp;
				else
					r = sinf((-l/wavelen)*PI2 + phase) * amp;
				rx = x - r * angsin;
				ry = y + r * angcos;
				DWORD col = DibGetColor24(tmpdib, rx, ry, loop);
				*pw++ = ((BYTE *)&col)[0];
				*pw++ = ((BYTE *)&col)[1];
				*pw++ = ((BYTE *)&col)[2];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++){
			pwdw = (DWORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++){
				dx = (float)x - cx;
				k = dx*angsin - dy*angcos;
				l = (dx*dx + dy*dy) - k*k;
				l = (l > 0) ? sqrtf(l) : 0;
				r = ABS(k);
				kx = r * angsin;
				ky = r * angcos;
				if(kx*(ky-dy) - ky*(kx+dx) >= 0)
					r = sinf(l/wavelen*PI2 + phase) * amp;
				else
					r = sinf((-l/wavelen)*PI2 + phase) * amp;
				rx = x - r * angsin;
				ry = y + r * angcos;
				*pwdw++ = DibGetColor32(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		return false;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectWave()

//---------------------------------------------------------------------------
//  DibEffectWhirl               2000/9/7
//    cx,cy      point of center
//    angle      angle of rotation per 1pixel (-PI,PI)radian
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectWhirl(bool callback(float,void*), void *data, DibStream *Dib,
	DEWhirlInfo *info)
{
	return DibEffectWhirl(callback, data, Dib,
		info->center_x, info->center_y, info->angle, info->loop);
}

bool DibEffectWhirl(bool callback(float,void*), void *data, DibStream *Dib,
	float cx, float cy, float angle, DWORD loop)
{
	int x,y,w,h,pitch;
	BYTE *pw;
	WORD *pww;
	DWORD *pwdw;
	DibStream *tmpdib;
	float dx,dy,rx,ry,r,ang;
	bool result = false;

	if(angle == 0) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	switch(Dib->GetBpp()) {
	case 1:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = r * angle;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				if(DibGetColor1(tmpdib, rx, ry, loop))
					*pw |= MASK1(x);
				if((x&7) == 7)
					pw++;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = r * angle;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				BYTE col = DibGetColor4(tmpdib, rx, ry, loop);
				if(x & 1) *pw++ |= col;
				else      *pw |= (BYTE)(col << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = r * angle;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				*pw++ = DibGetColor8(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++){
			pww = (WORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++){
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = r * angle;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				*pww++ = DibGetColor16(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++)
			{
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = r * angle;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				DWORD col = DibGetColor24(tmpdib, rx, ry, loop);
				*pw++ = ((BYTE *)&col)[0];
				*pw++ = ((BYTE *)&col)[1];
				*pw++ = ((BYTE *)&col)[2];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++) {
			pwdw = (DWORD *)Dib->ScanLine(y);
			dy = (float)y - cy;
			for(x=0; x<w; x++) {
				dx = (float)x - cx;
				r = sqrtf(dx*dx + dy*dy);
				ang = r * angle;
				rx = dx * cosf(ang) - dy * sinf(ang) + cx;
				ry = dx * sinf(ang) + dy * cosf(ang) + cy;
				*pwdw++ = DibGetColor32(tmpdib, rx, ry, loop);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		return false;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectWhirl()

//---------------------------------------------------------------------------
//  DibEffectMozaic              2000/9/7
//    pixel      size of square  (1,+NAN)pixel
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectMozaic(bool callback(float,void*), void *data, DibStream *Dib,
	DEMozaicInfo *info)
{
	return DibEffectMozaic(callback, data, Dib,
		info->pixel);
}

bool DibEffectMozaic(bool callback(float,void*), void *data, DibStream *Dib,
	int pixel)
{
	int x,y,dx,dy,w,h;
	BYTE **scan;
	BYTE *pw,*pr;
	WORD *pww,*prw;
	BYTE bb;
	DWORD cb,cg,cr,ca;
	int pc;
	bool result = false;

	if(pixel <= 0) return false;
	if(pixel == 1) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	initRow(Dib, scan);

	switch(Dib->GetBpp()) {
	case 1:
		for(y=0; y<h; y+=pixel)
		{
			for(x=0; x<w; x+=pixel)
			{
				// check at the left-top pixel
				if(scan[y][x>>3] & MASK1(x))
				{
					for(dy=0; dy<pixel && y+dy<h; dy++) {
						pw = scan[y+dy];
						for(dx=x; dx<x+pixel && dx<w; dx++)
							pw[dx>>3] |= MASK1(dx);
					}
				}
				else
				{
					for(dy=0; dy<pixel && y+dy<h; dy++) {
						pw = scan[y+dy];
						for(dx=x; dx<x+pixel && dx<w; dx++)
							pw[dx>>3] &= INV_MASK1(dx);
					}
				}
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		for(y=0; y<h; y+=pixel)
		{
			for(x=0; x<w; x+=pixel)
			{
				// get left-top pixel
				if(x & 1) bb = (BYTE)(scan[y][x>>1] & 0x0F);
				else      bb = (BYTE)((scan[y][x>>1] >> 4) & 0x0F);
				// set color
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					pw = scan[y+dy];
					for(dx=x; dx<x+pixel && dx<w; dx++) {
						if(dx & 1)
							pw[dx>>1] = (BYTE)((pw[dx>>1] & 0xF0) | bb);
						else
							pw[dx>>1] = (BYTE)((pw[dx>>1] & 0x0F) | (bb << 4));
					}
				}
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		for(y=0; y<h; y+=pixel)
		{
			for(x=0; x<w; x+=pixel)
			{
				// get left-top pixel
				bb = scan[y][x];
				// set color
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					pw = scan[y+dy] + x;
					for(dx=0; dx<pixel && x+dx<w; dx++)
						*pw++ = bb;
				}
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y+=pixel)
		{
			for(x=0; x<w; x+=pixel)
			{
				WORD col;
				cb=cg=cr=0; pc=0;
				// get color of average
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					prw = (WORD *)(scan[y+dy] + x*2);
					for(dx=0; dx<pixel && x+dx<w; dx++,pc++) {
						col = *prw++;
						cb += Get16B(col);
						cg += Get16G(col);
						cr += Get16R(col);
					}
				}
				if(!pc) continue;
				col = Set16Color(cb/pc, cg/pc, cr/pc);
				// set color
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					pww = (WORD *)(scan[y+dy] + x*2);
					for(dx=0; dx<pixel && x+dx<w; dx++) {
						*pww++ = col;
					}
				}
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y+=pixel)
		{
			for(x=0; x<w; x+=pixel)
			{
				cb=cg=cr=0; pc=0;
				// get color of average
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					pr = scan[y+dy] + x*3;
					for(dx=0; dx<pixel && x+dx<w; dx++,pc++) {
						cb += *pr++;
						cg += *pr++;
						cr += *pr++;
					}
				}
				if(!pc) continue;
				cb/=pc; cg/=pc; cr/=pc;
				// set color
				for(dy=0; dy<pixel && y+dy<h; dy++){
					pw = scan[y+dy] + x*3;
					for(dx=0; dx<pixel && x+dx<w; dx++){
						*pw++ = (BYTE)cb;
						*pw++ = (BYTE)cg;
						*pw++ = (BYTE)cr;
					}
				}
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y+=pixel)
		{
			for(x=0; x<w; x+=pixel)
			{
				cb=cg=cr=ca=0; pc=0;
				// get color of average
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					pr = scan[y+dy] + x*4;
					for(dx=0; dx<pixel && x+dx<w; dx++,pc++) {
						cb += *pr++;
						cg += *pr++;
						cr += *pr++;
						ca += *pr++;
					}
				}
				if(!pc) continue;
				cb/=pc; cg/=pc; cr/=pc; ca/=pc;
				// set color
				for(dy=0; dy<pixel && y+dy<h; dy++) {
					pw = scan[y+dy] + x*4;
					for(dx=0; dx<pixel && x+dx<w; dx++) {
						*pw++ = (BYTE)cb;
						*pw++ = (BYTE)cg;
						*pw++ = (BYTE)cr;
						*pw++ = (BYTE)ca;
					}
				}
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	freeRow(scan);

	return result;
}	//DibEffectMozaic()


//---------------------------------------------------------------------------
//  DibEffectDiffuse             2000/9/10
//    pixel      value of diffusion of pixel  (0,+NAN)pixel
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectDiffuse(bool callback(float,void*), void *data, DibStream *Dib,
	DEDiffuseInfo *info)
{
	return DibEffectDiffuse(callback, data, Dib,
		info->pixel, info->loop);
}

bool DibEffectDiffuse(bool callback(float,void*), void *data, DibStream *Dib,
	int pixel, DWORD loop)
{
	int x,y,dx,dy,w,h,pitch;
	int pixel2;
	BYTE *pw,*pr;
	WORD *pww;
	DWORD *pwdw;
	DibStream *tmpdib;
	bool result = false;
	BYTE **row;

	if(pixel < 0) return false;
	if(pixel == 0) return true;

	pixel2 = pixel*2 + 1;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	tmpdib = Dib->Clone();
	initRow(tmpdib, row);

	switch(Dib->GetBpp()) {
	case 1:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			for(x=0; x<w; x++)
			{
				dx = getLoopX(((rand() >> 8) % pixel2) - pixel + x, w, loop);
				dy = getLoopY(((rand() >> 8) % pixel2) - pixel + y, h, loop);
				pr = row[dy];
				if(pr[dx>>3] & MASK1(dx))
					*pw |= MASK1(x);
				if((x & 7) == 7) pw++;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			for(x=0; x<w; x++)
			{
				dx = getLoopX(((rand() >> 8) % pixel2) - pixel + x, w, loop);
				dy = getLoopY(((rand() >> 8) % pixel2) - pixel + y, h, loop);
				pr = row[dy];
				BYTE ch;
				if(dx & 1) ch = (BYTE)(pr[dx>>1] & 0x0F);
				else       ch = (BYTE)(pr[dx>>1] >> 4);
				if(x & 1) *pw++ |= ch;
				else      *pw |= (BYTE)(ch << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				dx = getLoopX(((rand() >> 8) % pixel2) - pixel + x, w, loop);
				dy = getLoopY(((rand() >> 8) % pixel2) - pixel + y, h, loop);
				*pw++ = row[dy][dx];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				dx = getLoopX(((rand() >> 8) % pixel2) - pixel + x, w, loop);
				dy = getLoopY(((rand() >> 8) % pixel2) - pixel + y, h, loop);
				*pww++ = row[dy][dx*2];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				dx = getLoopX(((rand() >> 8) % pixel2) - pixel + x, w, loop);
				dy = getLoopY(((rand() >> 8) % pixel2) - pixel + y, h, loop);
				pr = (BYTE *)&row[dy][dx*3];
				pw[0] = pr[0];
				pw[1] = pr[1];
				pw[2] = pr[2];
				pw += 3;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pwdw = (DWORD *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				dx = getLoopX(((rand() >> 8) % pixel2) - pixel + x, w, loop);
				dy = getLoopY(((rand() >> 8) % pixel2) - pixel + y, h, loop);
				*pwdw++ = *(DWORD *)&row[dy][dx*4];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	freeRow(row);
	delete tmpdib;

	return result;
}	//DibEffectDiffuse()


//---------------------------------------------------------------------------
//  DibEffectCompose             2000/9/10
//    value     value of composition  (0,1)
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectCompose(bool callback(float,void*), void *data, DibStream *Dib,
	DEComposeInfo *info)
{
	if(info->filename != NULL)
		return DibEffectCompose(callback, data, Dib,
			info->filename, info->value, info->type);

	return DibEffectCompose(callback, data, Dib,
		info->source, info->value, info->type);
}

bool DibEffectCompose(bool callback(float,void*), void *data, DibStream *DestDib,
	DibStream *SourceDib, float value, DWORD type)
{
	int x,y,w,h;
	int sx,sy,sw,sh;
	int dw;
	BYTE blend = (BYTE)(value*255+0.5);
	BYTE *pr,*pw;
	DibStream *tmpdib = NULL, *tardib;

	// check bpp
	switch(DestDib->GetBpp()){
	case 8:
	case 24:
	case 32:
		break;
	default:
		return false;
	}

	w = DestDib->GetWidth();
	h = DestDib->GetHeight();

	tardib = SourceDib;

	// fit bpp
	if(SourceDib->GetBpp() != DestDib->GetBpp()){
		if(tmpdib == NULL){
			tmpdib = SourceDib->Clone();
			tardib = tmpdib;
		}
		tmpdib->ChangeBpp(DestDib->GetBpp());
	}

	switch(type){
	case DIB_COMPOSE_RESIZE:
		if(SourceDib->GetWidth() != w || SourceDib->GetHeight() != h)
		{
			if(tmpdib == NULL){
				tmpdib = SourceDib->Clone();
				tardib = tmpdib;
			}
			tmpdib->Resize(w,h);
		}

		switch(DestDib->GetBpp()){
		case 8:
			dw = w;
			break;
		case 24:
			dw = w*3;
			break;
		case 32:
			dw = w*4;
			break;
		}
		for(y=0; y<h; y++){
			pr = (BYTE *)tardib->ScanLine(y);
			pw = (BYTE *)DestDib->ScanLine(y);
			for(x=0; x<dw; x++,pr++,pw++){
				*pw = (BYTE)(((*pr) * blend + (*pw) * (255-blend)) >> 8);
			}
		}
		break;
	case DIB_COMPOSE_TILE:
		sw = tardib->GetWidth();
		sh = tardib->GetHeight();
		switch(DestDib->GetBpp()){
		case 8:
			dw = w;
			break;
		case 24:
			sw *= 3;
			dw = w*3;
			break;
		case 32:
			sw *= 4;
			dw = w*4;
			break;
		}
		for(y=sy=0; y<h; y++){
			pr = (BYTE *)tardib->ScanLine(sy);
			pw = (BYTE *)DestDib->ScanLine(y);
			for(x=sx=0; x<dw; x++,pw++){
				*pw = (BYTE)((pr[sx] * blend + (*pw) * (255-blend)) >> 8);
				if(++sx >= sw) sx=0;
			}
			if(++sy >= sh) sy=0;
		}
		break;
	}

	if(tmpdib)
		delete tmpdib;

	return true;
}	//DibEffectCompose()


bool DibEffectCompose(bool callback(float,void*), void *data, DibStream *DestDib,
	const char *filename, float value, DWORD type)
{
	DibStream SourceDib;
	char ext[MAX_PATH];
	bool result = false;

	GetFileExtension(ext, filename);

#define CHK_EXT(e) (strcmpi(ext, e)==0)
	if(CHK_EXT("bmp")) {
		result = SourceDib.LoadFromFile(filename, DIB_LOAD_ALL);
	} else if(CHK_EXT("ppm")) {
		result = LoadPPM(filename, &SourceDib);
	} else if(CHK_EXT("pgm")) {
		result = LoadPGM(filename, &SourceDib);
	} else if(CHK_EXT("tga")) {
		result = LoadTGA(filename, &SourceDib);
	} else if(CHK_EXT("png")) {
		result = LoadPNG(filename, &SourceDib);
	} else if(CHK_EXT("jpg")) {
		result = LoadJPEG(filename, &SourceDib);
	} else if(CHK_EXT("iff")) {
		result = LoadIFF(filename, &SourceDib);
	}
#undef CHK_EXT

	if(result && (SourceDib.GetBuffer() != NULL))
		result = DibEffectCompose(callback, data, DestDib, &SourceDib, value, type);

	return result;
}	//DibEffectCompose()


//---------------------------------------------------------------------------
//  DibEffectMotionBlur          98.8.9
//    speed        speed of blur  (1,+NAN)pixel
//    angle        angle of blur  (-PI,+PI) or (0,+2PI)radian
//
//      corresponded bpp  - 8(M),16,24,32
//---------------------------------------------------------------------------
bool DibEffectMotionBlur(bool callback(float,void*), void *data, DibStream *Dib,
	DEMotionBlurInfo *info)
{
	return DibEffectMotionBlur(callback, data, Dib,
		info->speed, info->angle, info->loop);
}

bool DibEffectMotionBlur(bool callback(float,void*), void *data, DibStream *Dib,
	int speed, float angle, DWORD loop)
{
	typedef struct {
		float dx,dy;
	} BlurPixel;

	int i,x,y,w,h;
	DibStream *tmpdib;
	BYTE *pw;
	WORD *pww;
	float dx,dy;
	BlurPixel *blur,*cb;
	int b,g,r,a,div;
	bool result = false;

	if(speed < 1) return false;
	if(speed == 1) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	// set blur table
	blur = (BlurPixel *)malloc(sizeof(BlurPixel) * speed);
	for(i=0,div=0; i<(int)speed; i++){
		blur[i].dx = (float)i * cosf(angle);
		blur[i].dy = (float)i * sinf(angle);
		div += speed - i;
	}
	tmpdib = Dib->Clone();

	switch(Dib->GetBpp()) {
	case 8:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y;
			for(x=0; x<w; x++)
			{
				dx = (float)x;
				b=0;
				for(i=speed,cb=blur; i>0; i--,cb++){
					b += DibGetColor8(tmpdib,
						dx+cb->dx, dy+cb->dy, loop) * i;
				}
				*pw++ = (BYTE)(b / div);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			dy = (float)y;
			for(x=0; x<w; x++)
			{
				dx = (float)x;
				b=g=r=0;
				for(i=speed,cb=blur; i>0; i--,cb++){
					WORD col = DibGetColor16(tmpdib,
						dx+cb->dx, dy+cb->dy, loop);
					b += Get16B(col) * i;
					g += Get16G(col) * i;
					r += Get16R(col) * i;
				}
				*pww++ = Set16Color(b/div, g/div, r/div);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y;
			for(x=0; x<w; x++)
			{
				dx = (float)x;
				b=g=r=0;
				for(i=speed,cb=blur; i>0; i--,cb++){
					DWORD col = DibGetColor24(tmpdib,
						dx+cb->dx, dy+cb->dy, loop);
					b += (col & 0x0000FF) * i;
					g += ((col & 0x00FF00) >> 8) * i;
					r += ((col & 0xFF0000) >> 16) * i;
				}
				pw[0] = (BYTE)(b / div);
				pw[1] = (BYTE)(g / div);
				pw[2] = (BYTE)(r / div);
				pw += 3;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			dy = (float)y;
			for(x=0; x<w; x++)
			{
				dx = (float)x;
				b=g=r=a=0;
				for(i=speed,cb=blur; i>0; i--,cb++){
					DWORD col = DibGetColor32(tmpdib,
						dx+cb->dx, dy+cb->dy, loop);
					b += (col & 0x0000FF) * i;
					g += ((col & 0x00FF00) >> 8) * i;
					r += ((col & 0xFF0000) >> 16) * i;
					a += ((col & 0xFF000000) >> 24) * i;
				}
				pw[0] = (BYTE)(b / div);
				pw[1] = (BYTE)(g / div);
				pw[2] = (BYTE)(r / div);
				pw[3] = (BYTE)(a / div);
				pw += 4;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	free(blur);
	delete tmpdib;

	return result;
}	//DibEffectMotionBlur()


//---------------------------------------------------------------------------
//  DibEffectGaussianBlur        2000/9/15
//    pixel        pixel of effect  (0,255)pixel
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectGaussianBlur(bool callback(float,void*), void *data, DibStream *Dib,
	float pixel, DWORD loop)
{
	int x,y,dx,dy,ix,iy,w,h,coefw;
	unsigned r,g,b,a;
	DibStream *tmpdib;
	BYTE **rscan;
	BYTE *pr,*pw;
	float *fcoef, *fcurcoef;
	float coef_sum = 0.0f;
	unsigned *coef, *curcoef;
	clamp_rect crect;
	bool result = false;

	int ipixel = (int)floor(pixel);
	if(ipixel < 0) return false;
	if(ipixel == 0) return true;
	if(ipixel > 255) { ipixel = 255; pixel = 255.0f; }

#define GAUSS_COEF 3.0f

	// initialize gaussian mask
	coefw = ipixel * 2 + 1;
	fcoef = (float *)malloc(sizeof(float) * coefw * coefw);
	coef = (unsigned *)malloc(sizeof(unsigned) * coefw * coefw);
	fcurcoef = &fcoef[ipixel*(1+coefw)]; // center
	for(y=0; y<=ipixel; y++)
	{
		float dy = (float)y / pixel * GAUSS_COEF;
		dy *= dy;
		for(x=0; x<=ipixel; x++) {
			float dx = (float)x / pixel * GAUSS_COEF;
			float v = (float)exp((double)(dx * dx + dy) * (-0.5));
			fcurcoef[x+y*coefw] = v;
			coef_sum += v;
			if(x) { fcurcoef[-x+y*coefw] = v; coef_sum += v; }
			if(y) { fcurcoef[ x-y*coefw] = v; coef_sum += v; }
			if(x && y) { fcurcoef[-x+-y*coefw] = v; coef_sum += v; }
		}
	}
	if(coef_sum == 0.0f)
		return false;

#undef GAUSS_COEF

	coef_sum = 65536.0f / coef_sum;
	fcurcoef = fcoef;
	curcoef  = coef;
	for(y=-ipixel; y<=ipixel; y++) {
		for(x=-ipixel; x<=ipixel; x++) {
			*curcoef++ = (unsigned)(*fcurcoef++ * coef_sum);
		}
	}

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	tmpdib = Dib->Clone();
	initRow(tmpdib, rscan);

	setClampRect(crect, Dib, ipixel);

	switch(Dib->GetBpp()) {
	case 8:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				b = g = r = 0;
				curcoef = coef;
				for(dy=y-ipixel; dy<=y+ipixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = rscan[iy];
					for(dx=x-ipixel; dx<=x+ipixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 3;
						b += *curcoef++ * pr[ix];
					}
				}
				*pw++ = (BYTE)(b >> 16);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
#define GET_PR(_x)  b += c * pr[(_x)+0];  \
					g += c * pr[(_x)+1];  \
					r += c * pr[(_x)+2];
#define SET_PW  pw[0] = (BYTE)(b >> 16);  \
				pw[1] = (BYTE)(g >> 16);  \
				pw[2] = (BYTE)(r >> 16);  \
				pw += 3;

		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			if(loop)
			{
				for(x=0; x<w; x++)
				{
					b = g = r = 0;
					curcoef = coef;
					for(dy=y-ipixel; dy<=y+ipixel; dy++)
					{
						iy = getLoopY(dy, h, loop);
						pr = rscan[iy];
						for(dx=x-ipixel; dx<=x+ipixel; dx++)
						{
							ix = getLoopX(dx, w, loop) * 3;
							unsigned c = *curcoef++;
							GET_PR(ix);
						}
					}
					SET_PW;
				}
			}
			else
			{
				for(x=0; x<crect.x0; x++)
				{
					b = g = r = 0;
					curcoef = coef;
					for(dy=y-ipixel; dy<=y+ipixel; dy++)
					{
						pr = rscan[clamp(dy, 0, h-1)];
						for(dx=x-ipixel; dx<=x+ipixel; dx++)
						{
							unsigned c = *curcoef++;
							if(c) {
								ix = clamp(dx, 0, w-1) * 3;
								GET_PR(ix);
							}
						}
					}
					SET_PW;
				}
				for( ; x<crect.x1; x++)
				{
					b = g = r = 0;
					curcoef = coef;
					for(dy=y-ipixel; dy<=y+ipixel; dy++)
					{
						pr = rscan[clamp(dy, 0, h-1)] + (x-ipixel)*3;
						for(dx=ipixel*2; dx>=0; dx--)
						{
							unsigned c = *curcoef++;
							if(c) { GET_PR(0); }
							pr += 3;
						}
					}
					SET_PW;
				}
				for( ; x<w; x++)
				{
					b = g = r = 0;
					curcoef = coef;
					for(dy=y-ipixel; dy<=y+ipixel; dy++)
					{
						pr = rscan[clamp(dy, 0, h-1)];
						for(dx=x-ipixel; dx<=x+ipixel; dx++)
						{
							unsigned c = *curcoef++;
							if(c) {
								ix = clamp(dx, 0, w-1) * 3;
								GET_PR(ix);
							}
						}
					}
					SET_PW;
				}
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

#undef GET_PR
#undef SET_PW

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				b = g = r = a = 0;
				for(dy=y-ipixel; dy<=y+ipixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = rscan[iy];
					for(dx=x-ipixel; dx<=x+ipixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 4;
						unsigned c = *curcoef++;
						b += c * pr[ix+0];
						g += c * pr[ix+1];
						r += c * pr[ix+2];
						a += c * pr[ix+3];
					}
				}
				pw[0] = (BYTE)(b >> 16);
				pw[1] = (BYTE)(g >> 16);
				pw[2] = (BYTE)(r >> 16);
				pw[3] = (BYTE)(a >> 16);
				pw += 4;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	freeRow(rscan);
	free(fcoef);
	free(coef);
	delete tmpdib;

	return result;
}	//DibEffectGaussianBlur()


//---------------------------------------------------------------------------
//  DibEffectBlur                2000/9/13
//    pixel        pixel of effect  (0,+NAN)pixel
//
//      corresponded bpp  - 8(M),16,24,32
//---------------------------------------------------------------------------
bool DibEffectBlur(bool callback(float,void*), void *data, DibStream *Dib,
	DEBlurInfo *info)
{
	return DibEffectBlur(callback, data, Dib,
		info->pixel, info->loop);
}

bool DibEffectBlur(bool callback(float,void*), void *data, DibStream *Dib,
	int pixel, DWORD loop)
{
	int x,y,dx,dy,ix,iy,w,h;
	int r,g,b,a,count;
	DibStream *tmpdib;
	BYTE **rscan;
	BYTE *pr,*pw;
	WORD *prw,*pww;
	bool result = false;

	if(pixel < 0) return false;
	if(pixel == 0) return true;

	count = (pixel*2+1) * (pixel*2+1);

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	tmpdib = Dib->Clone();
	initRow(tmpdib, rscan);

	switch(Dib->GetBpp()) {
	case 8:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				b=g=r=0;
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = rscan[iy];
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						b += pr[ix];
					}
				}
				*pw++ = (BYTE)(b / count);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				b=g=r=0;
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					prw = (WORD *)rscan[iy];
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						b += Get16B(prw[ix]);
						g += Get16G(prw[ix]);
						r += Get16R(prw[ix]);
					}
				}
				*pww++ = Set16Color(b/count, g/count, r/count);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				b=g=r=0;
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = rscan[iy];
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 3;
						b += pr[ix+0];
						g += pr[ix+1];
						r += pr[ix+2];
					}
				}
				*pw++ = (BYTE)(b / count);
				*pw++ = (BYTE)(g / count);
				*pw++ = (BYTE)(r / count);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				b=g=r=a=0;
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = rscan[iy];
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 4;
						b += pr[ix+0];
						g += pr[ix+1];
						r += pr[ix+2];
						a += pr[ix+3];
					}
				}
				*pw++ = (BYTE)(b / count);
				*pw++ = (BYTE)(g / count);
				*pw++ = (BYTE)(r / count);
				*pw++ = (BYTE)(a / count);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	freeRow(rscan);
	delete tmpdib;

	return result;
}	//DibEffectBlur()


//---------------------------------------------------------------------------
//  DibEffectBlurLittle          98.8.11
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectBlurLittle(bool callback(float,void*), void *data, DibStream *Dib,
	DEBlurLittleInfo *info)
{
	return DibEffectBlurLittle(callback, data, Dib,
		info->loop);
}

bool DibEffectBlurLittle(bool callback(float,void*), void *data, DibStream *Dib,
	DWORD loop)
{
	int x,y,w,h;
	int dw;
	BYTE **rscan;
	BYTE *pw,*pr1,*pr2,*pr3;
	DibStream *tmpdib;
	long v;
	bool result = false;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	tmpdib = Dib->Clone();
	initRow(tmpdib, rscan);

	switch(Dib->GetBpp()){
	case 8:
		dw = w;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			if(loop & DIB_EFFECT_LOOP_X){
				for(x=0; x<dw; x++){
					v =           pr1[x]
					  + pr2[x>0 ? x-1 : dw-1] + pr2[x<dw-1 ? x+1 : 0]
								+ pr3[x];
					*pw++ = (BYTE)(v >> 2);
				}
			} else {
				for(x=0; x<dw; x++){
					v =           pr1[x]
					  + pr2[x>0 ? x-1 : 0] + pr2[x<dw-1 ? x+1 : dw-1]
								+ pr3[x];
					*pw++ = (BYTE)(v >> 2);
				}
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 24:
		dw = w * 3;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			if(loop & DIB_EFFECT_LOOP_X){
				for(x=0; x<dw; x+=3){
					v =           pr1[x]
					  + pr2[x>0 ? x-3 : dw-3] + pr2[x<dw-3 ? x+3 : 0]
								+ pr3[x];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+1]
					  + pr2[x>0 ? x-2 : dw-2] + pr2[x<dw-3 ? x+4 : 1]
								+ pr3[x+1];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+2]
					  + pr2[x>0 ? x-1 : dw-1] + pr2[x<dw-3 ? x+5 : 2]
								+ pr3[x+2];
					*pw++ = (BYTE)(v >> 2);
				}
			} else {
				for(x=0; x<dw; x+=3){
					v =           pr1[x]
					  + pr2[x>0 ? x-3 : 0] + pr2[x<dw-3 ? x+3 : dw-3]
								+ pr3[x];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+1]
					  + pr2[x>0 ? x-2 : 1] + pr2[x<dw-3 ? x+4 : dw-2]
								+ pr3[x+1];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+2]
					  + pr2[x>0 ? x-1 : 2] + pr2[x<dw-3 ? x+5 : dw-1]
								+ pr3[x+2];
					*pw++ = (BYTE)(v >> 2);
				}
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 32:
		dw = w * 4;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			if(loop & DIB_EFFECT_LOOP_X){
				for(x=0; x<dw; x+=4){
					v =           pr1[x]
					  + pr2[x>0 ? x-4 : dw-4] + pr2[x<dw-4 ? x+4 : 0]
								+ pr3[x];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+1]
					  + pr2[x>0 ? x-3 : dw-3] + pr2[x<dw-4 ? x+5 : 1]
								+ pr3[x+1];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+2]
					  + pr2[x>0 ? x-2 : dw-2] + pr2[x<dw-4 ? x+6 : 2]
								+ pr3[x+2];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+3]
					  + pr2[x>0 ? x-1 : dw-1] + pr2[x<dw-4 ? x+7 : 3]
								+ pr3[x+3];
					*pw++ = (BYTE)(v >> 2);
				}
			} else {
				for(x=0; x<dw; x+=4){
					v =           pr1[x]
					  + pr2[x>0 ? x-4 : 0] + pr2[x<dw-4 ? x+4 : dw-4]
								+ pr3[x];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+1]
					  + pr2[x>0 ? x-3 : 1] + pr2[x<dw-4 ? x+5 : dw-3]
								+ pr3[x+1];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+2]
					  + pr2[x>0 ? x-2 : 2] + pr2[x<dw-4 ? x+6 : dw-2]
								+ pr3[x+2];
					*pw++ = (BYTE)(v >> 2);
					v =           pr1[x+3]
					  + pr2[x>0 ? x-1 : 3] + pr2[x<dw-4 ? x+7 : dw-1]
								+ pr3[x+3];
					*pw++ = (BYTE)(v >> 2);
				}
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	freeRow(rscan);
	delete tmpdib;

	return result;
}	//DibEffectBlurLittle()


//---------------------------------------------------------------------------
//  DibEffectSharpLittle          98.8.11
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectSharpLittle(bool callback(float,void*), void *data, DibStream *Dib,
	DESharpLittleInfo *info)
{
	return DibEffectSharpLittle(callback, data, Dib,
		info->loop);
}

bool DibEffectSharpLittle(bool callback(float,void*), void *data, DibStream *Dib,
	DWORD loop)
{
	int x,y,w,h;
	int dw;
	BYTE **rscan;
	BYTE *pw,*pr1,*pr2,*pr3;
	DibStream *tmpdib;
	long v;
	bool result = false;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	tmpdib = Dib->Clone();
	initRow(tmpdib, rscan);

	switch(Dib->GetBpp()){
	case 8:
		dw = w;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			if(loop & DIB_EFFECT_LOOP_X){
				for(x=0; x<dw; x++){
					v =         - pr1[x]
					  - pr2[x>0 ? x-1 : dw-1] + pr2[x]*5 - pr2[x<dw-1 ? x+1 : 0]
								- pr3[x];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
				}
			} else {
				for(x=0; x<dw; x++){
					v =         - pr1[x]
					  - pr2[x>0 ? x-1 : 0] + pr2[x]*5 - pr2[x<dw-1 ? x+1 : dw-1]
								- pr3[x];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
				}
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 24:
		dw = w * 3;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			if(loop & DIB_EFFECT_LOOP_X){
				for(x=0; x<dw; x+=3){
					v =         - pr1[x]
					  - pr2[x>0 ? x-3 : dw-3] + pr2[x  ]*5 - pr2[x<dw-3 ? x+3 : 0]
								- pr3[x];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+1]
					  - pr2[x>0 ? x-2 : dw-2] + pr2[x+1]*5 - pr2[x<dw-3 ? x+4 : 1]
								- pr3[x+1];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+2]
					  - pr2[x>0 ? x-1 : dw-1] + pr2[x+2]*5 - pr2[x<dw-3 ? x+5 : 2]
								- pr3[x+2];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
				}
			} else {
				for(x=0; x<dw; x+=3){
					v =         - pr1[x]
					  - pr2[x>0 ? x-3 : 0] + pr2[x  ]*5 - pr2[x<dw-3 ? x+3 : dw-3]
								- pr3[x];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+1]
					  - pr2[x>0 ? x-2 : 1] + pr2[x+1]*5 - pr2[x<dw-3 ? x+4 : dw-2]
								- pr3[x+1];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+2]
					  - pr2[x>0 ? x-1 : 2] + pr2[x+2]*5 - pr2[x<dw-3 ? x+5 : dw-1]
								- pr3[x+2];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
				}
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 32:
		dw = w * 4;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			if(loop & DIB_EFFECT_LOOP_X){
				for(x=0; x<dw; x+=4)
				{
					v =         - pr1[x]
					  - pr2[x>0 ? x-4 : dw-4] + pr2[x  ]*5 - pr2[x<dw-4 ? x+4 : 0]
								- pr3[x];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+1]
					  - pr2[x>0 ? x-3 : dw-3] + pr2[x+1]*5 - pr2[x<dw-4 ? x+5 : 1]
								- pr3[x+1];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+2]
					  - pr2[x>0 ? x-2 : dw-2] + pr2[x+2]*5 - pr2[x<dw-4 ? x+6 : 2]
								- pr3[x+2];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+3]
					  - pr2[x>0 ? x-1 : dw-1] + pr2[x+3]*5 - pr2[x<dw-4 ? x+7 : 3]
								- pr3[x+3];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
				}
			} else {
				for(x=0; x<dw; x+=4)
				{
					v =         - pr1[x]
					  - pr2[x>0 ? x-4 : 0] + pr2[x  ]*5 - pr2[x<dw-4 ? x+4 : dw-4]
								- pr3[x];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+1]
					  - pr2[x>0 ? x-3 : 1] + pr2[x+1]*5 - pr2[x<dw-4 ? x+5 : dw-3]
								- pr3[x+1];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+2]
					  - pr2[x>0 ? x-2 : 2] + pr2[x+2]*5 - pr2[x<dw-4 ? x+6 : dw-2]
								- pr3[x+2];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
					v =         - pr1[x+3]
					  - pr2[x>0 ? x-1 : 3] + pr2[x+3]*5 - pr2[x<dw-4 ? x+7 : dw-1]
								- pr3[x+3];
					if(v>255)v=255; else if(v<0)v=0;
					*pw++ = (BYTE)v;
				}
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	freeRow(rscan);
	delete tmpdib;

	return result;
}	//DibEffectSharpLittle()

//---------------------------------------------------------------------------
//  DibEffectBrightMax          2000/9/10
//    pixel        pixel of effect  (0,+NAN)pixel
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectBrightMax(bool callback(float,void*), void *data, DibStream *Dib,
	DEBrightMaxInfo *info)
{
	return DibEffectBrightMax(callback, data, Dib,
		info->pixel, info->loop);
}

bool DibEffectBrightMax(bool callback(float,void*), void *data, DibStream *Dib,
	int pixel, DWORD loop)
{
	int x,y,dx,dy,ix,iy,w,h,pitch;
	DibStream *tmpdib;
	DibBGRA bgra,maxcol;
	BYTE maxpal;
	WORD maxpalw;
	int bright;
	BYTE *pr,*pw;
	WORD *prw,*pww;
	RGBQUAD *pal;
	bool result = false;

	if(pixel < 0) return false;
	if(pixel == 0) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	tmpdib = Dib->Clone();

	switch(Dib->GetBpp()) {
	case 1:
		pal = Dib->GetPalette();
		bright = (GetBright(pal[0]) > GetBright(pal[1])) ? 2 : 1;
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			for(x=0; x<w; x++)
			{
				maxpal = 0;
				// search maximum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						if(pr[ix>>3] & MASK1(ix)) maxpal |= 1;
						else                      maxpal |= 2;
						if(maxpal == 3) break;
					}
					if(maxpal == 3) break;
				}
				if(maxpal == 3) maxpal = (BYTE)bright;

				if(maxpal == 1)
					pw[x>>3] |= MASK1(x);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		pal = Dib->GetPalette();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			for(x=0; x<w; x++)
			{
				bright = -1;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						BYTE ch;
						if(ix & 1) ch = (BYTE)(pr[ix>>1] >> 4);
						else       ch = (BYTE)(pr[ix>>1] & 0x0F);
						int cb = GetBright(pal[ch]);
						if(bright < cb) {
							bright = cb;
							maxpal = ch;
						}
					}
				}
				if(x & 1) *pw++ |= maxpal;
				else      *pw |= (BYTE)(maxpal << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		pal = Dib->GetPalette();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = -1;
				// search maximum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						int cb = GetBright(pal[pr[ix]]);
						if(bright < cb) {
							bright = cb;
							maxpal = pr[ix];
						}
					}
				}
				*pw++ = maxpal;
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = -1;
				// search maximum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					prw = (WORD *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						WORD col = prw[ix];
						int cb = GetBright(
							(BYTE)Get16B(col),
							(BYTE)Get16G(col),
							(BYTE)Get16R(col));
						if(bright < cb){
							bright = cb;
							maxpalw = col;
						}
					}
				}
				*pww++ = maxpalw;
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = 0;
				maxcol.b = maxcol.g = maxcol.r = 0;
				// search maximum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 3;
						bgra.b = pr[ix+0];
						bgra.g = pr[ix+1];
						bgra.r = pr[ix+2];
						int cb = GetBright(bgra);
						if(bright < cb) {
							bright = cb;
							maxcol = bgra;
						}
					}
				}
				*pw++ = maxcol.b;
				*pw++ = maxcol.g;
				*pw++ = maxcol.r;
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = 0;
				maxcol.b = maxcol.g = maxcol.r = 0;
				// search maximum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 4;
						bgra = DibBGRA(&pr[ix]);
						int cb = GetBright(bgra);
						if(bright < cb){
							bright = cb;
							maxcol = bgra;
						}
					}
				}
				*pw++ = maxcol.b;
				*pw++ = maxcol.g;
				*pw++ = maxcol.r;
				*pw++ = maxcol.a;
			}
			if(callback != NULL) {
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectBrightMax()

//---------------------------------------------------------------------------
//  DibEffectBrightMin          2000/9/10
//    pixel        pixel of effect  (0,+NAN)pixel
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectBrightMin(bool callback(float,void*), void *data, DibStream *Dib,
	DEBrightMaxInfo *info)
{
	return DibEffectBrightMin(callback, data, Dib,
		info->pixel, info->loop);
}

bool DibEffectBrightMin(bool callback(float,void*), void *data, DibStream *Dib,
	int pixel, DWORD loop)
{
	int x,y,dx,dy,ix,iy,w,h,pitch;
	DibStream *tmpdib;
	DibBGRA bgra,maxcol;
	BYTE maxpal;
	WORD maxpalw;
	int bright;
	BYTE *pr,*pw;
	WORD *prw,*pww;
	RGBQUAD *pal;
	bool result = false;

	if(pixel < 0) return false;
	if(pixel == 0) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();
	pitch = Dib->GetPitch();

	tmpdib = Dib->Clone();

	switch(Dib->GetBpp()) {
	case 1:
		pal = Dib->GetPalette();
		bright = (GetBright(pal[0]) > GetBright(pal[1])) ? 1 : 2;
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				maxpal = 0;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						if(pr[ix>>3] & MASK1(ix)) maxpal |= 1;
						else                      maxpal |= 2;
						if(maxpal == 3) break;
					}
					if(maxpal == 3) break;
				}
				if(maxpal == 3) maxpal = (BYTE)bright;

				if(maxpal == 1) pw[x>>3] |= MASK1(x);
				else            pw[x>>3] &= INV_MASK1(x);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		pal = Dib->GetPalette();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			memset(pw, 0, pitch);
			for(x=0; x<w; x++)
			{
				bright = 256*255 + 1;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						BYTE ch;
						if(ix & 1) ch = (BYTE)(pr[ix>>1] >> 4);
						else       ch = (BYTE)(pr[ix>>1] & 0x0F);
						int cb = GetBright(pal[ch]);
						if(bright > cb) {
							bright = cb;
							maxpal = ch;
						}
					}
				}
				if(x & 1) *pw++ |= maxpal;
				else      *pw |= (BYTE)(maxpal << 4);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		pal = Dib->GetPalette();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = 256*255 + 1;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						int cb = GetBright(pal[pr[ix]]);
						if(bright > cb){
							bright = cb;
							maxpal = pr[ix];
						}
					}
				}
				*pw++ = maxpal;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = 256*255 + 1;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					prw = (WORD *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						WORD col = prw[ix];
						int cb = GetBright(
							(BYTE)Get16B(col),
							(BYTE)Get16G(col),
							(BYTE)Get16R(col));
						if(bright > cb) {
							bright = cb;
							maxpalw = col;
						}
					}
				}
				*pww++ = maxpalw;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = 256*255;
				maxcol.b = maxcol.g = maxcol.r = 255;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 3;
						bgra.b = pr[ix+0];
						bgra.g = pr[ix+1];
						bgra.r = pr[ix+2];
						int cb = GetBright(bgra);
						if(bright > cb) {
							bright = cb;
							maxcol = bgra;
						}
					}
				}
				pw[0] = maxcol.b;
				pw[1] = maxcol.g;
				pw[2] = maxcol.r;
				pw += 3;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				bright = 256*255;
				maxcol.b = maxcol.g = maxcol.r = 255;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop) * 4;
						bgra = DibBGRA(&pr[ix]);
						int cb = GetBright(bgra);
						if(bright > cb) {
							bright = cb;
							maxcol = bgra;
						}
					}
				}
				pw[0] = maxcol.b;
				pw[1] = maxcol.g;
				pw[2] = maxcol.r;
				pw[3] = maxcol.a;
				pw += 4;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectBrightMin()

//---------------------------------------------------------------------------
//  DibEffectBrightMiddle        2000/9/9
//    pixel        pixel of effect  (0,+NAN)pixel
//
//      corresponded bpp  - all
//---------------------------------------------------------------------------
bool DibEffectBrightMiddle(bool callback(float,void*), void *data, DibStream *Dib,
	DEBrightMiddleInfo *info)
{
	return DibEffectBrightMiddle(callback, data, Dib,
		info->pixel, info->loop);
}

struct bmidinfo
{
	union {
		BYTE b;
		WORD w;
		DWORD dw;
		BYTE col[4];
	};
	int bright;
};

static bmidinfo *bright_sort(bmidinfo *info, int size, int middle)
{
	int itemp;
	DWORD dwtemp;

	// use 'bubble sort'
	for(int i=0; i<=middle; i++)
	{
		for(int j=size-1; j>i; j--)
		{
			if(info[j].bright > info[j-1].bright) {
				itemp = info[j].bright;
				info[j].bright = info[j-1].bright;
				info[j-1].bright = itemp;
				dwtemp = info[j].dw;
				info[j].dw = info[j-1].dw;
				info[j-1].dw = dwtemp;
			}
		}
	}

	return &info[middle];
}

bool DibEffectBrightMiddle(bool callback(float,void*), void *data, DibStream *Dib,
	int pixel, DWORD loop)
{
	int x,y,dx,dy,ix,iy,w,h;
	DibStream *tmpdib;
	BYTE *pr,*pw;
	bmidinfo *median, *curmed;
	int medmax, mididx;
	RGBQUAD *pal;
	bool result = false;

	if(pixel < 0) return false;
	if(pixel == 0) return true;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	tmpdib = Dib->Clone();

	medmax = (pixel*2+1) * (pixel*2+1);
	mididx = (medmax + 1) / 2;
	median = (bmidinfo *)malloc(sizeof(bmidinfo) * medmax);

	switch(Dib->GetBpp()) {
	case 1:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int count = 0;
				// search minimum bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++)
					{
						ix = getLoopX(dx, w, loop);
						if(pr[ix>>3] & MASK1(ix)) count++;
					}
				}
				if(count >= mididx) pw[x>>3] |= MASK1(x);
				else                pw[x>>3] &= INV_MASK1(x);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 4:
		pal = Dib->GetPalette();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				curmed = median;
				// search midium bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++,curmed++)
					{
						ix = getLoopX(dx, w, loop);
						BYTE ch;
						if(ix & 1) ch = (BYTE)(pr[ix>>1] & 0x0F);
						else       ch = (BYTE)((pr[ix>>1] >> 4) & 0x0F);
						curmed->col[0] = ch;
						curmed->bright = GetBright(
								(BYTE)pal[ch].rgbBlue,
								(BYTE)pal[ch].rgbGreen,
								(BYTE)pal[ch].rgbRed);
					}
				}
				// sort
				curmed = bright_sort(median, medmax, mididx);
				if(x & 1) { *pw = (BYTE)((*pw & 0xF0) | curmed->col[0]); pw++; }
				else *pw = (BYTE)((*pw & 0x0F) | (curmed->col[0] << 4));
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 8:
		pal = Dib->GetPalette();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				curmed = median;
				// search midium bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++,curmed++)
					{
						ix = getLoopX(dx, w, loop);
						curmed->col[0] = pr[ix];
						curmed->bright = GetBright(
								(BYTE)pal[pr[ix]].rgbBlue,
								(BYTE)pal[pr[ix]].rgbGreen,
								(BYTE)pal[pr[ix]].rgbRed);
					}
				}
				// sort
				curmed = bright_sort(median, medmax, mididx);
				*pw++ = curmed->col[0];
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				curmed = median;
				// search midium bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++,curmed++)
					{
						ix = getLoopX(dx, w, loop) * 2;
						WORD cw = *(WORD *)&pr[ix];
						curmed->w = cw;
						curmed->bright = GetBright(
							(BYTE)Get16B(cw), (BYTE)Get16G(cw), (BYTE)Get16R(cw));
					}
				}
				// sort
				curmed = bright_sort(median, medmax, mididx);
				*(DWORD *)&pw[0] = curmed->w;
				pw += 2;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				curmed = median;
				// search midium bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++,curmed++)
					{
						ix = getLoopX(dx, w, loop) * 3;
						curmed->col[0] = pr[ix+0];
						curmed->col[1] = pr[ix+1];
						curmed->col[2] = pr[ix+2];
						curmed->bright = GetBright(pr[ix+2], pr[ix+1], pr[ix+0]);
					}
				}
				// sort
				curmed = bright_sort(median, medmax, mididx);
				pw[0] = curmed->col[0];
				pw[1] = curmed->col[1];
				pw[2] = curmed->col[2];
				pw += 3;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				curmed = median;
				// search midium bright
				for(dy=y-pixel; dy<=y+pixel; dy++)
				{
					iy = getLoopY(dy, h, loop);
					pr = (BYTE *)tmpdib->ScanLine(iy);
					for(dx=x-pixel; dx<=x+pixel; dx++,curmed++)
					{
						ix = getLoopX(dx, w, loop) * 4;
						curmed->dw = *(DWORD *)&pr[ix];
						curmed->bright = GetBright(pr[ix+2], pr[ix+1], pr[ix+0]);
					}
				}
				// sort
				curmed = bright_sort(median, medmax, mididx);
				*(DWORD *)&pw[0] = curmed->dw;
				pw += 4;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		goto EFFECT_EXIT;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;
	free(median);

	return result;
}	//DibEffectBrightMiddle()


//---------------------------------------------------------------------------
//  DibEffectEmboss              98.8.9
//    power        power of emboss (0,+NAN)
//    pixel        pixel (1,+NAN)pixel
//    angle        angle of emboss  (-PI,+PI) or (0,+2PI)radian
//
//      corresponded bpp  - 8(M),16,24,32
//---------------------------------------------------------------------------
bool DibEffectEmboss(bool callback(float,void*), void *data, DibStream *Dib,
	DEEmbossInfo *info)
{
	return DibEffectEmboss(callback, data, Dib,
		info->power, info->pixel, info->angle, info->loop);
}

bool DibEffectEmboss(bool callback(float,void*), void *data, DibStream *Dib,
	float power, int pixel, float angle, DWORD loop)
{
	int i,x,y,w,h;
	float cx,cy,dx,dy,fv;
	int b,g,r,a;
	float fb,fg,fr;
	BYTE *pw;
	WORD *pww;
	DibStream *tmpdib;
	bool result = false;

	if(pixel < 1)
		return false;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	dx = cosf(angle);
	dy = sinf(angle);
	power /= pixel;

	switch(Dib->GetBpp()) {
	case 8:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			cy = (float)y;
			for(x=0; x<w; x++)
			{
				cx = (float)x;
				b=0;
				for(i=pixel; i>0; i--)
				{
					b -= DibGetColor8(tmpdib,
						cx-dx*(float)i, cy-dy*(float)i, loop);
					b += DibGetColor8(tmpdib,
						cx+dx*(float)i, cy+dy*(float)i, loop);
				}
				fv = (float)b*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 16:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pww = (WORD *)Dib->ScanLine(y);
			cy = (float)y;
			for(x=0; x<w; x++)
			{
				cx = (float)x;
				b=g=r=0;
				for(i=pixel; i>0; i--)
				{
					WORD col;
					col = DibGetColor16(tmpdib,
						cx-dx*(float)i, cy-dy*(float)i, loop);
					b -= Get16B(col);
					g -= Get16G(col);
					r -= Get16R(col);
					col = DibGetColor16(tmpdib,
						cx+dx*(float)i, cy+dy*(float)i, loop);
					b += Get16B(col);
					g += Get16G(col);
					r += Get16R(col);
				}
				fb = (float)b*power + 15.5f;
				if(fb<0)fb=0; else if(fb>31) fb=31;
				fg = (float)g*power + 15.5f;
				if(fg<0)fg=0; else if(fg>31) fg=31;
				fr = (float)r*power + 15.5f;
				if(fr<0)fr=0; else if(fr>31) fr=31;
				*pww++ = Set16Color((int)fb, (int)fg, (int)fr);
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			cy = (float)y;
			for(x=0; x<w; x++)
			{
				cx = (float)x;
				b=g=r=0;
				for(i=pixel; i>0; i--)
				{
					DWORD col;
					col = DibGetColor24(tmpdib,
						cx-dx*(float)i, cy-dy*(float)i, loop);
					b -= (col & 0x0000FF);
					g -= ((col & 0x00FF00) >> 8);
					r -= ((col & 0xFF0000) >> 16);
					col = DibGetColor24(tmpdib,
						cx+dx*(float)i, cy+dy*(float)i, loop);
					b += (col & 0x0000FF);
					g += ((col & 0x00FF00) >> 8);
					r += ((col & 0xFF0000) >> 16);
				}
				fv = (float)b*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
				fv = (float)g*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
				fv = (float)r*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		tmpdib = Dib->Clone();
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			cy = (float)y;
			for(x=0; x<w; x++)
			{
				cx = (float)x;
				b=g=r=a=0;
				for(i=pixel; i>0; i--)
				{
					DWORD col;
					col = DibGetColor32(tmpdib,
						cx-dx*(float)i, cy-dy*(float)i, loop);
					b -= (col & 0x0000FF);
					g -= ((col & 0x00FF00) >> 8);
					r -= ((col & 0xFF0000) >> 16);
					a -= ((col & 0xFF000000) >> 24);
					col = DibGetColor32(tmpdib,
						cx+dx*(float)i, cy+dy*(float)i, loop);
					b += (col & 0x0000FF);
					g += ((col & 0x00FF00) >> 8);
					r += ((col & 0xFF0000) >> 16);
					a += ((col & 0xFF000000) >> 24);
				}
				fv = (float)b*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
				fv = (float)g*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
				fv = (float)r*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
				fv = (float)a*power + 127.5f;
				if(fv<0)fv=0; else if(fv>255) fv=255;
				*pw++ = (BYTE)fv;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		return false;
	}

	result = true;

EFFECT_EXIT:
	delete tmpdib;

	return result;
}	//DibEffectEmboss()

//---------------------------------------------------------------------------
//  DibEffectNoise               98.8.9
//    power      power of noise  (0,+NAN)
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectNoise(bool callback(float,void*), void *data, DibStream *Dib,
	DENoiseInfo *info)
{
	return DibEffectNoise(callback, data, Dib,
		info->power);
}

bool DibEffectNoise(bool callback(float,void*), void *data, DibStream *Dib,
	int power)
{
	int x,y,w,h;
	BYTE *pw;
	int v;
	int ip = power * 2 + 1;
	bool result = false;

	if(power < 0)
		return false;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	switch(Dib->GetBpp()) {
	case 8:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int r = ((rand() >> 4) % ip) - power;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int r = ((rand() >> 4) % ip) - power;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int r = ((rand() >> 4) % ip) - power;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = (BYTE)v;
#if 0
				v = (*pw + r);
				if(v<0)v=0; else if(v>255)v=255;
				*pw++ = v;
#else
				pw++;
#endif
			}
			if(callback != NULL) {
				if( !callback((float)(y+1)/h, data) )
					goto EFFECT_EXIT;
			}
		}
		break;

	default:
		return false;
	}

	result = true;

EFFECT_EXIT:

	return result;
}	//DibEffectNoise()

//---------------------------------------------------------------------------
//  DibEffectColorNoise          98.8.9
//    color      color of noise
//    power      power of noise  (0,1)
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectColorNoise(bool callback(float,void*), void *data, DibStream *Dib,
	DEColorNoiseInfo *info)
{
	return DibEffectColorNoise(callback, data, Dib,
		info->color, info->power);
}

bool DibEffectColorNoise(bool callback(float,void*), void *data, DibStream *Dib,
	DibBGRA color, float power)
{
	int x,y,w,h;
	BYTE *pw;
	BYTE bright;
	int ip = (int)(power * 256);
	bool result = false;

	if(ip<0 || ip>256)
		return false;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	switch(Dib->GetBpp()){
	case 8:
		bright = (BYTE)(GetBright(color)>>8);
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int r = ((rand() >> 4) & 0xFF) * ip;
				*pw = (BYTE)((bright * r + *pw * (255*256-r)) / (255*256)); pw++;
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 24:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int r = ((rand() >> 4) & 0xFF) * ip;
				*pw = (BYTE)((color.b * r + *pw * (255*256-r)) / (255*256)); pw++;
				*pw = (BYTE)((color.g * r + *pw * (255*256-r)) / (255*256)); pw++;
				*pw = (BYTE)((color.r * r + *pw * (255*256-r)) / (255*256)); pw++;
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 32:
		for(y=0; y<h; y++)
		{
			pw = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++)
			{
				int r = ((rand() >> 4) & 0xFF) * ip;
				*pw = (BYTE)((color.b * r + *pw * (255*256-r)) / (255*256)); pw++;
				*pw = (BYTE)((color.g * r + *pw * (255*256-r)) / (255*256)); pw++;
				*pw = (BYTE)((color.r * r + *pw * (255*256-r)) / (255*256)); pw++;
				*pw = (BYTE)((color.a * r + *pw * (255*256-r)) / (255*256)); pw++;
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		return false;
	}

	result = true;

EFFECT_EXIT:

	return result;
}	//DibEffectColorNoise()

bool DibEffectColorNoise(bool callback(float,void*), void *data, DibStream *Dib,
	DWORD color, float power)
{
	DibBGRA bgra;
	bgra.b = (BYTE)Get32B(color);
	bgra.g = (BYTE)Get32G(color);
	bgra.r = (BYTE)Get32R(color);
	bgra.a = (BYTE)Get32R(color);
	return DibEffectColorNoise(callback, data, Dib, bgra, power);
}	//DibEffectColorNoise()

//---------------------------------------------------------------------------
//  DibEffectPickEdge            98.10.1
//
//      corresponded bpp  - 8(M),24,32
//---------------------------------------------------------------------------
bool DibEffectPickEdge(bool callback(float,void*), void *data, DibStream *Dib,
	DEPickEdgeInfo *info)
{
	return DibEffectPickEdge(callback, data, Dib,
		info->loop);
}

bool DibEffectPickEdge(bool callback(float,void*), void *data, DibStream *Dib,
	DWORD loop)
{
	int x,y,w,h;
	int dw;
	BYTE *pw,*pr1,*pr2,*pr3;
	unsigned vx,vy;
	DibStream *tmpdib;
	BYTE **rscan;
	bool result = false;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	tmpdib = Dib->Clone();
	initRow(tmpdib, rscan);

	switch(Dib->GetBpp()){
	case 8:
		dw = w;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			for(x=0; x<dw; x++){
				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-1 : dw-1] - pr2[x<dw-1 ? x+1 : 0];
				else
					vx = pr2[x>0 ? x-1 : 0] - pr2[x<dw-1 ? x+1 : dw-1];
				vy = pr1[x] - pr3[x];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 24:
		dw = w*3;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			for(x=0; x<dw; x+=3)
			{
				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-3 : dw-3] - pr2[x<dw-3 ? x+3 : 0];
				else
					vx = pr2[x>0 ? x-3 : 0] - pr2[x<dw-3 ? x+3 : dw-3];
				vy = pr1[x] - pr3[x];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);

				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-2 : dw-2] - pr2[x<dw-3 ? x+4 : 0];
				else
					vx = pr2[x>0 ? x-2 : 0] - pr2[x<dw-3 ? x+4 : dw-2];
				vy = pr1[x+1] - pr3[x+1];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);

				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-1 : dw-1] - pr2[x<dw-3 ? x+5 : 0];
				else
					vx = pr2[x>0 ? x-1 : 0] - pr2[x<dw-3 ? x+5 : dw-1];
				vy = pr1[x+2] - pr3[x+2];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	case 32:
		dw = w*4;
		for(y=0; y<h; y++)
		{
			// get line
			pw = (BYTE *)Dib->ScanLine(y);
			pr2 = rscan[y];
			if(loop & DIB_EFFECT_LOOP_Y){
				pr1 = rscan[y>0 ? y-1 : h-1];
				pr3 = rscan[y+1<h ? y+1 : 0];
			} else {
				pr1 = rscan[y>0 ? y-1 : 0];
				pr3 = rscan[y+1<h ? y+1 : h-1];
			}
			// set pixel
			for(x=0; x<dw; x+=4)
			{
				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-4 : dw-4] - pr2[x<dw-4 ? x+4 : 0];
				else
					vx = pr2[x>0 ? x-4 : 0] - pr2[x<dw-4 ? x+4 : dw-4];
				vy = pr1[x] - pr3[x];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);

				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-3 : dw-3] - pr2[x<dw-4 ? x+5 : 0];
				else
					vx = pr2[x>0 ? x-3 : 0] - pr2[x<dw-4 ? x+5 : dw-3];
				vy = pr1[x+1] - pr3[x+1];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);

				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-2 : dw-2] - pr2[x<dw-4 ? x+6 : 0];
				else
					vx = pr2[x>0 ? x-2 : 0] - pr2[x<dw-4 ? x+6 : dw-2];
				vy = pr1[x+2] - pr3[x+2];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);

				if(loop & DIB_EFFECT_LOOP_X)
					vx = pr2[x>0 ? x-1 : dw-1] - pr2[x<dw-4 ? x+7 : 0];
				else
					vx = pr2[x>0 ? x-1 : 0] - pr2[x<dw-4 ? x+7 : dw-1];
				vy = pr1[x+3] - pr3[x+3];
				*pw++ = (BYTE)sqrt(vx*vx + vy*vy);
			}
			if(callback != NULL){
				if(!callback((float)(y+1)/h, data))
					goto EFFECT_EXIT;
			}
		}
		break;
	default:
		goto EFFECT_EXIT;
	}
	result = true;

EFFECT_EXIT:
	freeRow(rscan);
	delete tmpdib;

	return result;
}	//DibEffectPickEdge()


//---------------------------------------------------------------------------
//  changeBright                 2000/9/11
//---------------------------------------------------------------------------
static bool changeBright(DibStream *Dib, BYTE *coef)
{
	RGBQUAD *pal;
	BYTE *ptr;
	int x,y,w,h;

	w = Dib->GetWidth();
	h = Dib->GetHeight();

	switch(Dib->GetBpp()) {
	case 1:
	case 4:
	case 8:
		pal = Dib->GetPalette();
		w = Dib->GetNumColors();
		for(x=0; x<w; x++,pal++) {
			pal->rgbRed   = coef[pal->rgbRed];
			pal->rgbGreen = coef[pal->rgbGreen];
			pal->rgbBlue  = coef[pal->rgbBlue];
		}
		break;

	case 24:
		for(y=0; y<h; y++)
		{
			ptr = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++) {
				ptr[0] = coef[ptr[0]];
				ptr[1] = coef[ptr[1]];
				ptr[2] = coef[ptr[2]];
				ptr += 3;
			}
		}
		break;

	case 32:
		for(y=0; y<h; y++)
		{
			ptr = (BYTE *)Dib->ScanLine(y);
			for(x=0; x<w; x++) {
				ptr[0] = coef[ptr[0]];
				ptr[1] = coef[ptr[1]];
				ptr[2] = coef[ptr[2]];
				ptr += 4;
			}
		}
		break;

	default:
		return false;
	}

	return true;
}

//---------------------------------------------------------------------------
//  DibEffectGammaCorrect        2000/9/11
//    bright_min   minimam value of bright (0,255)
//    bright_max   maximam value of bright (0,255)   bright_min <= bright_max
//    gamma        gamma value  (0,+NAN)
//
//      corresponded bpp  - 1,4,8,24,32
//---------------------------------------------------------------------------
bool DibEffectGammaCorrect(bool callback(float,void*), void *data, DibStream *Dib,
	int bright_min, int bright_max, float gamma)
{
	bool result;
	int div = bright_max - bright_min + 1;
	gamma = 1.0f / gamma;

	BYTE *coef = (BYTE *)malloc(sizeof(BYTE) * 256);
	for(int i=0; i<256; i++)
	{
		int cb = i - bright_min;
		if(cb >= bright_max)      coef[i] = 0xFF;
		else if(cb <= bright_min) coef[i] = 0;
		else
			coef[i] = (BYTE)((float)pow((double)cb/(double)div, gamma) * (float)div + 0.5f);
	}

	result = changeBright(Dib, coef);

	free(coef);
	return result;
}


//---------------------------------------------------------------------------
//  DibEffectContrast            2000/9/11
//    bright       bright (-NAN,+NAN)
//    contrast     contrast  (0,+NAN)
//
//      corresponded bpp  - 1,4,8,24,32
//---------------------------------------------------------------------------
bool DibEffectContrast(bool callback(float,void*), void *data, DibStream *Dib,
	float bright, float contrast)
{
	bool result;
	float offset = (1.0f - contrast) * 0.5f + bright;
	BYTE *coef = (BYTE *)malloc(sizeof(BYTE) * 256);
	for(int i=0; i<256; i++)
		coef[i] = clampByte(
			(int)(255.0f * ((float)i / 255.0f * contrast + offset) + 0.5f) );

	result = changeBright(Dib, coef);

	free(coef);
	return result;
}




