//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-2000, O.Mizno     All rights reserved.
//
//     DibJpeg.cpp
//       LoadJPEG()
//       SaveJPEG()
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "Dib.h"
#include "mlibjpg.h"

typedef struct {
	DibStream *dib;
	int height;
	bool (*callback)(float,void*);
	void *callback_data;
} CallbackInfo;

//---------------------------------------------------------------------------
//  LoadJPEG
//---------------------------------------------------------------------------
static int HeaderCallback(LoadJPGInfo *info, void *data)
{
	CallbackInfo *call = (CallbackInfo *)data;
	bool result = call->dib->Create(info->width, info->height, info->bpp);

	call->height = info->height;

	if(info->bpp == 8 && info->num_palette > 0){
		memcpy(call->dib->GetPalette(), info->palette,
			sizeof(RGBQUAD) * info->num_palette);
	}

	return result;
}

static void *ReadCallback(int line_number, void *data)
{
	CallbackInfo *call = (CallbackInfo *)data;

	if(call->callback != NULL){
		if(!call->callback((float)line_number/call->height, call->callback_data))
			return NULL;
	}

	return call->dib->ScanLine(line_number);
}


bool LoadJPEG(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	lpLoadJPG *jpg_proc;
	CallbackInfo call;
	HINSTANCE hInst;
	bool result;

	hInst = LoadLibrary("mlibjpg.dll");
	if(hInst == NULL)
		return false;

	*(FARPROC *)&jpg_proc = GetProcAddress(hInst, "LoadJPG");
	if(jpg_proc == NULL)
	{
		FreeLibrary(hInst);
		return false;
	}

	call.dib = Dib;
	call.callback = callback;
	call.callback_data = data;

	result = (jpg_proc(FileName, HeaderCallback, ReadCallback, &call) != 0);

	FreeLibrary(hInst);

	return result;
}

//---------------------------------------------------------------------------
//  SaveJPEG
//---------------------------------------------------------------------------
static void *WriteCallback(int line_number, void *data)
{
	CallbackInfo *call = (CallbackInfo *)data;

	if(call->callback != NULL){
		if(!call->callback((float)line_number/call->height, call->callback_data))
			return NULL;
	}

	return call->dib->ScanLine(line_number);
}


bool SaveJPEG(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	lpSaveJPG *jpg_proc;
	CallbackInfo call;
	HINSTANCE hInst;
	bool result;
	SaveJPGInfo info;

	hInst = LoadLibrary("mlibjpg.dll");
	if(hInst == NULL)
		return false;

	*(FARPROC *)&jpg_proc = GetProcAddress(hInst, "SaveJPG");
	if(jpg_proc == NULL)
	{
		FreeLibrary(hInst);
		return false;
	}

	memset(&info, 0, sizeof(info));
	info.width = Dib->GetWidth();
	info.height = Dib->GetHeight();
	info.bpp = 24;

	call.height = info.height;
	call.callback = callback;
	call.callback_data = data;

	if(Dib->GetBpp() != 24)
	{
		DibStream *tmp = Dib->Clone();
		tmp->ChangeBpp(24);
		call.dib = tmp;
		result = (jpg_proc(FileName, &info, WriteCallback, &call) != 0);
		delete tmp;
	} else {
		call.dib = Dib;
		result = (jpg_proc(FileName, &info, WriteCallback, &call) != 0);
	}

	FreeLibrary(hInst);

	return result;
}



