//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-1999, O.Mizno     All rights reserved.
//
//     DibPng.cpp
//       LoadPNG()
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "Dib.h"

#define  USE_MLIBPNG_DLL   1

#if USE_MLIBPNG_DLL
	#include "mlibpng.h"
#endif


#define streams
#define entire
#define USE_GAMMA 0


#if !USE_MLIBPNG_DLL

//---------------------------------------------------------------------------
//  check_if_png
//---------------------------------------------------------------------------
#define PNG_BYTES_TO_CHECK 4
static int check_if_png(char *file_name, FILE **fh)
{
	char buf[PNG_BYTES_TO_CHECK];

	/* Open the prospective PNG file. */
	if((*fh = fopen(file_name, "rb")) != NULL)
		return 0;

	/* Read in the signature bytes */
	if(fread(buf, 1, PNG_BYTES_TO_CHECK, *fh) != PNG_BYTES_TO_CHECK)
		return 0;

	/* Compare the first PNG_BYTES_TO_CHECK bytes of the signature. */
	return png_sig_cmp((png_bytep)buf, (png_size_t)0, PNG_BYTES_TO_CHECK);
}


//---------------------------------------------------------------------------
//  LoadPNG
//---------------------------------------------------------------------------
bool LoadPNG(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	bool result=false;
	png_structp png_ptr;
	png_infop info_ptr;
	unsigned int sig_read = 0;
	png_uint_32 width, height, row;
	int bit_depth, color_type, interlace_type;
	FILE *fh;
	int i;

	if((fh = fopen(FileName, "rb")) == NULL)
		return false;

	png_ptr = png_create_read_struct(
		PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);

	if(png_ptr == NULL)
	{
		fclose(fh);
		return false;
	}

	// Allocate/initialize the memory for image information.  REQUIRED.
	info_ptr = png_create_info_struct(png_ptr);
	if(info_ptr == NULL)
	{
		fclose(fh);
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		return false;
	}

	// Set error handling if you are using the setjmp/longjmp method (this is
	// the normal method of doing things with libpng).  REQUIRED unless you
	// set up your own error handlers in the png_create_read_struct() earlier.
	if(setjmp(png_ptr->jmpbuf))
		goto FILE_EXIT;

	// One of the following I/O initialization methods is REQUIRED */
#ifdef streams /* PNG file I/O method 1 */
	// Set up the input control if you are using standard C streams */
	png_init_io(png_ptr, fh);

#else no_streams /* PNG file I/O method 2 */
	// If you are using replacement read functions, instead of calling
	// png_init_io() here you would call:
	png_set_read_fn(png_ptr, (void *)user_io_ptr, user_read_fn);
	// where user_io_ptr is a structure you want available to the callbacks */
#endif no_streams /* Use only one I/O method! */

	// If we have already read some of the signature
	png_set_sig_bytes(png_ptr, sig_read);

	// The call to png_read_info() gives us all of the information from the
	// PNG file before the first IDAT (image data chunk).  REQUIRED
	png_read_info(png_ptr, info_ptr);

	png_get_IHDR(png_ptr, info_ptr, &width, &height, &bit_depth, &color_type,
		&interlace_type, NULL, NULL);


/**** Set up the data transformations you want.  Note that these are all
 **** optional.  Only call them if you want/need them.  Many of the
 **** transformations only work on specific types of images, and many
 **** are mutually exclusive.
 ****/

	// tell libpng to strip 16 bit/color files down to 8 bits/color */
	png_set_strip_16(png_ptr);

	int num_palette;
	png_colorp palette;
	RGBQUAD *rgb;
	switch(color_type){
	case PNG_COLOR_TYPE_GRAY: //0
		Dib->Create(width, height, 8);
		// Set gray palette
		Dib->SetMonoPalette();
		// Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel
		if(bit_depth < 8)
			png_set_expand(png_ptr);
		break;
	case PNG_COLOR_TYPE_RGB: //2
		Dib->Create(width, height, 24);
		break;
	case PNG_COLOR_TYPE_PALETTE: //3
		Dib->Create(width, height, 8);
		// Load palette
		if(png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette))
		{
			rgb = Dib->GetPalette();
			for(i=0; i<num_palette; i++)
			{
				rgb[i].rgbRed   = palette[i].red;
				rgb[i].rgbGreen = palette[i].green;
				rgb[i].rgbBlue  = palette[i].blue;
				rgb[i].rgbReserved = 0;
			}
		}
		break;
	case PNG_COLOR_TYPE_GRAY_ALPHA: //4
		Dib->Create(width, height, 8);
		// Load palette
		if(png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette))
		{
			rgb = Dib->GetPalette();
			for(i=0; i<num_palette; i++)
			{
				rgb[i].rgbRed = rgb[i].rgbGreen = rgb[i].rgbBlue = (BYTE)i;
				rgb[i].rgbReserved = 0;
			}
		}
		// Strip alpha bytes from the input data without combining with th
		// background (not recommended).
		png_set_strip_alpha(png_ptr);
		break;
	case PNG_COLOR_TYPE_RGB_ALPHA: //6
		Dib->Create(width, height, 32);
		break;
	}

	// Strip alpha bytes from the input data without combining with th
	// background (not recommended).
	//png_set_strip_alpha(png_ptr);

	// Extract multiple pixels with bit depths of 1, 2, and 4 from a single
	// byte into separate bytes (useful for paletted and grayscale images).
	png_set_packing(png_ptr);

	// Change the order of packed pixels to least significant bit first
	// (not useful if you are using png_set_packing).
	png_set_packswap(png_ptr);

	// Expand paletted colors into true RGB triplets
	//if(color_type == PNG_COLOR_TYPE_PALETTE)
	//	png_set_expand(png_ptr);

	// Expand grayscale images to the full 8 bits from 1, 2, or 4 bits/pixel
	//if(color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
	//	png_set_expand(png_ptr);

	// Expand paletted or RGB images with transparency to full alpha channels
	// so the data will be available as RGBA quartets.
	if(png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
		png_set_expand(png_ptr);

#if USE_GAMMA

   /* Set the background color to draw transparent and alpha images over.
	* It is possible to set the red, green, and blue components directly
	* for paletted images instead of supplying a palette index.  Note that
	* even if the PNG file supplies a background, you are not required to
	* use it - you should use the (solid) application background if it has one.
	*/

	png_color_16 my_background, *image_background;

	if (png_get_bKGD(png_ptr, info_ptr, &image_background))
		png_set_background(png_ptr, image_background,
			PNG_BACKGROUND_GAMMA_FILE, 1, 1.0);
	else
		png_set_background(png_ptr, &my_background,
			PNG_BACKGROUND_GAMMA_SCREEN, 0, 1.0);

	// Some suggestions as to how to get a screen gamma value

	// Note that screen gamma is (display_gamma/viewing_gamma)
	if (/* We have a user-defined screen gamma value */)
	{
		screen_gamma = user-defined screen_gamma;
	}
	// This is one way that applications share the same screen gamma value
	else if ((gamma_str = getenv("SCREEN_GAMMA")) != NULL)
	{
		screen_gamma = atof(gamma_str);
	}
	// If we don't have another value
	else
	{
		screen_gamma = 2.2;  /* A good guess for a PC monitors in a dimly
							  lit room */
		screen_gamma = 1.7 or 1.0;  /* A good guess for Mac systems */
	}

	/* Tell libpng to handle the gamma conversion for you.  The second call
	 * is a good guess for PC generated images, but it should be configurable
	 * by the user at run time by the user.  It is strongly suggested that
	 * your application support gamma correction.
	 */

	int intent;

	if (png_get_sRGB(png_ptr, info_ptr, &intent))
		png_set_sRGB(png_ptr, intent, 0);
	else
		if (png_get_gAMA(png_ptr, info_ptr, &image_gamma))
			png_set_gamma(png_ptr, screen_gamma, image_gamma);
		else
			png_set_gamma(png_ptr, screen_gamma, 0.50);
#endif //USE_GAMMA

#if 0
	// Dither RGB files down to 8 bit palette or reduce palettes
	// to the number of colors available on your screen.
	if (color_type & PNG_COLOR_MASK_COLOR)
	{
		png_uint_32 num_palette;
		png_colorp palette;

		// This reduces the image to the application supplied palette
		if (/* we have our own palette */)
		{
			// An array of colors to which the image should be dithered
			png_color std_color_cube[MAX_SCREEN_COLORS];

			png_set_dither(png_ptr, std_color_cube, MAX_SCREEN_COLORS,
				MAX_SCREEN_COLORS, NULL, 0);
		}
		/* This reduces the image to the palette supplied in the file */
		else if (png_get_PLTE(png_ptr, info_ptr, &palette, &num_palette))
		{
			png_color16p histogram;

			png_get_hIST(png_ptr, info_ptr, &histogram);

			png_set_dither(png_ptr, palette, num_palette,
				max_screen_colors, histogram, 0);
		}
	}

	// invert monocrome files to have 0 as white and 1 as black
	png_set_invert_mono(png_ptr);

	/* If you want to shift the pixel values from the range [0,255] or
	 * [0,65535] to the original [0,7] or [0,31], or whatever range the
	 * colors were originally in:
	 */
	if (png_get_valid(png_ptr, info_ptr, PNG_INFO_sBIT))
	{
		png_color8p sig_bit;

		png_get_sBIT(png_ptr, info_ptr, &sig_bit);
		png_set_shift(png_ptr, sig_bit);
	}
#endif

	// flip the RGB pixels to BGR (or RGBA to BGRA)
	//png_set_bgr(png_ptr);

	// swap the RGBA or GA data to ARGB or AG (or BGRA to ABGR)
	//png_set_swap_alpha(png_ptr);

	// swap bytes of 16 bit files to least significant byte first
	//png_set_swap(png_ptr);

	// Add filler (or alpha) byte (before/after each RGB triplet)
	//png_set_filler(png_ptr, 0xff, PNG_FILLER_AFTER);

	// Turn on interlace handling.  REQUIRED if you are not using
	// png_read_image().  To see how to handle interlacing passes,
	// see the png_read_row() method below:
	//number_passes = png_set_interlace_handling(png_ptr);

	// Optional call to gamma correct and add the background to the palette
	// and update info structure.  REQUIRED if you are expecting libpng to
	// update the palette for you (ie you selected such a transform above).
	//png_read_update_info(png_ptr, info_ptr);

	// Allocate the memory to hold the image using the fields of info_ptr.

	// The easiest way to read the image:
	png_bytep *row_pointers;
	row_pointers = (png_bytep *)malloc(sizeof(png_bytep) * height);
	for(row=0; row<height; row++)
		row_pointers[row] = (png_bytep)Dib->ScanLine(row);

	// Now it's time to read the image.  One of these methods is REQUIRED
#ifdef entire
	// Read the entire image in one go
	png_read_image(png_ptr, row_pointers);
#else
	if(callback){
		for(row=0; row<height; row++) {
			png_read_rows(png_ptr, &row_pointers[row], NULL, 1);
			if(!callback((float)row/height,data))
				goto FILE_EXIT;
		}
	} else {
		// Read the entire image in one go
		png_read_image(png_ptr, row_pointers);
	}
#endif

	free(row_pointers);

	// read rest of file, and get additional chunks in info_ptr - REQUIRED
	png_read_end(png_ptr, info_ptr);

	// RGB(A) -> BGR(A)
	switch(Dib->GetBpp()){
	case 24: {
		BYTE *p,t;
		for(unsigned y=0; y<height; y++){
			p = (BYTE *)Dib->ScanLine(y);
			for(unsigned x=0; x<width; x++,p+=3){
				t=p[2]; p[2]=p[0]; p[0]=t;
			}
		}
		break; }
	case 32: {
		BYTE *p,t;
		for(unsigned y=0; y<height; y++){
			p = (BYTE *)Dib->ScanLine(y);
			for(unsigned x=0; x<width; x++,p+=4){
				t=p[2]; p[2]=p[0]; p[0]=t;
			}
		}
		break; }
	}

	result = true;

FILE_EXIT:
	// clean up after the read, and free any memory allocated - REQUIRED
	png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);

	// close the file
	fclose(fh);

	return result;
}	//LoadPNG()

//---------------------------------------------------------------------------
//  SavePNG
//---------------------------------------------------------------------------
bool SavePNG(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	DibStream *TmpDib;
	png_structp png_ptr;
	png_infop info_ptr;
	png_uint_32 width, height, bit_depth;
	png_uint_32 x,y;
	int i,color_type;
	FILE *fh;
	BYTE *p,t;
	bool result = false;

	// create file
	fh = fopen(FileName, "wb");
	if(fh == NULL) return false;

	// Create and initialize the png_struct with the desired error handler
	// functions.  If you want to use the default stderr and longjump method,
	// you can supply NULL for the last three parameters.  We also check that
	// the library version is compatible with the one used at compile time,
	// in case we are using dynamically linked libraries.  REQUIRED.
	png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL);

	if(png_ptr == NULL)
	{
		fclose(fh);
		return false;
	}

	// Allocate/initialize the image information data.  REQUIRED
	info_ptr = png_create_info_struct(png_ptr);
	if(info_ptr == NULL)
	{
		fclose(fh);
		png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
		return false;
	}

	// Set error handling.  REQUIRED if you aren't supplying your own
	// error hadnling functions in the png_create_write_struct() call.
	if(setjmp(png_ptr->jmpbuf))
	{
		// If we get here, we had a problem reading the file
		fclose(fh);
		png_destroy_write_struct(&png_ptr,  (png_infopp)NULL);
		return false;
	}

	// create clone of DIB
	TmpDib = Dib->Clone();

	width  = TmpDib->GetWidth();
	height = TmpDib->GetHeight();

	switch(TmpDib->GetBpp()){
#if 1
	case 1:
		bit_depth = 1;
		color_type = PNG_COLOR_TYPE_PALETTE;
		break;
	case 4:
		bit_depth = 4;
		color_type = PNG_COLOR_TYPE_PALETTE;
		break;
	case 8:
		bit_depth = 8;
		color_type = PNG_COLOR_TYPE_PALETTE;
		break;
#else
	case 1:
	case 4:
	case 8:
		TmpDib->ChangeBpp(24); // PNG not support 555 format, so change 888
		goto BPP_24;
#endif
	case 16:
		TmpDib->ChangeBpp(24); // PNG not support 555 format, so change 888
		goto BPP_24;
	case 24:
	BPP_24:
		// BGR(A) -> RGB(A)
		for(y=0; y<height; y++){
			p = (BYTE *)TmpDib->ScanLine(y);
			for(x=0; x<width; x++,p+=3){
				t=p[2]; p[2]=p[0]; p[0]=t;
			}
		}
		bit_depth = 8;
		color_type = PNG_COLOR_TYPE_RGB;
		break;
	case 32:
		// BGR(A) -> RGB(A)
		for(y=0; y<height; y++){
			p = (BYTE *)TmpDib->ScanLine(y);
			for(x=0; x<width; x++,p+=4){
				t=p[2]; p[2]=p[0]; p[0]=t;
			}
		}
		bit_depth = 8;
		color_type = PNG_COLOR_TYPE_RGB_ALPHA;
		break;
	}

	// One of the following I/O initialization functions is REQUIRED
#ifdef streams /* I/O initialization method 1 */
	// set up the output control if you are using standard C streams */
	png_init_io(png_ptr, fh);
#else no_streams /* I/O initialization method 2 */
	// If you are using replacement read functions, instead of calling
	// png_init_io() here you would call
	png_set_write_fn(png_ptr, (void *)user_io_ptr, user_write_fn,
		user_IO_flush_function);
	// where user_io_ptr is a structure you want available to the callbacks
#endif no_streams /* only use one initialization method */

	// Set the image information here.  Width and height are up to 2^31,
	// bit_depth is one of 1, 2, 4, 8, or 16, but valid values also depend on
	// the color_type selected. color_type is one of PNG_COLOR_TYPE_GRAY,
	// PNG_COLOR_TYPE_GRAY_ALPHA, PNG_COLOR_TYPE_PALETTE, PNG_COLOR_TYPE_RGB,
	// or PNG_COLOR_TYPE_RGB_ALPHA.  interlace is either PNG_INTERLACE_NONE or
	// PNG_INTERLACE_ADAM7, and the compression_type and filter_type MUST
	// currently be PNG_COMPRESSION_TYPE_BASE and PNG_FILTER_TYPE_BASE. REQUIRED
	png_set_IHDR(png_ptr, info_ptr, width, height, bit_depth, color_type,
		PNG_INTERLACE_NONE, PNG_COMPRESSION_TYPE_BASE, PNG_FILTER_TYPE_BASE);

	// set palette
	if(TmpDib->GetBpp() <= 8)
	{
		int colnum = TmpDib->GetNumColors();
		if(colnum > 0)
		{
			png_color *pal = (png_color *)malloc(colnum * sizeof(png_color));
			RGBQUAD *rgb = TmpDib->GetColors();
			for(i=0; i<colnum; i++){
				pal[i].red   = rgb[i].rgbRed;
				pal[i].green = rgb[i].rgbGreen;
				pal[i].blue  = rgb[i].rgbBlue;
			}

			png_set_PLTE(png_ptr, info_ptr, pal, colnum);
		}
	}

#if 0
	// optional significant bit chunk
	// if we are dealing with a grayscale image then
	sig_bit.gray = true_bit_depth;
	// otherwise, if we are dealing with a color image then
	sig_bit.red = true_red_bit_depth;
	sig_bit.green = true_green_bit_depth;
	sig_bit.blue = true_blue_bit_depth;
	// if the image has an alpha channel then
	sig_bit.alpha = true_alpha_bit_depth;
	png_set_sBIT(png_ptr, info_ptr, sig_bit);
#endif

	// Optional gamma chunk is strongly suggested if you have any guess
	// as to the correct gamma of the image.
	//png_set_gAMA(png_ptr, info_ptr, gamma);

#if 0
	// Optionally write comments into the image
	text_ptr[0].key = "Title";
	text_ptr[0].text = "Mona Lisa";
	text_ptr[0].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[1].key = "Author";
	text_ptr[1].text = "Leonardo DaVinci";
	text_ptr[1].compression = PNG_TEXT_COMPRESSION_NONE;
	text_ptr[2].key = "Description";
	text_ptr[2].text = "<long text>";
	text_ptr[2].compression = PNG_TEXT_COMPRESSION_zTXt;
	png_set_text(png_ptr, info_ptr, text_ptr, 2);
#endif

	// other optional chunks like cHRM, bKGD, tRNS, tIME, oFFs, pHYs,
	// note that if sRGB is present the cHRM chunk must be ignored
	// on read and must be written in accordance with the sRGB profile

	// Write the file header information.  REQUIRED
	png_write_info(png_ptr, info_ptr);

	// Once we write out the header, the compression type on the text
	// chunks gets changed to PNG_TEXT_COMPRESSION_NONE_WR or
	// PNG_TEXT_COMPRESSION_zTXt_WR, so it doesn't get written out again
	// at the end.

	// set up the transformations you want.  Note that these are
	// all optional.  Only call them if you want them.

	// invert monocrome pixels
	//png_set_invert_mono(png_ptr);

	// Shift the pixels up to a legal bit depth and fill in
	// as appropriate to correctly scale the image.
	//png_set_shift(png_ptr, &sig_bit);

	// pack pixels into bytes
	//png_set_packing(png_ptr);

	// swap location of alpha bytes from ARGB to RGBA
	//png_set_swap_alpha(png_ptr);

	// Get rid of filler (OR ALPHA) bytes, pack XRGB/RGBX/ARGB/RGBA into
	// RGB (4 channels -> 3 channels). The second parameter is not used.
	//png_set_filler(png_ptr, 0, PNG_FILLER_BEFORE);

	// flip BGR pixels to RGB
	//png_set_bgr(png_ptr);

	// swap bytes of 16-bit files to most significant byte first
	//png_set_swap(png_ptr);

	// swap bits of 1, 2, 4 bit packed pixel formats
	//png_set_packswap(png_ptr);

#if 0
	// turn on interlace handling if you are not using png_write_image()
	if(interlacing)
		number_passes = png_set_interlace_handling(png_ptr);
	else
		number_passes = 1;
#endif

	// The easiest way to write the image (you may have a different memory
	// layout, however, so choose what fits your needs best).  You need to
	// use the first method if you aren't handling interlacing yourself.

	//One of the following output methods is REQUIRED

	png_bytep *row_pointers;
	row_pointers = (png_bytep *)malloc(sizeof(png_bytep) * height);
	// set line (top-down)
	for(y=0; y<height; y++)
		row_pointers[y] = (png_bytep)TmpDib->ScanLine(y);
#ifdef entire
	png_write_image(png_ptr, row_pointers);
#else
	// write out the entire image data in one call
	if(callback)
	{
		// If you are only writing one row at a time, this works
		for(y=0; y<height; y++)
		{
			png_write_rows(png_ptr, &row_pointers[y], 1);
			if(!callback((float)y/height,data))
				goto FILE_EXIT;
		}
	}
	else
	{
		png_write_image(png_ptr, row_pointers);
	}
#endif
	free(row_pointers);

	// You can write optional chunks like tEXt, zTXt, and tIME at the end
	// as well.

	// It is REQUIRED to call this to finish writing the rest of the file
	png_write_end(png_ptr, info_ptr);

	result = true;

#ifndef entire
FILE_EXIT:
#endif
	// clean up after the write, and free any memory allocated
	png_destroy_write_struct(&png_ptr, (png_infopp)NULL);

	delete TmpDib;

	// close the file
	fclose(fh);

	return result;
}	//SavePNG()

#else

typedef struct png_callback_info {
	bool (*callback)(float,void*);
	void *data;
} png_callback_info;

static BOOL PngCallback(float progress, void *data)
{
	png_callback_info *info = (png_callback_info*)data;
	if(info->callback != NULL)
		return info->callback(progress, info->data);
	return TRUE;
}

//---------------------------------------------------------------------------
//  LoadPNG
//---------------------------------------------------------------------------
bool LoadPNG(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	lpLoadPNG *png_proc;
	HINSTANCE hInst;
	bool result = false;

	hInst = LoadLibrary("mlibpng.dll");
	if(hInst == NULL)
		return false;

	*(FARPROC *)&png_proc = GetProcAddress(hInst, "LoadPNG");
	if(png_proc != NULL)
	{
		void *Header, *Buffer;
		png_callback_info info;

		info.callback = callback;
		info.data = data;
		if(png_proc(FileName, &Header, &Buffer, PngCallback, &info))
		{
			Dib->SetDib(
				Header, DIB_MEMORYTYPE_GLOBALFIXED,
				Buffer, DIB_MEMORYTYPE_GLOBALFIXED);
			result = true;
		}
	}

	FreeLibrary(hInst);

	return result;
}

//---------------------------------------------------------------------------
//  SavePNG
//---------------------------------------------------------------------------
bool SavePNG(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	lpSavePNG *png_proc;
	HINSTANCE hInst;
	bool result = false;

	hInst = LoadLibrary("mlibpng.dll");
	if(hInst == NULL)
		return false;

	*(FARPROC *)&png_proc = GetProcAddress(hInst, "SavePNG");
	if(png_proc != NULL)
	{
		png_callback_info info;
		info.callback = callback;
		info.data = data;
		if(png_proc(FileName, Dib->GetHeader(), Dib->GetBuffer(), PngCallback, &info))
		{
			result = true;
		}
	}

	FreeLibrary(hInst);

	return result;
}

#endif
