//---------------------------------------------------------------------------
//   DIB Library
//         Copyright(c)1998-99, O.Mizno     All rights reserved.
//
//     DibTim.cpp
//       LoadTIM()
//---------------------------------------------------------------------------

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <stdio.h>
#include "Dib.h"


//---------------------------------------------------------------------------
//  LoadTIM
//---------------------------------------------------------------------------
bool LoadTIM(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	bool result=false;
	FILE *fh;
	DWORD flag;
	WORD dw;
	RGBQUAD rgb[256];
	int bpp;
	int width,height,pitch;
	long chunk_begin, chunk_size;
	int x,y,i;

	if((fh = fopen(FileName, "rb")) == NULL)
		return false;

	// check ID
	fread(&flag, 4, 1, fh);
	if(flag != 0x00000010)
		goto LOAD_END;

	// check flag
	fread(&flag, 4, 1, fh);
	switch(flag & 0x07){
	case 0: bpp= 4; break;
	case 1: bpp= 8; break;
	case 2: bpp=16; break;
	//case 3: bpp=24; break;
	default: goto LOAD_END;
	}
	if(flag & 0x08)
	{	// read CULT (palette)
		chunk_begin = ftell(fh);
		fread(&chunk_size, 4, 1, fh);
		fseek(fh, 4, SEEK_CUR); //ignore DX,DY

		fread(&dw, 2, 1, fh); width = (int)dw;
		fread(&dw, 2, 1, fh); width *= (int)dw;

		for(i=0; i<width; i++)
		{
			fread(&dw, 2, 1, fh);
			rgb[i].rgbRed =   (BYTE)((dw & 0x001F) << 3);
			rgb[i].rgbGreen = (BYTE)((dw & 0x03E0) >> 2);
			rgb[i].rgbBlue =  (BYTE)((dw & 0x7C00) >> 7);
			rgb[i].rgbReserved = 0;
		}

		fseek(fh, chunk_begin+chunk_size, SEEK_SET);
	}

	chunk_begin = ftell(fh);
	fread(&chunk_size, 4, 1, fh);
	fseek(fh, 4, SEEK_CUR); //ignore DX,DY
	fread(&dw, 2, 1, fh); pitch  = (int)dw;
	fread(&dw, 2, 1, fh); height = (int)dw;
	switch(bpp){
	case  4: width = pitch << 2; break;
	case  8: width = pitch << 1; break;
	case 16: width = pitch; break;
	}
	Dib->Create(width, height, bpp);

	switch(bpp){
	case 4:
		memcpy(Dib->GetPalette(), rgb, sizeof(RGBQUAD) * 16);
		break;
	case 8:
		memcpy(Dib->GetPalette(), rgb, sizeof(RGBQUAD) * 256);
		break;
	}

	for(y=0; y<height; y++)
	{
		WORD *wp = (WORD *)Dib->ScanLine(y);
		switch(bpp){
		case 4:
			for(x=0; x<pitch; x++,wp++){
				fread(&dw, 2, 1, fh);
				*wp = (WORD)(
					((dw & 0x0F0F) << 4) |
					((dw & 0xF0F0) >> 4));
			}
			break;
		case 8:
			for(x=0; x<pitch; x++,wp++){
				fread(&dw, 2, 1, fh);
				*wp = dw;
			}
			break;
		case 16:
			for(x=0; x<pitch; x++,wp++){
				fread(&dw, 2, 1, fh);
				*wp = (WORD)(
					((dw & 0x001F) << 10) |
					((dw & 0x03E0)      ) |
					((dw & 0x7C00) >> 10));
			}
			break;
		}
	}

	result = true;

LOAD_END:
	fclose(fh);

	return result;
}

//---------------------------------------------------------------------------
//  SaveTIM
//---------------------------------------------------------------------------
#define W2(d) {dw=(d); fwrite(&dw, 2, 1, fh);}
#define W4(d) {ddw=(d); fwrite(&ddw, 4, 1, fh);}


bool SaveTIM(LPCSTR FileName, DibStream *Dib, bool callback(float,void*), void *data)
{
	bool result=false;
	FILE *fh;
	DWORD ddw;
	WORD dw,*rp;
	DibStream *TmpDib = Dib;
	RGBQUAD *rgb;
	int palette_count;
	WORD height,pitch,x,y;
	int i;

	fh = fopen(FileName, "wb");
	if(fh == NULL)
		return false;

	// check ID
	W4(0x00000010);

	switch(Dib->GetBpp()){
	case 4:
	case 8:
	case 16:
		break;
	default:
		TmpDib = Dib->Clone();
		TmpDib->ChangeBpp(16);
	}

	// flag
	switch(TmpDib->GetBpp()){
	case  4: ddw=8; break;
	case  8: ddw=9; break;
	case 16: ddw=2; break;
	}
	fwrite(&ddw, 4, 1, fh);

	//CLUT
	switch(TmpDib->GetBpp()){
	case 4:
		W4(4+8+2*16);
		W2(0); W2(0);
		W2(16); W2(1);
		palette_count = TmpDib->GetNumColors();
		rgb = TmpDib->GetPalette();
		for(i=0; i<palette_count; i++){
			W2((WORD)(
				((rgb[i].rgbRed   >> 3) & 0x001F) |
				((rgb[i].rgbGreen << 2) & 0x03E0) |
				((rgb[i].rgbBlue  << 7) & 0x7C00)));
		}
		for(; i<16; i++){
			W2(0);
		}
		break;
	case 8:
		W4(4+8+2*256);
		W2(0); W2(0);
		W2(256); W2(1);
		palette_count = TmpDib->GetNumColors();
		rgb = TmpDib->GetPalette();
		for(i=0; i<palette_count; i++){
			W2((WORD)(
				((rgb[i].rgbRed   >> 3) & 0x001F) |
				((rgb[i].rgbGreen << 2) & 0x03E0) |
				((rgb[i].rgbBlue  << 7) & 0x7C00)));
		}
		for(; i<256; i++){
			W2(0);
		}
		break;
	}

	//PIXEL
	height = (WORD)TmpDib->GetHeight();
	switch(TmpDib->GetBpp()){
	case 4:
		pitch = (WORD)((TmpDib->GetWidth() + 3) >> 2);
		W4(4+8+pitch*height*2);
		W2(0); W2(0);
		W2(pitch); W2(height);
		for(y=0; y<height; y++){
			rp = (WORD *)TmpDib->ScanLine(y);
			for(x=0; x<pitch; x++,rp++){
				W2((WORD)(((*rp & 0xF0F0) >> 4) | ((*rp & 0x0F0F) << 4)));
			}
		}
		break;
	case 8:
		pitch = (WORD)((TmpDib->GetWidth() + 1) >> 1);
		W4(4+8+pitch*height*2);
		W2(0); W2(0);
		W2(pitch); W2(height);
		for(y=0; y<height; y++){
			rp = (WORD *)TmpDib->ScanLine(y);
			for(x=0; x<pitch; x++,rp++){
				W2(*rp);
			}
		}
		break;
	case 16:
		pitch = (WORD)TmpDib->GetWidth();
		W4(4+8+pitch*height*2);
		W2(0); W2(0);
		W2(pitch); W2(height);
		for(y=0; y<height; y++){
			rp = (WORD *)TmpDib->ScanLine(y);
			for(x=0; x<pitch; x++,rp++){
				W2((WORD)(
					((*rp & 0x001F) << 10) |
					((*rp & 0x03E0)      ) |
					((*rp & 0x7C00) >> 10)));
			}
		}
		break;
	}

	if(TmpDib != Dib)
		delete TmpDib;

	fclose(fh);

	return result;
}




